/*

                          Argonne National Laboratory
                            APS Operations Division
                     Beamline Controls and Data Acquisition

                     EPICS Brick Scaler Asyn Device Support



 -----------------------------------------------------------------------------
                                COPYRIGHT NOTICE
 -----------------------------------------------------------------------------
   Copyright (c) 2002 The University of Chicago, as Operator of Argonne
      National Laboratory.
   Copyright (c) 2002 The Regents of the University of California, as
      Operator of Los Alamos National Laboratory.
   Synapps Versions 4-5
   and higher are distributed subject to a Software License Agreement found
   in file LICENSE that is included with this distribution.
 -----------------------------------------------------------------------------

 Description
    This module provides device support support for a multiple device port
    driver drvScaler to be used in conjunction with the scaler record. To
    initialize device support, call the method below with the following
    calling sequence.

        devScalerInit(port,type,inst,scls,rate)

        Where:
            port - Scaler port driver name (i.e. "SCAL")
            type - Board type (see boardtypes.h)
            inst - Board instance
            scls - Number of scalers
            rate - Poll frequency (i.e. 5 is 5Hz)


 Developer notes:

 Source control info:
    Modified by:    $Author: dkline $
                    $Date: 2006/07/24 19:02:29 $
                    $Revision: 1.1 $

 =============================================================================
 History:
 Author: David M. Kline
 -----------------------------------------------------------------------------
 2006-Jul-01  DMK  Derived from the Joerger Scaler devices support.
 2006-Jul-24  DMK  Initial development complete.
 -----------------------------------------------------------------------------

*/


/* EPICS base version-specific definitions (must be performed first) */
#include <epicsVersion.h>
#define LT_EPICSBASE(v,r,l) (EPICS_VERSION<(v)||(EPICS_VERSION==(v)&&(EPICS_REVISION<(r)||(EPICS_REVISION==(r)&&EPICS_MODIFICATION<(l)))))


/* Evaluate EPICS base */
#if LT_EPICSBASE(3,14,6)
    #error "EPICS base must be 3.14.6 or greater"
#endif

/* System related include files */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* EPICS system related include files */
#include <iocsh.h>
#include <epicsStdio.h>
#include <cantProceed.h>
#include <epicsString.h>
#include <epicsExport.h>
#include <epicsThread.h>

#include <cadef.h>
#include <recGbl.h>
#include <devSup.h>
#include <dbEvent.h>
#include <callback.h>


/* EPICS synApps/Asyn related include files */
#include <asynInt32.h>
#include <devScaler.h>
#include <asynDriver.h>
#include <asynDrvUser.h>
#include <scalerRecord.h>
#include <asynUInt32Digital.h>


/* EPICS Brick related include files */
#include "dscsup.h"
#include "drvScaler.h"


/* Forward references for device support struct */
static long scal_report(int level);
static long scal_hdw_init(int after);
static long scal_rec_done(int inst);
static long scal_rec_reset(int inst);
static long scal_rec_arm(int inst,int val);
static long scal_rec_read(int inst,long* val);
static long scal_rec_init(scalerRecord* pscal);
static long scal_rec_preset(int inst,int signal,long val);

static SCALERDSET devAsynScaler =
{
    8,
    (long int (*)(void*))scal_report,
    (long int (*)(void*))scal_hdw_init,
    (long int (*)(void*))scal_rec_init,
    NULL,
    (long int (*)(void*))scal_rec_reset,
    (long int (*)(void*))scal_rec_read,
    (long int (*)(void*))scal_rec_preset,
    (long int (*)(void*))scal_rec_arm,
    (long int (*)(void*))scal_rec_done,
};
epicsExportAddress(dset,devAsynScaler);


/* Forward struct declarations */
typedef struct Port Port;
typedef struct Scaler Scaler;

/* Declare Scaler structure */
struct Scaler
{
    int                conn;
    int                addr;
    epicsInt32         value;
    epicsInt32         preset;

    asynUser*          pasynUser;
    asynInt32*         pasynInt32;
    void*              pasynInt32Pvt;
    asynUInt32Digital* pasynUInt32;
    void*              pasynUInt32Pvt;
};

/* Declare Port structure */
struct Port
{
    char*   name;
    int     inst;
    int     rate;
    int     scls;
    double  freq;
    Scaler  mgnt;
    Scaler* scal;

    int canscan;
    int ispreset;

    int done;
    CALLBACK* pcallback;
};


/* Public interface forward references */
static int devScalerInit(const char* port,int type,int inst,int scls,int rate);


/* Private methods forward references */
static void scanScaler(Port* pport);
static int connectScaler(Port* pport);

/* Define macros */
#define ISOK(s)		(asynSuccess==(s))
#define ISNOTOK(s)	(!ISOK(s))


/* Define local variants */
static epicsUInt32 mask = 0xFF;
static Port* pports[] = {NULL,NULL,NULL,NULL};


/****************************************************************************
 * Define public interface methods
 ****************************************************************************/
static int devScalerInit(const char* port,int type,int inst,int scls,int rate)
{
    int len;
    uint prio;
    Port* pport;

    if( inst > 3 )
    {
        printf("devScalerInit::init %s: invalid instance count %d\n",port,inst);
        return( 0 );
    }

    if( scls > 16 )
    {
        printf("devScalerInit::init %s: invalid scaler count %d\n",port,scls);
        return( 0 );
    }

    switch( type )
    {
    case SCL_GDIO:
        break;

    default:
        printf("devScalerInit::init %s: unknown board type %d\n",port,type);
        return( -1 );
    }

    if( pports[inst] )
    {
        printf("devScalerInit::init %s: instance %d already defined\n",port,inst);
        return( 0 );
    }

    len = strlen(port)+1;
    len += sizeof(Port)+(sizeof(Scaler)*scls);
    pport = (Port*)callocMustSucceed(len,sizeof(char),"devScalerInit");

    pport->scal = (Scaler*)(pport+1);
    pport->name = (char*)(pport->scal+scls);

    pport->inst = inst;
    pport->scls = scls;
    pport->rate = rate;
    strcpy(pport->name,port);

    if( connectScaler(pport) )
    {
        printf("devScalerInit::init %s: failure to connect to scalers\n",port);
        free(pport);
        return( -1 );
    }

    pports[inst] = pport;
    epicsThreadLowestPriorityLevelAbove(epicsThreadPriorityScanHigh,&prio);
    epicsThreadCreate("scanScaler",prio,epicsThreadGetStackSize(epicsThreadStackSmall),(EPICSTHREADFUNC)scanScaler,pport);

    return( 0 );
}


static int connectScaler(Port* pport)
{
    int i;
    Scaler* psclr;
    asynStatus sts;
    asynUser* pasynUser;
    asynInterface* pasynIface;


    /* Connect to management scaler (global control) */
    i = 32;
    psclr = &pport->mgnt;
    pasynUser = pasynManager->createAsynUser(NULL,NULL);
    if( pasynUser == NULL )
        return( -1 );

    sts = pasynManager->connectDevice(pasynUser,pport->name,i);
    if( ISNOTOK(sts) )
    {
        printf("connectScaler::failure to connect with device %s @ %d\n",pport->name,i);
        pasynManager->freeAsynUser(pasynUser);
        return( -1 );
    }

    pasynIface = pasynManager->findInterface(pasynUser,asynUInt32DigitalType,1);
    if( pasynIface )
    {
        psclr->pasynUInt32 = (asynUInt32Digital*)pasynIface->pinterface;
        psclr->pasynUInt32Pvt = pasynIface->drvPvt;
    }
    else
    {
        printf("connectScaler::failure to find interface %s\n",asynUInt32DigitalType);
        pasynManager->disconnect(pasynUser);
        pasynManager->freeAsynUser(pasynUser);
        return( -1 );
    }

    pasynIface = pasynManager->findInterface(pasynUser,asynInt32Type,1);
    if( pasynIface )
    {
        psclr->pasynInt32 = (asynInt32*)pasynIface->pinterface;
        psclr->pasynInt32Pvt = pasynIface->drvPvt;
    }
    else
    {
        printf("connectScaler::failure to find interface %s\n",asynInt32Type);
        pasynManager->disconnect(pasynUser);
        pasynManager->freeAsynUser(pasynUser);
        return( -1 );
    }

    psclr->addr = i;
    psclr->conn = TRUE;
    psclr->pasynUser = pasynUser;

    /* Connect to individual scalers */
    for( i=0; i<pport->scls; ++i )
    {
        psclr = &pport->scal[i];

        pasynUser = pasynManager->createAsynUser(NULL,NULL);
        if( pasynUser == NULL )
            return( -1 );

        sts = pasynManager->connectDevice(pasynUser,pport->name,i);
        if( ISNOTOK(sts) )
        {
            printf("connectScaler::failure to connect with device %s @ %d\n",pport->name,i);
            pasynManager->freeAsynUser(pasynUser);
            return( -1 );
        }

        pasynIface = pasynManager->findInterface(pasynUser,asynUInt32DigitalType,1);
        if( pasynIface )
        {
            psclr->pasynUInt32 = (asynUInt32Digital*)pasynIface->pinterface;
            psclr->pasynUInt32Pvt = pasynIface->drvPvt;
        }
        else
        {
            printf("connectScaler::failure to find interface %s\n",asynUInt32DigitalType);
            pasynManager->disconnect(pasynUser);
            pasynManager->freeAsynUser(pasynUser);
            return( -1 );
        }

        pasynIface = pasynManager->findInterface(pasynUser,asynInt32Type,1);
        if( pasynIface )
        {
            psclr->pasynInt32 = (asynInt32*)pasynIface->pinterface;
            psclr->pasynInt32Pvt = pasynIface->drvPvt;
        }
        else
        {
            printf("connectScaler::failure to find interface %s\n",asynInt32Type);
            pasynManager->disconnect(pasynUser);
            pasynManager->freeAsynUser(pasynUser);
            return( -1 );
        }

        psclr->addr = i;
        psclr->conn = TRUE;
        psclr->pasynUser = pasynUser;
    }    

    return( 0 );
}


static void scanScaler(Port* pport)
{
    int i;
    Scaler* psclr;
    double freq = (0.95/pport->rate);

    while(0==0)
    {
        epicsThreadSleep(freq);

        if( pport->canscan == FALSE )
            continue;

        for( i=0,psclr=&pport->scal[0]; i<pport->scls; ++i,++psclr )
            psclr->pasynInt32->read(psclr->pasynInt32Pvt,psclr->pasynUser,&psclr->value);

        if( pport->ispreset == FALSE )
            continue;

        for( i=0,psclr=&pport->scal[0]; i<pport->scls; ++i,++psclr )
            if( psclr->preset && psclr->value == 0 )
            {
                pport->done = 1;
                pport->ispreset = FALSE;
                callbackRequest(pport->pcallback);

                break;
            }
    }
   
}


/****************************************************************************
 * Define device support private interface
 ****************************************************************************/
static long scal_report(int level)
{
    int i,j;
    Port* pport;
    Scaler* psclr;

    for( i=0,pport=pports[0]; pports[i]; ++i,pport=pports[i] )
    {
        printf("%s-%d has %d scalers, %fMHz clock freq, poll rate of %d/sec\n",pport->name,pport->inst,pport->scls,(pport->freq/1000000.0),pport->rate);
        for( j=0,psclr=&pport->scal[0]; j<pport->scls; ++j,++psclr )
            printf("    Scaler %2.2d connected %s, preset %d, value %d\n",psclr->addr,(psclr->conn)?"Yes":"No",psclr->preset,psclr->value);
        printf("    Scaler 32 connected %s (module management)\n",(pport->mgnt.conn)?"Yes":"No");
    }

    return( 0 );
}


static long scal_hdw_init(int after)
{
    int i,j;
    Port* pport;
    Scaler *psclr;

    if( after )
        return( 0 );

    for( i=0,pport=pports[0]; pports[i]; ++i,pport=pports[i] )
    {
        pport->done = 0;

        psclr = &pport->mgnt;
        psclr->pasynUInt32->write(psclr->pasynUInt32Pvt,psclr->pasynUser,CMD_RES,mask);

        for( j=0,psclr=&pport->scal[0]; j<pport->scls; ++j,++psclr )
            psclr->pasynUInt32->write(psclr->pasynUInt32Pvt,psclr->pasynUser,CMD_RES,mask);
    }

    return( 0 );
}


static long scal_rec_init(scalerRecord* pscal)
{
    Port* pport;
    CALLBACK *pcallbacks = (CALLBACK*)pscal->dpvt;

    switch( pscal->out.type )
    {
    case VME_IO:
        if( pscal->out.value.vmeio.card > 3 )
        {
            recGblRecordError(S_dev_badBus,(void*)pscal,"devAsynScaler: scal_rec_init - bad card");
            return(S_dev_badCard);
        }

        pport = pports[pscal->out.value.vmeio.card];
        break;

    case INST_IO:
        recGblRecordError(S_dev_badBus,(void*)pscal,"devAsynScaler: scal_rec_init - unsupported bus type");
        return(S_dev_badBus);

    default:
        recGblRecordError(S_dev_badBus,(void*)pscal,"devAsynScaler: scal_rec_init - bad bus type");
        return(S_dev_badBus);
    }

    pscal->nch = pport->scls;

    pport->pcallback = (CALLBACK*)&(pcallbacks[3]);
    pport->freq = pscal->freq;
    pport->ispreset = FALSE;
    pport->canscan = TRUE;

    return( 0 );
}


static long scal_rec_reset(int inst)
{
    int i;
    Scaler* psclr;
    Port* pport = pports[inst];

    for( i=0,psclr=&pport->scal[0]; i<pport->scls; ++i,++psclr )
    {
        psclr->preset = 0;
        psclr->pasynUInt32->write(psclr->pasynUInt32Pvt,psclr->pasynUser,CMD_RES,mask);
        psclr->pasynUInt32->write(psclr->pasynUInt32Pvt,psclr->pasynUser,CMD_UP,mask);
    }
    pport->done = 0;
    pport->ispreset = FALSE;

    return( 0 );
}


static long scal_rec_read(int inst,long* val)
{
    int i;
    Port* pport = pports[inst];

    for( i=0; i<pport->scls; ++i )
        val[i] = (long)pport->scal[i].preset + (long)pport->scal[i].value;

    return( 0 );
}


static long scal_rec_preset(int inst,int signal,long val)
{
    epicsInt32 preset;
    Port* pport = pports[inst];
    Scaler* psclr = &pport->scal[signal];

    preset = (epicsInt32)val * -1;
    psclr->preset = (epicsInt32)val;
    psclr->pasynInt32->write(psclr->pasynInt32Pvt,psclr->pasynUser,preset);
    if( val > 0 )
    {
        psclr->pasynUInt32->write(psclr->pasynUInt32Pvt,psclr->pasynUser,CMD_GEA,mask);
        psclr->pasynUInt32->write(psclr->pasynUInt32Pvt,psclr->pasynUser,CMD_UP,mask);
    }
    else if ( val < 0 )
    {
        psclr->pasynUInt32->write(psclr->pasynUInt32Pvt,psclr->pasynUser,CMD_GEA,mask);
        psclr->pasynUInt32->write(psclr->pasynUInt32Pvt,psclr->pasynUser,CMD_DWN,mask);
    }
    else
        psclr->pasynUInt32->write(psclr->pasynUInt32Pvt,psclr->pasynUser,CMD_UP,mask);

    return( 0 );
}


static long scal_rec_arm(int inst,int val)
{
    int i;
    Scaler* psclr;
    epicsUInt32 value;
    Port* pport = pports[inst];

    if( val )
    {
        value = CMD_ENA;
        pport->done = 0;
        pport->ispreset = TRUE;
    }
    else
    {
        value = CMD_DIS;
        pport->done = 1;
        pport->ispreset = FALSE;
    }

    for( i=0,psclr=&pport->scal[0]; i<pport->scls; ++i, ++psclr )
        psclr->pasynUInt32->write(psclr->pasynUInt32Pvt,psclr->pasynUser,value,mask);

    psclr = &pport->mgnt;
    psclr->pasynUInt32->write(psclr->pasynUInt32Pvt,psclr->pasynUser,value,mask);

    return( 0 );
}


static long scal_rec_done(int inst)
{
    if( pports[inst]->done )
    {
        pports[inst]->done = 0;
        pports[inst]->ispreset = FALSE;

        return( 1 );
    }

    return( 0 );
}


/****************************************************************************
 * Register public methods
 ****************************************************************************/
/* Initialization method definitions */
static const iocshArg devScalerInitArg0 = {"port",iocshArgString};
static const iocshArg devScalerInitArg1 = {"type",iocshArgInt};
static const iocshArg devScalerInitArg2 = {"inst",iocshArgInt};
static const iocshArg devScalerInitArg3 = {"scls",iocshArgInt};
static const iocshArg devScalerInitArg4 = {"rate",iocshArgInt};
static const iocshArg* devScalerInitArgs[]= {&devScalerInitArg0,&devScalerInitArg1,&devScalerInitArg2,&devScalerInitArg3,&devScalerInitArg4};
static const iocshFuncDef devScalerInitFuncDef = {"devScalerInit",5,devScalerInitArgs};
static void devScalerInitCallFunc(const iocshArgBuf* args)
{
    devScalerInit(args[0].sval,args[1].ival,args[2].ival,args[3].ival,args[4].ival);
}

/* Registration method */
static void devScalerRegister(void)
{
    static int firstTime = 1;

    if( firstTime )
    {
        firstTime = 0;
        iocshRegister(&devScalerInitFuncDef,devScalerInitCallFunc);
    }
}
epicsExportRegistrar( devScalerRegister );
