/*

                          Argonne National Laboratory
                            APS Operations Division
                     Beamline Controls and Data Acquisition

                          EPICS Brick ADC Port Driver



 -----------------------------------------------------------------------------
                                COPYRIGHT NOTICE
 -----------------------------------------------------------------------------
   Copyright (c) 2002 The University of Chicago, as Operator of Argonne
      National Laboratory.
   Copyright (c) 2002 The Regents of the University of California, as
      Operator of Los Alamos National Laboratory.
   Synapps Versions 4-5
   and higher are distributed subject to a Software License Agreement found
   in file LICENSE that is included with this distribution.
 -----------------------------------------------------------------------------

 Description
    This module provides support for a multiple device port driver. To
    initialize the driver, the method drvAdcInit() is called from the
    startup script with the following calling sequence.

        drvAdcInit(port,type,inst,addr,freq,conv,gain)

        Where:
            port - Pport driver name (i.e. "ADC" )
            type - Board type (see dscsup.h)
            inst - Board instance
            addr - Board address, zero for Athena
            freq - Sample frequency
            conv - Conversion, either Bipolar (0) or Unipolar (1)
            gain - Input gain (1,2,4,8)


    The method dbior can be called from the IOC shell to display the current
    status of the driver.


 Developer notes:


 Source control info:
    Modified by:    $Author: dkline $
                    $Date: 2006/05/26 20:36:41 $
                    $Revision: 1.2 $

 =============================================================================
 History:
 Author: David M. Kline
 -----------------------------------------------------------------------------
 2005-Nov-09  DMK  Derived from love multidevice driver.
 2005-Nov-15  DMK  Initial development complete.
 2005-Nov-30  DMK  Add scan thread @ 20Hz.
 2005-Dec-01  DMK  Provide configuration flexibility to the initialization.
 2006-May-19  DMK  Added parameters to initialization method.
 -----------------------------------------------------------------------------

*/


/* EPICS base version-specific definitions (must be performed first) */
#include <epicsVersion.h>
#define LT_EPICSBASE(v,r,l) (EPICS_VERSION<(v)||(EPICS_VERSION==(v)&&(EPICS_REVISION<(r)||(EPICS_REVISION==(r)&&EPICS_MODIFICATION<(l)))))


/* Evaluate EPICS base */
#if LT_EPICSBASE(3,14,6)
    #error "EPICS base must be 3.14.6 or greater"
#endif

/* System related include files */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* EPICS system related include files */
#include <iocsh.h>
#include <epicsStdio.h>
#include <cantProceed.h>
#include <epicsString.h>
#include <epicsExport.h>
#include <epicsThread.h>


/* EPICS synApps/Asyn related include files */
#include <asynDriver.h>
#include <asynInt32.h>
#include <asynDrvUser.h>


/* DSC related include files */
#include "dscsup.h"


/* Define symbolic constants */
#define TRUE       (1)
#define FALSE      (0)
#define K_COMTMO   (1)

/* Forward struct declarations */
typedef struct Ainp Ainp;
typedef struct Port Port;


/* Declare register (A,B,C) info structure */
struct Ainp
{
    int isConn;
    int regNum;
};


/* Declare DSC port structure */
struct Port
{
    char*         name;
    int           type;
    int           inst;
    int           addr;
    int           freq;
    int           conv;
    char          gain;
    char          chns;
    int           isConn;
    epicsMutexId  syncLock;
    asynInterface asynInt32;
    asynInterface asynCommon;
    asynInterface asynDrvUser;
    Ainp          adc[ATH_ADC_CHANS];
};


/* Define local variants */
static Port* pports = NULL;
static void scanADC(Port* pport);
static epicsInt32 adcData[ATH_ADC_CHANS];
static epicsFloat32 adcGain[][4] = {{10.0,5.0,2.5,1.25},{0.0,8.3,5.0,2.5}};


/* Public interface forward references */
static int drvAdcInit(const char* port,int type,int inst,int addr,int freq,int conv,int gain);


/* Forward references for asynCommon methods */
static void report(void* ppvt,FILE* fp,int details);
static asynStatus connect(void* ppvt,asynUser* pasynUser);
static asynStatus disconnect(void* ppvt,asynUser* pasynUser);
static asynCommon common = {report,connect,disconnect};


/* Forward references for asynDrvUser methods */
static asynStatus create(void* ppvt,asynUser* pasynUser,const char* drvInfo, const char** pptypeName,size_t* psize);
static asynStatus destroy(void* ppvt,asynUser* pasynUser);
static asynStatus gettype(void* ppvt,asynUser* pasynUser,const char** pptypeName,size_t* psize);
static asynDrvUser drvuser = {create,gettype,destroy};


/* Forward references for asynInt32 methods */
static asynStatus readInt32(void* ppvt,asynUser* pasynUser,epicsInt32* value);
static asynStatus writeInt32(void* ppvt,asynUser* pasynUser,epicsInt32 value);


/* Define macros */
#define ISOK(s) (asynSuccess==(s))
#define ISNOTOK(s) (!ISOK(s))


/****************************************************************************
 * Define public interface methods
 ****************************************************************************/
static int drvAdcInit(const char* port,int type,int inst,int addr,int freq,int conv,int gain)
{
    int len;
    Port* pport;
    unsigned int prio;
    asynInt32* pasynInt32;

    if( pports )
    {
        printf("drvAdcInit::init %s: already initialized\n",port);
        return( 0 );
    }

    len = sizeof(Port)+sizeof(asynInt32)+strlen(port)+1;
    pport = (Port*)callocMustSucceed(len,sizeof(char),"drvAdcInit");

    pasynInt32 = (asynInt32*)(pport + 1);
    pport->name = (char*)(pasynInt32 + 1);

    pport->type = type;
    pport->inst = inst;
    pport->isConn = FALSE;
    strcpy(pport->name,port);
    pport->syncLock = epicsMutexMustCreate();

    switch( type )
    {
    case DSC_ATHENA:
        pport->addr = ATH_DAQ_BASE;
        pport->chns = ATH_ADC_CHANS;
        break;

    default:
        printf("drvAdcInit::init %s: unknown board type %d\n",port,type);
        free(pport);
        return( - 1 );
    }

    switch(gain)
    {
    case 1:
        pport->gain = 0;
        break;
    case 2:
        pport->gain = 1;
        break;
    case 4:
        pport->gain = 2;
        break;
    case 8:
        pport->gain = 3;
        break;
    default:
        printf("drvAdc::init %s: invalid gain %d\n",port,gain);
        return( 0 );
    }

    if( freq < 1 || freq > 100 )
    {
        printf("drvAdc::init %s: invalid sample freqency %d\n",port,freq);
        return( 0 );
    }
    else
        pport->freq = freq;

    if( conv == 0 )
        pport->conv = conv;
    else if( conv == 1 && gain != 1 )
        pport->conv = conv;
    else
    {
        printf("drvAdc::init %s: invalid conversion type %d\n",port,conv);
        return( 0 );
    }

    if( pasynManager->registerPort(port,ASYN_MULTIDEVICE|ASYN_CANBLOCK,1,0,0) )
    {
        printf("drvAdc::init %s: failure to register port\n",port);
        free(pport);
        return( -1 );
    }

    pport->asynCommon.interfaceType = asynCommonType;
    pport->asynCommon.pinterface = &common;
    pport->asynCommon.drvPvt = pport;

    if( pasynManager->registerInterface(port,&pport->asynCommon) )
    {
        printf("drvAdc::init %s: failure to register asynCommon\n",port);
        return( -1 );
    }

    pport->asynDrvUser.interfaceType = asynDrvUserType;
    pport->asynDrvUser.pinterface = &drvuser;
    pport->asynDrvUser.drvPvt = pport;

    if( pasynManager->registerInterface(port,&pport->asynDrvUser) )
    {
        printf("drvAdc::init %s: failure to register asynDrvUser\n",port);
        return( -1 );
    }

    pasynInt32->read = readInt32;
    pasynInt32->write = writeInt32;
    pport->asynInt32.interfaceType = asynInt32Type;
    pport->asynInt32.pinterface = pasynInt32;
    pport->asynInt32.drvPvt = pport;

    if( pasynInt32Base->initialize(port,&pport->asynInt32) )
    {
        printf("drvAdc::init %s: failure to initialize asynInt32Base\n",port);
        return( -1 );
    }

    pports = pport;
    epicsThreadLowestPriorityLevelAbove(epicsThreadPriorityScanHigh,&prio);
    epicsThreadCreate("ScanADC",prio,epicsThreadGetStackSize(epicsThreadStackSmall),(EPICSTHREADFUNC)scanADC,pport);

    return( 0 );
}


static void scanADC(Port* pport)
{
    short val;
    float fval;
    double freq = (0.9 / pport->freq);
    unsigned char lsb,msb,fifoDepth;

    if( (getdaq(pport->addr,ATH_INB_ANALOG_INPUT_STATUS) & ATH_ANALOG_INPUT_STATUS_SD) == 0 )
        pport->chns /= 2;

    putdaq(pport->addr,ATH_OUTB_COMMAND,ATH_CMD_RSTFIFO);
    while(getdaq(pport->addr,ATH_INB_ANALOG_INPUT_STATUS) & ATH_ANALOG_INPUT_STATUS_WAIT);

    putdaq(pport->addr,ATH_OUTB_FIFO_THRESHOLD,pport->chns);
    while(getdaq(pport->addr,ATH_INB_ANALOG_INPUT_STATUS) & ATH_ANALOG_INPUT_STATUS_WAIT);

    putdaq(pport->addr,ATH_OUTB_ADC_CHANNEL,((pport->chns-1) << 4));
    while(getdaq(pport->addr,ATH_INB_ANALOG_INPUT_STATUS) & ATH_ANALOG_INPUT_STATUS_WAIT);

    putdaq(pport->addr,ATH_OUTB_ANALOG_INPUT_GAIN,ATH_ANALOG_INPUT_GAIN_SCANEN|pport->gain);
    while(getdaq(pport->addr,ATH_INB_ANALOG_INPUT_STATUS) & ATH_ANALOG_INPUT_STATUS_WAIT);

    while(0==0)
    {
        epicsThreadSleep(freq);

        putdaq(pport->addr,ATH_OUTB_COMMAND,ATH_CMD_RSTFIFO);
        while(getdaq(pport->addr,ATH_INB_ANALOG_INPUT_STATUS) & ATH_ANALOG_INPUT_STATUS_WAIT);

        putdaq(pport->addr,ATH_OUTB_COMMAND,ATH_CMD_STRTAD);
        while(getdaq(pport->addr,ATH_INB_ANALOG_INPUT_STATUS) & ATH_ANALOG_INPUT_STATUS_STS);

        fifoDepth = getdaq(pport->addr,ATH_INB_FIFO_DEPTH);
        if( fifoDepth >= ATH_ADC_FIFOMAX )
        {
            putdaq(pport->addr,ATH_OUTB_COMMAND,ATH_CMD_RSTFIFO);
            printf("ScanADC: FIFO overflow - %d\n",fifoDepth);
            epicsThreadSleep(0.250);

            continue;
        }

        for(int c = 0; c < pport->chns; c++ )
        {
            lsb = getdaq(pport->addr,ATH_INB_ADC_LSB);
            msb = getdaq(pport->addr,ATH_INB_ADC_MSB);
            val = msb * 256 + lsb;
            if( pport->conv == 0 )  
                fval = (float)val / 32768.0;
            else
                fval = ((float)val + 32768) / 65536;
            adcData[c] = (int)(fval * COM_ANA_VAL * adcGain[pport->conv][(int)pport->gain]);
        }

    }
    
}


/****************************************************************************
 * Define private interface asynCommon methods
 ****************************************************************************/
static void report(void* ppvt,FILE* fp,int details)
{
    int i;
    Port* pport = (Port*)ppvt;

    fprintf(fp,"    EPICS Brick analog input @ 0x%4.4X has %d channels\n",pport->addr,pport->chns);
    fprintf(fp,"        Conversion is %s\n", (pport->conv==0)?"bipolar":"unipolar");
    fprintf(fp,"        Sampling rate is %dHz\n",pport->freq);

    for(i = 0; i < pport->chns; ++i)
        fprintf(fp,"           chan %i connected %s\n",i,(pport->adc[i].isConn==TRUE)?"Yes":"No");

}


static asynStatus connect(void* ppvt,asynUser* pasynUser)
{
    int addr;
    asynStatus sts;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvAdc::connect %s:\n",pport->name);

    if( (sts = pasynManager->getAddr(pasynUser,&addr)) )
        return( sts );

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvAdc::connect %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    if( addr > pport->chns )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvAdc::connect %s: illegal addr %d\n",pport->name,addr);
        return( asynError );
    }

    if( addr >= 0 )
    {
        if( pport->adc[addr].isConn == TRUE )
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvAdc::connect %s: already connected to %d\n",pport->name,addr);
            return( asynError );
        }
    
        pport->adc[addr].isConn = TRUE;
        pasynManager->exceptionConnect(pasynUser);
        return( asynSuccess );
    }

    if( pport->isConn )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvAdc::connect %s: device %d already connected\n",pport->name,addr);
        return( asynError );
    }

    pport->isConn = TRUE;
    pasynManager->exceptionConnect(pasynUser);
    return( asynSuccess );
}


static asynStatus disconnect(void* ppvt,asynUser* pasynUser)
{
    int addr;
    asynStatus sts;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvAdc::disconnect %s:\n",pport->name);

    if( (sts = pasynManager->getAddr(pasynUser,&addr)) )
        return( sts );

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvAdc::disconnect %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    if( addr > pport->chns )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvAdc::disconnect %s: illegal addr %d\n",pport->name,addr);
        return( asynError );
    }

    if( addr >= 0 )
    {
        if( pport->adc[addr].isConn == FALSE )
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvAdc::disconnect %s: not connected to %d\n",pport->name,addr);
            return( asynError );
        }

        pport->adc[addr].isConn = FALSE;
        pasynManager->exceptionDisconnect(pasynUser);
        return( asynSuccess );
    }

    if( pport->isConn == FALSE )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvAdc::disconnect %s: not connected to %d\n",pport->name,addr);
        return( asynError );
    }

    pport->isConn = FALSE;
    pasynManager->exceptionDisconnect(pasynUser);
    return( asynSuccess );
}


/****************************************************************************
 * Define private interface asynDrvUser methods
 ****************************************************************************/
static asynStatus create(void* ppvt,asynUser* pasynUser,const char* drvInfo, const char** pptypeName,size_t* psize)
{
    int addr;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvAdc::create %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvAdc::create %s: getAddr failure\n",pport->name);
        return( asynError );
    }

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvAdc::create %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    if( addr > pport->chns )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvAdc::create %s: illegal channel %d\n",pport->name,addr);
        return( asynError );
    }

    if( epicsStrCaseCmp("Input",drvInfo) )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvAdc::create %s: illegal channel %d direction\n",pport->name,addr);
        return( asynError );
    }

    pport->adc[addr].regNum = addr;
    pasynUser->drvUser = &pport->adc[addr];

    return( asynSuccess );
}


static asynStatus gettype(void* ppvt,asynUser* pasynUser,const char** pptypeName,size_t* psize)
{
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvAdc::gettype %s:\n",pport->name);

    if( pptypeName )
        *pptypeName = NULL;
    if( psize )
        *psize = 0;

    return( asynSuccess );
}


static asynStatus destroy(void* ppvt,asynUser* pasynUser)
{
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvAdc::destroy %s:\n",pport->name);

    if( pport )
    {
        pasynUser->drvUser = NULL;
        return( asynSuccess );
    }
    else
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvAdc::destroy %s: called before create\n",pport->name);
        return( asynError );
    }

}


/****************************************************************************
 * Define private interface asynInt32 methods
 ****************************************************************************/
static asynStatus writeInt32(void* ppvt,asynUser* pasynUser,epicsInt32 value)
{
    Port* pport = (Port*)ppvt;
    Ainp* padc = (Ainp*)pasynUser->drvUser;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvAdc::writeInt32 %s:\n",pport->name);

    epicsMutexMustLock(pport->syncLock);
    adcData[padc->regNum] = value;
    epicsMutexUnlock(pport->syncLock);

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvAdc::writeInt32 %s: asyn - 0x%8.8X, addr - %d, value - %d\n",pport->name,(int)pasynUser,(int)padc->regNum,(int)value);

    return( asynSuccess );
}


static asynStatus readInt32(void* ppvt,asynUser* pasynUser,epicsInt32* value)
{
    Port* pport = (Port*)ppvt;
    Ainp* padc = (Ainp*)pasynUser->drvUser;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvAdc::readInt32 %s:\n",pport->name);

    epicsMutexMustLock(pport->syncLock);
    *value = adcData[padc->regNum];
    epicsMutexUnlock(pport->syncLock);

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvAdc::readInt32 %s: asyn - 0x%8.8X, addr - %d, value - %d\n",pport->name,(int)pasynUser,(int)padc->regNum,(int)*value);

    return( asynSuccess );
}


/****************************************************************************
 * Register public methods
 ****************************************************************************/

/* Initialization method definitions */
static const iocshArg drvAdcInitArg0 = {"port",iocshArgString};
static const iocshArg drvAdcInitArg1 = {"type",iocshArgInt};
static const iocshArg drvAdcInitArg2 = {"inst",iocshArgInt};
static const iocshArg drvAdcInitArg3 = {"addr",iocshArgInt};
static const iocshArg drvAdcInitArg4 = {"freq",iocshArgInt};
static const iocshArg drvAdcInitArg5 = {"conv",iocshArgInt};
static const iocshArg drvAdcInitArg6 = {"gain",iocshArgInt};
static const iocshArg* drvAdcInitArgs[]= {&drvAdcInitArg0,&drvAdcInitArg1,&drvAdcInitArg2,&drvAdcInitArg3,&drvAdcInitArg4,&drvAdcInitArg5,&drvAdcInitArg6};
static const iocshFuncDef drvAdcInitFuncDef = {"drvAdcInit",7,drvAdcInitArgs};
static void drvAdcInitCallFunc(const iocshArgBuf* args)
{
    drvAdcInit(args[0].sval,args[1].ival,args[2].ival,args[3].ival,args[4].ival,args[5].ival,args[6].ival);
}

/* Registration method */
static void drvAdcRegister(void)
{
    static int firstTime = 1;

    if( firstTime )
    {
        firstTime = 0;
        iocshRegister( &drvAdcInitFuncDef, drvAdcInitCallFunc );
    }
}
epicsExportRegistrar( drvAdcRegister );
