/*

                          Argonne National Laboratory
                            APS Operations Division
                     Beamline Controls and Data Acquisition

                          EPICS Brick DAC Port Driver



 -----------------------------------------------------------------------------
                                COPYRIGHT NOTICE
 -----------------------------------------------------------------------------
   Copyright (c) 2002 The University of Chicago, as Operator of Argonne
      National Laboratory.
   Copyright (c) 2002 The Regents of the University of California, as
      Operator of Los Alamos National Laboratory.
   Synapps Versions 4-5
   and higher are distributed subject to a Software License Agreement found
   in file LICENSE that is included with this distribution.
 -----------------------------------------------------------------------------

 Description
    This module provides support for a multiple device port driver. To
    initialize the driver, the method drvDacInit() is called from the
    startup script with the following calling sequence.

        drvDacInit(port,type,inst,addr,chns)

        Where:
            port - DSC port driver name (i.e. "DAC" )
            type - DSC board type (see dscsup.h)
            inst - Board instance
            addr - Board address, zero for Athena
            chns - Register count


    The method dbior can be called from the IOC shell to display the current
    status of the driver.


 Developer notes:
    The supported conversion is bipolar 10V for all board types.

 Source control info:
    Modified by:    $Author: dkline $
                    $Date: 2006/05/26 20:36:41 $
                    $Revision: 1.2 $

 =============================================================================
 History:
 Author: David M. Kline
 -----------------------------------------------------------------------------
 2005-Nov-09  DMK  Derived from love multidevice driver.
 2005-Nov-15  DMK  Initial development complete.
 2006-Feb-28  DMK  Augmented to support board types and instances.
 2006-May-19  DMK  Added 'addr' parameter to initialization method.
 -----------------------------------------------------------------------------

*/


/* EPICS base version-specific definitions (must be performed first) */
#include <epicsVersion.h>
#define LT_EPICSBASE(v,r,l) (EPICS_VERSION<(v)||(EPICS_VERSION==(v)&&(EPICS_REVISION<(r)||(EPICS_REVISION==(r)&&EPICS_MODIFICATION<(l)))))


/* Evaluate EPICS base */
#if LT_EPICSBASE(3,14,6)
    #error "EPICS base must be 3.14.6 or greater"
#endif

/* System related include files */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* EPICS system related include files */
#include <iocsh.h>
#include <epicsStdio.h>
#include <cantProceed.h>
#include <epicsString.h>
#include <epicsExport.h>
#include <epicsThread.h>


/* EPICS synApps/Asyn related include files */
#include <asynDriver.h>
#include <asynInt32.h>
#include <asynDrvUser.h>


/* DSC related include files */
#include "dscsup.h"


/* Define symbolic constants */
#define TRUE       (1)
#define FALSE      (0)

/* Forward struct declarations */
typedef struct Dac Dac;
typedef struct Port Port;


/* Declare register (A,B,C) info structure */
struct Dac
{
    int isConn;
    int regNum;
};


/* Declare DSC port structure */
struct Port
{
    char*         name;
    int           type;
    int           inst;
    int           addr;
    int           isConn;
    epicsMutexId  syncLock;
    asynInterface asynInt32;
    asynInterface asynCommon;
    asynInterface asynDrvUser;

    /* Channel info */
    int           chns;
    Dac*          dacs;
};


/* Public interface forward references */
static int drvDacInit(const char* port,int type,int inst,int addr,int chns);


/* Forward references for asynCommon methods */
static void report(void* ppvt,FILE* fp,int details);
static asynStatus connect(void* ppvt,asynUser* pasynUser);
static asynStatus disconnect(void* ppvt,asynUser* pasynUser);
static asynCommon common = {report,connect,disconnect};


/* Forward references for asynDrvUser methods */
static asynStatus create(void* ppvt,asynUser* pasynUser,const char* drvInfo, const char** pptypeName,size_t* psize);
static asynStatus destroy(void* ppvt,asynUser* pasynUser);
static asynStatus gettype(void* ppvt,asynUser* pasynUser,const char** pptypeName,size_t* psize);
static asynDrvUser drvuser = {create,gettype,destroy};


/* Forward references for asynInt32 methods */
static asynStatus readInt32(void* ppvt,asynUser* pasynUser,epicsInt32* value);
static asynStatus writeInt32(void* ppvt,asynUser* pasynUser,epicsInt32 value);


/* Define macros */
#define ISOK(s) (asynSuccess==(s))
#define ISNOTOK(s) (!ISOK(s))


/****************************************************************************
 * Define public interface methods
 ****************************************************************************/
static int drvDacInit(const char* port,int type,int inst,int addr,int chns)
{
    int len,ns;
    Port* pport;
    asynInt32* pasynInt32;

    ns = strlen(port)+1;
    len = sizeof(Port)+sizeof(asynInt32)+ns+(sizeof(Dac)*chns);
    pport = (Port*)callocMustSucceed(len,sizeof(char),"drvDac");

    pasynInt32 = (asynInt32*)(pport + 1);
    pport->name = (char*)(pasynInt32 + 1);
    pport->dacs = (Dac*)(pport->name + ns);

    pport->type = type;
    pport->inst = inst;
    pport->chns = chns;
    pport->isConn = FALSE;
    strcpy(pport->name,port);
    pport->syncLock = epicsMutexMustCreate();

    switch( type )
    {
    case DSC_ATHENA:
        pport->addr = ATH_DAQ_BASE;
        break;

    case DSC_RMM416:
        pport->addr = addr;
        break;

    default:
        printf("drvDac::init %s: unknown board type %d\n",port,type);
        free(pport);
        return( - 1 );
    }

    if( pasynManager->registerPort(port,ASYN_MULTIDEVICE|ASYN_CANBLOCK,1,0,0) )
    {
        printf("drvDac::init %s: failure to register DSC port\n",port);
        free(pport);
        return( -1 );
    }

    pport->asynCommon.interfaceType = asynCommonType;
    pport->asynCommon.pinterface = &common;
    pport->asynCommon.drvPvt = pport;

    if( pasynManager->registerInterface(port,&pport->asynCommon) )
    {
        printf("drvDac::init %s: failure to register asynCommon\n",port);
        return( -1 );
    }

    pport->asynDrvUser.interfaceType = asynDrvUserType;
    pport->asynDrvUser.pinterface = &drvuser;
    pport->asynDrvUser.drvPvt = pport;

    if( pasynManager->registerInterface(port,&pport->asynDrvUser) )
    {
        printf("drvDac::init %s: failure to register asynDrvUser\n",port);
        return( -1 );
    }

    pasynInt32->read = readInt32;
    pasynInt32->write = writeInt32;
    pport->asynInt32.interfaceType = asynInt32Type;
    pport->asynInt32.pinterface = pasynInt32;
    pport->asynInt32.drvPvt = pport;

    if( pasynInt32Base->initialize(port,&pport->asynInt32) )
    {
        printf("drvDac::init %s: failure to initialize asynInt32Base\n",port);
        return( -1 );
    }

    return( 0 );
}


/****************************************************************************
 * Define private interface asynCommon methods
 ****************************************************************************/
static void report(void* ppvt,FILE* fp,int details)
{
    int i;
    Port* pport = (Port*)ppvt;

    fprintf(fp,"    EPICS Brick analog output @ 0x%4.4X has %d bipolar (10V) channels\n",pport->addr,pport->chns);
    for( i = 0; i < pport->chns; ++i )
        fprintf(fp,"        chan %i connected %s\n",i,(pport->dacs[i].isConn==TRUE)?"Yes":"No");

}


static asynStatus connect(void* ppvt,asynUser* pasynUser)
{
    int addr;
    asynStatus sts;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvDac::connect %s:\n",pport->name);

    if( (sts = pasynManager->getAddr(pasynUser,&addr)) )
        return( sts );

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvDac::connect %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    if( addr > pport->chns )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvDac::connect %s: illegal addr %d\n",pport->name,addr);
        return( asynError );
    }

    if( addr >= 0 )
    {
        if( pport->dacs[addr].isConn == TRUE )
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvDac::connect %s: already connected to %d\n",pport->name,addr);
            return( asynError );
        }
    
        pport->dacs[addr].isConn = TRUE;
        pasynManager->exceptionConnect(pasynUser);
        return( asynSuccess );
    }

    if( pport->isConn )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvDac::connect %s: device %d already connected\n",pport->name,addr);
        return( asynError );
    }

    pport->isConn = TRUE;
    pasynManager->exceptionConnect(pasynUser);
    return( asynSuccess );
}


static asynStatus disconnect(void* ppvt,asynUser* pasynUser)
{
    int addr;
    asynStatus sts;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvDac::disconnect %s:\n",pport->name);

    if( (sts = pasynManager->getAddr(pasynUser,&addr)) )
        return( sts );

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvDac::disconnect %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    if( addr > pport->chns )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvDac::disconnect %s: illegal addr %d\n",pport->name,addr);
        return( asynError );
    }

    if( addr >= 0 )
    {
        if( pport->dacs[addr].isConn == FALSE )
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvDac::disconnect %s: not connected to %d\n",pport->name,addr);
            return( asynError );
        }

        pport->dacs[addr].isConn = FALSE;
        pasynManager->exceptionDisconnect(pasynUser);
        return( asynSuccess );
    }

    if( pport->isConn == FALSE )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvDac::disconnect %s: not connected to addr %d\n",pport->name,addr);
        return( asynError );
    }

    pport->isConn = FALSE;
    pasynManager->exceptionDisconnect(pasynUser);
    return( asynSuccess );
}


/****************************************************************************
 * Define private interface asynDrvUser methods
 ****************************************************************************/
static asynStatus create(void* ppvt,asynUser* pasynUser,const char* drvInfo, const char** pptypeName,size_t* psize)
{
    int addr;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvDac::create %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvDac::create %s: getAddr failure\n",pport->name);
        return( asynError );
    }

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvDac::create %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    if( addr > pport->chns )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvDac::create %s: illegal channel %d\n",pport->name,addr);
        return( asynError );
    }

    if( epicsStrCaseCmp("Output",drvInfo) )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvDac::create %s: illegal channel %d direction\n",pport->name,addr);
        return( asynError );
    }

    pport->dacs[addr].regNum = addr;
    pasynUser->drvUser = &pport->dacs[addr];

    return( asynSuccess );
}


static asynStatus gettype(void* ppvt,asynUser* pasynUser,const char** pptypeName,size_t* psize)
{
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvDac::gettype %s:\n",pport->name);

    if( pptypeName )
        *pptypeName = NULL;
    if( psize )
        *psize = 0;

    return( asynSuccess );
}


static asynStatus destroy(void* ppvt,asynUser* pasynUser)
{
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvDac::destroy %s:\n",pport->name);

    if( pport )
    {
        pasynUser->drvUser = NULL;
        return( asynSuccess );
    }
    else
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvDac::destroy %s: called before create\n",pport->name);
        return( asynError );
    }

}


/****************************************************************************
 * Define private interface asynInt32 methods
 ****************************************************************************/
static asynStatus writeInt32(void* ppvt,asynUser* pasynUser,epicsInt32 value)
{
    Port* pport = (Port*)ppvt;
    Dac* pdac = (Dac*)pasynUser->drvUser;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvDac::writeInt32 %s:\n",pport->name);

    epicsMutexMustLock(pport->syncLock);
    dacWrite(pport->addr,pport->type,pdac->regNum,value);
    epicsMutexUnlock(pport->syncLock);

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvDac::writeInt32 %s: asyn - 0x%8.8X, addr - %d, value - %d\n",pport->name,(int)pasynUser,(int)pdac->regNum,(int)value);

    return( asynSuccess );
}


static asynStatus readInt32(void* ppvt,asynUser* pasynUser,epicsInt32* value)
{
    Port* pport = (Port*)ppvt;
    Dac* pdac = (Dac*)pasynUser->drvUser;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvDac::readInt32 %s:\n",pport->name);
    *value = 0;
    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvDac::readInt32 %s: asyn - 0x%8.8X, addr - %d, value - %d\n",pport->name,(int)pasynUser,(int)pdac->regNum,(int)*value);

    return( asynSuccess );
}


/****************************************************************************
 * Register public methods
 ****************************************************************************/

/* Initialization method definitions */
static const iocshArg drvDacInitArg0 = {"port",iocshArgString};
static const iocshArg drvDacInitArg1 = {"type",iocshArgInt};
static const iocshArg drvDacInitArg2 = {"inst",iocshArgInt};
static const iocshArg drvDacInitArg3 = {"addr",iocshArgInt};
static const iocshArg drvDacInitArg4 = {"chns",iocshArgInt};
static const iocshArg* drvDacInitArgs[]= {&drvDacInitArg0,&drvDacInitArg1,&drvDacInitArg2,&drvDacInitArg3,&drvDacInitArg4};
static const iocshFuncDef drvDacInitFuncDef = {"drvDacInit",5,drvDacInitArgs};
static void drvDacInitCallFunc(const iocshArgBuf* args)
{
    drvDacInit(args[0].sval,args[1].ival,args[2].ival,args[3].ival,args[4].ival);
}

/* Registration method */
static void drvDacRegister(void)
{
    static int firstTime = 1;

    if( firstTime )
    {
        firstTime = 0;
        iocshRegister( &drvDacInitFuncDef, drvDacInitCallFunc );
    }
}
epicsExportRegistrar( drvDacRegister );
