/*

                          Argonne National Laboratory
                            APS Operations Division
                     Beamline Controls and Data Acquisition

                          EPICS Brick DIO Port Driver



 -----------------------------------------------------------------------------
                                COPYRIGHT NOTICE
 -----------------------------------------------------------------------------
   Copyright (c) 2002 The University of Chicago, as Operator of Argonne
      National Laboratory.
   Copyright (c) 2002 The Regents of the University of California, as
      Operator of Los Alamos National Laboratory.
   Synapps Versions 4-5
   and higher are distributed subject to a Software License Agreement found
   in file LICENSE that is included with this distribution.
 -----------------------------------------------------------------------------

 Description
    This module provides support for a multiple device port driver. To
    initialize the driver, the method drvDioInit() is called from the
    startup script with the following calling sequence.

        drvDioInit(port,type,inst,addr,regs)

        Where:
            port - Port driver name (i.e. "DIO" )
            type - Board type (see dscsup.h)
            inst - Board instance
            addr - Board address, zero for Athena
            regs - Register count

    The method dbior can be called from the IOC shell to display the current
    status of the driver.


 Developer notes:

 Source control info:
    Modified by:    $Author: dkline $
                    $Date: 2006/07/31 14:57:37 $
                    $Revision: 1.5 $

 =============================================================================
 History:
 Author: David M. Kline
 -----------------------------------------------------------------------------
 2005-Nov-09  DMK  Derived from love multidevice driver.
 2005-Nov-15  DMK  Initial development complete.
 2006-Feb-28  DMK  Added support for Ruby-RMM-416.
                   Made methods more generic.
 2006-May-19  DMK  Added 'addr' parameter to initialization method.
                   Added support for Pearl-MM modules.
 2006-Jun-03  DMK  Added support for Onyx-MM modules.
 2006-Jul-31  DMK  Added support for IR-104 modules.
 -----------------------------------------------------------------------------

*/


/* EPICS base version-specific definitions (must be performed first) */
#include <epicsVersion.h>
#define LT_EPICSBASE(v,r,l) (EPICS_VERSION<(v)||(EPICS_VERSION==(v)&&(EPICS_REVISION<(r)||(EPICS_REVISION==(r)&&EPICS_MODIFICATION<(l)))))


/* Evaluate EPICS base */
#if LT_EPICSBASE(3,14,6)
    #error "EPICS base must be 3.14.6 or greater"
#endif

/* System related include files */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* EPICS system related include files */
#include <iocsh.h>
#include <epicsStdio.h>
#include <cantProceed.h>
#include <epicsString.h>
#include <epicsExport.h>
#include <epicsThread.h>


/* EPICS synApps/Asyn related include files */
#include <asynDriver.h>
#include <asynDrvUser.h>
#include <asynUInt32Digital.h>


/* DSC related include files */
#include "dscsup.h"


/* Define symbolic constants */
#define IN         (1)
#define OUT        (0)
#define TRUE       (1)
#define FALSE      (0)
#define K_COMTMO   (1)


/* Forward struct declarations */
typedef struct Reg Reg;
typedef struct Port Port;


/* Declare register info structure */
struct Reg
{
    int isConn;
    int cfgIdx;
    int cfgOff;
    int regOff;
    int regNum;
    int regDir;
};


/* Declare DSC port structure */
struct Port
{
    char*         name;
    int           type;
    int           inst;
    int           addr;
    int           isConn;
    epicsMutexId  syncLock;
    asynInterface asynUInt32;
    asynInterface asynCommon;
    asynInterface asynDrvUser;

    /* Register info */
    int            cnt;
    unsigned char* cfgs;
    Reg*           regs;
};


/* Public interface forward references */
static int drvDioInit(const char* port,int type,int inst,int addr,int regs);


/* Forward references for asynCommon methods */
static void report(void* ppvt,FILE* fp,int details);
static asynStatus connect(void* ppvt,asynUser* pasynUser);
static asynStatus disconnect(void* ppvt,asynUser* pasynUser);
static asynCommon common = {report,connect,disconnect};


/* Forward references for asynDrvUser methods */
static asynStatus create(void* ppvt,asynUser* pasynUser,const char* drvInfo, const char** pptypeName,size_t* psize);
static asynStatus destroy(void* ppvt,asynUser* pasynUser);
static asynStatus gettype(void* ppvt,asynUser* pasynUser,const char** pptypeName,size_t* psize);
static asynDrvUser drvuser = {create,gettype,destroy};


/* Forward references for asynUInt32Digital methods */
static asynStatus readUInt32(void* ppvt,asynUser* pasynUser,epicsUInt32* value,epicsUInt32 mask);
static asynStatus writeUInt32(void* ppvt,asynUser* pasynUser,epicsUInt32 value,epicsUInt32 mask);


/* Define macros */
#define ISOK(s) (asynSuccess==(s))
#define ISNOTOK(s) (!ISOK(s))


/****************************************************************************
 * Define public interface methods
 ****************************************************************************/
static int drvDioInit(const char* port,int type,int inst,int addr,int regs)
{
    Port* pport;
    int rmax,len,ns,cs;
    asynUInt32Digital* pasynUInt32;

    ns = strlen(port)+1;
    cs = regs/COM_DIO_REGS;
    len = sizeof(Port)+sizeof(asynUInt32Digital)+ns+cs+(sizeof(Reg)*regs);
    pport = (Port*)callocMustSucceed(len,sizeof(char),"drvDioInit");

    pasynUInt32 = (asynUInt32Digital*)(pport + 1);
    pport->name = (char*)(pasynUInt32 + 1);
    pport->cfgs = (unsigned char*)(pport->name + ns);
    pport->regs = (Reg*)(pport->cfgs + cs);

    pport->cnt = regs;
    pport->inst = inst;
    pport->type = type;
    pport->isConn = FALSE;
    strcpy(pport->name,port);
    pport->syncLock = epicsMutexMustCreate();

    switch( type )
    {
    case DSC_ATHENA:
        rmax = 3;
        pport->addr = ATH_DAQ_BASE;
        break;

    case DSC_RMM416:
        rmax = 3;
        pport->addr = addr;
        break;

    case DSC_PMM:
        rmax = 2;
        pport->addr = addr;
        break;

    case DSC_OMM:
        rmax = 6;
        pport->addr = addr;
        break;

    case DSC_GMM_PIO:
        rmax = 6;
        pport->addr = addr;
        break;

    case DSC_GMM_FIX:
        rmax = 2;
        pport->addr = addr;
        break;

    case DSC_IR4:
        rmax = 6;
        pport->addr = addr;
        break;

    default:
        printf("drvDioInit::init %s: unknown board type %d\n",port,type);
        free(pport);
        return( - 1 );
    }

    if( regs > rmax )
    {
        printf("drvDioInit::init %s: board type %d cannot have > %d registers\n",port,type,rmax);
        free(pport);
        return( - 1 );
    }

    if( pasynManager->registerPort(port,ASYN_MULTIDEVICE|ASYN_CANBLOCK,1,0,0) )
    {
        printf("drvDioInit::init %s: failure to register port\n",port);
        free(pport);
        return( -1 );
    }

    pport->asynCommon.interfaceType = asynCommonType;
    pport->asynCommon.pinterface = &common;
    pport->asynCommon.drvPvt = pport;

    if( pasynManager->registerInterface(port,&pport->asynCommon) )
    {
        printf("drvDioInit::init %s: failure to register asynCommon\n",port);
        return( -1 );
    }

    pport->asynDrvUser.interfaceType = asynDrvUserType;
    pport->asynDrvUser.pinterface = &drvuser;
    pport->asynDrvUser.drvPvt = pport;

    if( pasynManager->registerInterface(port,&pport->asynDrvUser) )
    {
        printf("drvDioInit::init %s: failure to register asynDrvUser\n",port);
        return( -1 );
    }

    pasynUInt32->read = readUInt32;
    pasynUInt32->write = writeUInt32;
    pport->asynUInt32.interfaceType = asynUInt32DigitalType;
    pport->asynUInt32.pinterface = pasynUInt32;
    pport->asynUInt32.drvPvt = pport;

    if( pasynUInt32DigitalBase->initialize(port,&pport->asynUInt32) )
    {
        printf("drvDioInit::init %s: failure to initialize asynUInt32DigitalBase\n",port);
        return( -1 );
    }

    return( 0 );
}


/****************************************************************************
 * Define private interface asynCommon methods
 ****************************************************************************/
static void report(void* ppvt,FILE* fp,int details)
{
    int i,con;
    Reg* preg;
    Port* pport = (Port*)ppvt;

    fprintf(fp,"    EPICS Brick digital I/O @ 0x%4.4X is type %d and has %d registers\n",pport->addr,pport->type,pport->cnt);
    for( i = 0; i < pport->cnt; ++i )
    {
        preg = &pport->regs[i];
        con = preg->isConn==TRUE;
        fprintf(fp,"        reg %d connected %s dir %s\n",i,(con)?"Yes":"No",(preg->regDir)?"Inp":"Out");
    }
}


static asynStatus connect(void* ppvt,asynUser* pasynUser)
{
    int addr;
    asynStatus sts;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvDio::connect %s:\n",pport->name);

    if( (sts = pasynManager->getAddr(pasynUser,&addr)) )
        return( sts );

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvDio::connect %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    if( addr > pport->cnt )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvDio::connect %s: illegal addr %d\n",pport->name,addr);
        return( asynError );
    }

    if( addr >= 0 )
    {
        if( pport->regs[addr].isConn == TRUE )
	{
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvDio::connect %s: already connected to %d\n",pport->name,addr);
            return( asynError );
	}
    
        pport->regs[addr].isConn = TRUE;
        pport->regs[addr].regNum = addr;
        pasynManager->exceptionConnect(pasynUser);
        return( asynSuccess );
    }

    if( pport->isConn )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvDio::connect %s: already connected to %d\n",pport->name,addr);
        return( asynError );
    }

    pport->isConn = TRUE;
    pasynManager->exceptionConnect(pasynUser);
    return( asynSuccess );
}


static asynStatus disconnect(void* ppvt,asynUser* pasynUser)
{
    int addr;
    asynStatus sts;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvDio::disconnect %s:\n",pport->name);

    if( (sts = pasynManager->getAddr(pasynUser,&addr)) )
        return( sts );

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvDio::disconnect %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    if( addr > pport->cnt )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvDio::disconnect %s: illegal addr %d\n",pport->name,addr);
        return( asynError );
    }

    if( addr >= 0 )
    {
        if( pport->regs[addr].isConn == FALSE )
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvDio::disconnect %s: not connected to %d\n",pport->name,addr);
            return( asynError );
        }

        pport->regs[addr].isConn = FALSE;
        pasynManager->exceptionDisconnect(pasynUser);
        return( asynSuccess );
    }

    if( pport->isConn == FALSE )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvDio::disconnect %s: not connected to %d\n",pport->name,addr);
        return( asynError );
    }

    pport->isConn = FALSE;
    pasynManager->exceptionDisconnect(pasynUser);
    return( asynSuccess );
}


/****************************************************************************
 * Define private interface asynDrvUser methods
 ****************************************************************************/
static asynStatus create(void* ppvt,asynUser* pasynUser,const char* drvInfo, const char** pptypeName,size_t* psize)
{
    int addr;
    Reg* preg;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvDio::create %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        preg = &pport->regs[addr];
    else
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvDio::create %s: getAddr failure\n",pport->name);
        return( asynError );
    }

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvDio::create %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    preg->cfgIdx = addr/COM_DIO_REGS;
    switch( pport->type )
    {
    case DSC_ATHENA:
        preg->cfgOff = ATH_OUTB_DIOCR;
        if( addr == 0 )
            preg->regOff = ATH_OUTB_DPORT_A;
        else if( addr == 1 )
            preg->regOff = ATH_OUTB_DPORT_B;
        else if( addr == 2 )
            preg->regOff = ATH_OUTB_DPORT_C;
        else
        {
            asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvDio::create %s: invalid register number %d\n",pport->name,addr);
            return( asynError );
        }
        break;

    case DSC_RMM416:
        preg->cfgOff = RMM_OUTB_DIOCR;
        if( addr == 0 )
            preg->regOff = RMM_OUTB_DPORT_A;
        else if( addr == 1 )
            preg->regOff = RMM_OUTB_DPORT_B;
        else if( addr == 2 )
            preg->regOff = RMM_OUTB_DPORT_C;
        else
        {
            asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvDio::create %s: invalid register number %d\n",pport->name,addr);
            return( asynError );
        }
        break;

    case DSC_PMM:
        if( epicsStrCaseCmp("Input",drvInfo) == 0 )
        {
            asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvDio::create %s: invalid register %d direction\n",pport->name,addr);
            return( asynError );
        }

        preg->cfgOff = PMM_OUTB_DIOCR;
        if( addr == 0 )
            preg->regOff = PMM_OUTB_DPORT_A;
        else if( addr == 1 )
            preg->regOff = PMM_OUTB_DPORT_B;
        else
        {
            asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvDio::create %s: invalid register number %d\n",pport->name,addr);
            return( asynError );
        }
        break;

    case DSC_OMM:
        if( addr < 3 )
           preg->cfgOff = OMM_OUTB_DIOCR1;
        else if( addr > 2 )
           preg->cfgOff = OMM_OUTB_DIOCR2;

        if( addr == 0 )
            preg->regOff = OMM_OUTB_DPORT1_A;
        else if( addr == 1 )
            preg->regOff = OMM_OUTB_DPORT1_B;
        else if( addr == 2 )
            preg->regOff = OMM_OUTB_DPORT1_C;
        else if( addr == 3 )
            preg->regOff = OMM_OUTB_DPORT2_A;
        else if( addr == 4 )
            preg->regOff = OMM_OUTB_DPORT2_B;
        else if( addr == 5 )
            preg->regOff = OMM_OUTB_DPORT2_C;
        else
        {
            asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvDio::create %s: invalid register number %d\n",pport->name,addr);
            return( asynError );
        }
        break;

    case DSC_GMM_PIO:
        if( addr < 3 )
           preg->cfgOff = GMM_PIO_DIOCR1;
        else if( addr > 2 )
           preg->cfgOff = GMM_PIO_DIOCR2;

        if( addr == 0 )
            preg->regOff = GMM_PIO_DPORT1_A;
        else if( addr == 1 )
            preg->regOff = GMM_PIO_DPORT1_B;
        else if( addr == 2 )
            preg->regOff = GMM_PIO_DPORT1_C;
        else if( addr == 3 )
            preg->regOff = GMM_PIO_DPORT2_A;
        else if( addr == 4 )
            preg->regOff = GMM_PIO_DPORT2_B;
        else if( addr == 5 )
            preg->regOff = GMM_PIO_DPORT2_C;
        else
        {
            asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvDio::create %s: invalid register number %d\n",pport->name,addr);
            return( asynError );
        }
        break;

    case DSC_GMM_FIX:
        preg->cfgOff = GMM_FIX_DIOCR;
        if( addr == 0 )
            preg->regOff = GMM_FIX_DPORT1_A;
        else if( addr == 1 )
            preg->regOff = GMM_FIX_DPORT1_B;
        else
        {
            asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvDio::create %s: invalid register number %d\n",pport->name,addr);
            return( asynError );
        }
        break;

    case DSC_IR4:
        if( addr < 3 )
           preg->cfgOff = IR4_OUTB_DIOCR;
        else if( addr > 2 )
           preg->cfgOff = IR4_INPB_DIOCR;

        if( addr == 0 )
            preg->regOff = IR4_OUTB_DPORT_A;
        else if( addr == 1 )
            preg->regOff = IR4_OUTB_DPORT_B;
        else if( addr == 2 )
            preg->regOff = IR4_OUTB_DPORT_C;
        else if( addr == 3 )
            preg->regOff = IR4_INPB_DPORT_A;
        else if( addr == 4 )
            preg->regOff = IR4_INPB_DPORT_B;
        else if( addr == 5 )
            preg->regOff = IR4_INPB_DPORT_C;
        else
        {
            asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvDio::create %s: invalid register number %d\n",pport->name,addr);
            return( asynError );
        }
        break;

    }

    if( epicsStrCaseCmp("Input",drvInfo) == 0 )
    {
        preg->regDir = IN;
        dioSetDir(pport->addr,preg->regOff,preg->cfgOff,&pport->cfgs[preg->cfgIdx],1);
    }
    else if( epicsStrCaseCmp("Output",drvInfo) == 0 )
    {
        preg->regDir = OUT;
        dioSetDir(pport->addr,preg->regOff,preg->cfgOff,&pport->cfgs[preg->cfgIdx],0);
    }
    else
    {
        asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvDio::create %s: failure to determine register %d direction\n",pport->name,addr);
        return( asynError );
    }

    preg->regNum = addr;
    pasynUser->drvUser = preg;

    return( asynSuccess );
}


static asynStatus gettype(void* ppvt,asynUser* pasynUser,const char** pptypeName,size_t* psize)
{
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvDio::gettype %s:\n",pport->name);

    if( pptypeName )
        *pptypeName = NULL;
    if( psize )
        *psize = 0;

    return( asynSuccess );
}


static asynStatus destroy(void* ppvt,asynUser* pasynUser)
{
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvDio::destroy %s:\n",pport->name);

    if( pport )
    {
        pasynUser->drvUser = NULL;
        return( asynSuccess );
    }
    else
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvDio::destroy %s: called before create\n",pport->name);
        return( asynError );
    }

}


/****************************************************************************
 * Define private interface asynUInt32Digital methods
 ****************************************************************************/
static asynStatus writeUInt32(void* ppvt,asynUser* pasynUser,epicsUInt32 value,epicsUInt32 mask)
{
    int addr;
    Reg* preg;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvDio::writeUInt32 %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        preg = &pport->regs[addr];
    else
        return( asynError );

    value &= mask;
    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvDio::writeUInt32 %s: asyn - 0x%8.8X, addr - %d, mask - 0x%8.8X, value - 0x%8.8X\n",pport->name,(int)pasynUser,(int)preg->regNum,(int)mask,(int)value);

    epicsMutexMustLock(pport->syncLock);
    dioWrite(pport->addr,preg->regOff,value);
    epicsMutexUnlock(pport->syncLock);

    return( asynSuccess );
}


static asynStatus readUInt32(void* ppvt,asynUser* pasynUser,epicsUInt32* value,epicsUInt32 mask)
{
    int addr;
    Reg* preg;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvDio::readUInt32 %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        preg = &pport->regs[addr];
    else
        return( asynError );

    epicsMutexMustLock(pport->syncLock);
    dioRead(pport->addr,preg->regOff,(unsigned char*)value);
    epicsMutexUnlock(pport->syncLock);

    *value &= mask;
    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvDio::readUInt32 %s: asyn - 0x%8.8X, addr - %d, mask - 0x%8.8X, value - 0x%8.8X\n",pport->name,(int)pasynUser,(int)preg->regNum,(int)mask,(int)*value);

    return( asynSuccess );
}


/****************************************************************************
 * Register public methods
 ****************************************************************************/

/* Initialization method definitions */
static const iocshArg drvDioInitArg0 = {"port",iocshArgString};
static const iocshArg drvDioInitArg1 = {"type",iocshArgInt};
static const iocshArg drvDioInitArg2 = {"inst",iocshArgInt};
static const iocshArg drvDioInitArg3 = {"addr",iocshArgInt};
static const iocshArg drvDioInitArg4 = {"regs",iocshArgInt};
static const iocshArg* drvDioInitArgs[]= {&drvDioInitArg0,&drvDioInitArg1,&drvDioInitArg2,&drvDioInitArg3,&drvDioInitArg4};
static const iocshFuncDef drvDioInitFuncDef = {"drvDioInit",5,drvDioInitArgs};
static void drvDioInitCallFunc(const iocshArgBuf* args)
{
    drvDioInit(args[0].sval,args[1].ival,args[2].ival,args[3].ival,args[4].ival);
}

/* Registration method */
static void drvDioRegister(void)
{
    static int firstTime = 1;

    if( firstTime )
    {
        firstTime = 0;
        iocshRegister( &drvDioInitFuncDef, drvDioInitCallFunc );
    }
}
epicsExportRegistrar( drvDioRegister );
