/*

                          Argonne National Laboratory
                            APS Operations Division
                     Beamline Controls and Data Acquisition

                          EPICS Brick GPIB Port Driver



 -----------------------------------------------------------------------------
                                COPYRIGHT NOTICE
 -----------------------------------------------------------------------------
   Copyright (c) 2002 The University of Chicago, as Operator of Argonne
      National Laboratory.
   Copyright (c) 2002 The Regents of the University of California, as
      Operator of Los Alamos National Laboratory.
   Synapps Versions 4-5
   and higher are distributed subject to a Software License Agreement found
   in file LICENSE that is included with this distribution.
 -----------------------------------------------------------------------------

 Description
    This module provides support for a multiple device port driver. To
    initialize the driver, the method drvGpibInit() is called from the
    startup script with the following calling sequence.

        drvGpibInit(port,gport)
        Where:
            port     - Port driver name (i.e. "I0" )
            gport - GPIB port driver name (i.e. "G0" )

        drvGpibConf(port,addr,type,chns)
        Where:
            port - Port driver name (i.e. "I0" )
            addr - Instrument address
            type - Instrument type (boardtypes.h)
            chns - Channel count

    The method dbior can be called from the IOC shell to display the current
    status of the driver.


 Developer notes:

 Source control info:
    Modified by:    $Author: dkline $
                    $Date: 2006/08/22 12:57:03 $
                    $Revision: 1.1 $

 =============================================================================
 History:
 Author: David M. Kline
 -----------------------------------------------------------------------------
 2006-Aug-21  DMK  Development begins. Only thermocouple support initially.
 -----------------------------------------------------------------------------

*/


/* EPICS base version-specific definitions (must be performed first) */
#include <epicsVersion.h>
#define LT_EPICSBASE(v,r,l) (EPICS_VERSION<(v)||(EPICS_VERSION==(v)&&(EPICS_REVISION<(r)||(EPICS_REVISION==(r)&&EPICS_MODIFICATION<(l)))))


/* Evaluate EPICS base */
#if LT_EPICSBASE(3,14,6)
    #error "EPICS base must be 3.14.6 or greater"
#endif

/* System related include files */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* EPICS system related include files */
#include <iocsh.h>
#include <epicsStdio.h>
#include <cantProceed.h>
#include <epicsString.h>
#include <epicsExport.h>
#include <epicsThread.h>


/* EPICS synApps/Asyn related include files */
#include <asynDriver.h>
#include <asynInt32.h>
#include <asynOctet.h>
#include <asynDrvUser.h>
#include <asynFloat64.h>
#include <asynUInt32Digital.h>


/* Boardtypes related include files */
#include "boardtypes.h"


/* Define symbolic constants */
#define OUT      (0)
#define INP      (1)
#define TRUE     (1)
#define FALSE    (0)
#define COMTMO   (1)
#define INSTRMAX (15)


/* Forward struct declarations */
typedef struct Chn Chn;
typedef struct Port Port;
typedef struct Instr Instr;
typedef struct Gpibport Gpibport;


/* Define controller type enum */
typedef enum {DIO=200,TCI,SSR,SMD} Type;


/* Declare register info structure */
struct Chn
{
    int con;
    int chn;
};


/* Declare instrument info structure */
struct Instr
{
    int  con;
    Type type;
    int  addr;
    int  chns;
    Chn* chn;
};


/* Declare port structure */
struct Port
{
    Port*         pport;

    char*         name;
    int           con;
    Gpibport*     pgpib;
    asynUser*     pasynUser;
    asynInterface asynInt32;
    asynInterface asynUInt32;
    asynInterface asynCommon;
    asynInterface asynDrvUser;
    asynInterface asynFloat64;
    asynInterface asynLockPort;

    Instr instr[INSTRMAX];
};


/* Declare GPIB port structure */
struct Gpibport
{
    char*      name;
    int        addr;
    int        con;
    int        canBlock;
    int        multiDevice;
    int        autoConnect;
    asynUser*  pasynUser;
    asynOctet* pasynOctet;
    void*      pasynOctetPvt;
};


/* Defaine local variants */
static Port* pports;

/* Public interface forward references */
static int drvGpibInit(const char* port,const char* gport);
static int drvGpibConf(const char* port,int addr,int type,int chns);
static asynStatus initGpibPort(Port* port,const char* gport,int gpibaddr);


/* Forward references for asynCommon methods */
static void report(void* ppvt,FILE* fp,int details);
static asynStatus connect(void* ppvt,asynUser* pasynUser);
static asynStatus disconnect(void* ppvt,asynUser* pasynUser);
static asynCommon common = {report,connect,disconnect};


/* Forward references for asynDrvUser methods */
static asynStatus create(void* ppvt,asynUser* pasynUser,const char* drvInfo, const char** pptypeName,size_t* psize);
static asynStatus destroy(void* ppvt,asynUser* pasynUser);
static asynStatus gettype(void* ppvt,asynUser* pasynUser,const char** pptypeName,size_t* psize);
static asynDrvUser drvuser = {create,gettype,destroy};


/* asynLockPortNotify methods */
static asynStatus lockPort(void *drvPvt,asynUser *pasynUser);
static asynStatus unlockPort(void *drvPvt,asynUser *pasynUser);
static asynLockPortNotify lockport = {lockPort,unlockPort};


/* Forward references for asynInt32Digital methods */
static asynStatus readInt32(void* ppvt,asynUser* pasynUser,epicsInt32* value);
static asynStatus writeInt32(void* ppvt,asynUser* pasynUser,epicsInt32 value);


/* Forward references for asynUInt32Digital methods */
static asynStatus readUInt32(void* ppvt,asynUser* pasynUser,epicsUInt32* value,epicsUInt32 mask);
static asynStatus writeUInt32(void* ppvt,asynUser* pasynUser,epicsUInt32 value,epicsUInt32 mask);


/* Forward references for asynFloat64 methods */
static asynStatus readFloat64(void* ppvt,asynUser* pasynUser,epicsFloat64* value);
static asynStatus writeFloat64(void* ppvt,asynUser* pasynUser,epicsFloat64 value);


/* Define macros */
#define ISOK(s) (asynSuccess==(s))
#define ISNOTOK(s) (!ISOK(s))


/****************************************************************************
 * Define public interface methods
 ****************************************************************************/
static int drvGpibInit(const char* port,const char* gport)
{
    int len,attr;
    Port* pport;
    Gpibport* pgpib;
    asynUser* pasynUser;
    asynInt32* pasynInt32;
    asynFloat64* pasynFloat64;
    asynUInt32Digital* pasynUInt32;

    len = strlen(port)+strlen(gport)+2;
    len += sizeof(Port)+sizeof(Gpibport)+sizeof(asynInt32)+sizeof(asynFloat64)+sizeof(asynUInt32Digital);
    pport = (Port*)callocMustSucceed(len,sizeof(char),"drvGpibInit");

    pgpib = (Gpibport*)(pport + 1);
    pasynInt32 = (asynInt32*)(pgpib + 1);
    pasynUInt32 = (asynUInt32Digital*)(pasynInt32 + 1);
    pasynFloat64 = (asynFloat64*)(pasynUInt32 + 1);

    pport->name = (char*)(pasynFloat64 + 1);
    pgpib->name = (pport->name + strlen(port) + 1);

    pport->con = FALSE;
    pport->pgpib = pgpib;
    strcpy(pport->name,port);

    if( initGpibPort(pport,gport,-1) )
    {
        printf("drvGpibInit::failure to initialize GPIB port %s\n",gport);
        free(pport);
        return( -1 );
    }

    attr = (pgpib->canBlock)?(ASYN_MULTIDEVICE|ASYN_CANBLOCK):ASYN_MULTIDEVICE;
    if( pasynManager->registerPort(port,attr,pgpib->autoConnect,0,0) )
    {
        printf("drvGpibInit::init %s: failure to register port\n",port);
        free(pport);
        return( -1 );
    }

    pport->asynCommon.interfaceType = asynCommonType;
    pport->asynCommon.pinterface = &common;
    pport->asynCommon.drvPvt = pport;

    if( pasynManager->registerInterface(port,&pport->asynCommon) )
    {
        printf("drvGpibInit::init %s: failure to register asynCommon\n",port);
        return( -1 );
    }

    pport->asynDrvUser.interfaceType = asynDrvUserType;
    pport->asynDrvUser.pinterface = &drvuser;
    pport->asynDrvUser.drvPvt = pport;

    if( pasynManager->registerInterface(port,&pport->asynDrvUser) )
    {
        printf("drvGpibInit::init %s: failure to register asynDrvUser\n",port);
        return( -1 );
    }

    pport->asynLockPort.interfaceType = asynLockPortNotifyType;
    pport->asynLockPort.pinterface = &lockport;
    pport->asynLockPort.drvPvt = pport;

    if( pasynManager->registerInterface(port,&pport->asynLockPort) )
    {
        printf("drvGpibInit::failure to register asynLockPort\n");
        return( -1 );
    }

    pasynInt32->read = readInt32;
    pasynInt32->write = writeInt32;
    pport->asynInt32.interfaceType = asynInt32Type;
    pport->asynInt32.pinterface = pasynInt32;
    pport->asynInt32.drvPvt = pport;

    if( pasynInt32Base->initialize(port,&pport->asynInt32) )
    {
        printf("drvGpibInit::init %s: failure to initialize asynInt32Base\n",port);
        return( -1 );
    }

    pasynUInt32->read = readUInt32;
    pasynUInt32->write = writeUInt32;
    pport->asynUInt32.interfaceType = asynUInt32DigitalType;
    pport->asynUInt32.pinterface = pasynUInt32;
    pport->asynUInt32.drvPvt = pport;

    if( pasynUInt32DigitalBase->initialize(port,&pport->asynUInt32) )
    {
        printf("drvDioInit::init %s: failure to initialize asynUInt32DigitalBase\n",port);
        return( -1 );
    }

    pasynFloat64->read = readFloat64;
    pasynFloat64->write = writeFloat64;
    pport->asynFloat64.interfaceType = asynFloat64Type;
    pport->asynFloat64.pinterface = pasynFloat64;
    pport->asynFloat64.drvPvt = pport;

    if( pasynFloat64Base->initialize(port,&pport->asynFloat64) )
    {
        printf("drvSenInit::init %s: failure to initialize asynFloat64Base\n",port);
        return( -1 );
    }

    if( (pasynUser = pasynManager->createAsynUser(NULL,NULL)) )
    {
        pport->pasynUser = pasynUser;
        pasynUser->userPvt = pport;
        pasynUser->timeout = COMTMO;
    }
    else
    {
        printf("drvGpibInit::create asynUser failure\n");
        return( -1 );
    }

    if( pasynManager->connectDevice(pasynUser,port,-1) )
    {
        printf("drvGpibInit::failure to connect with device %s\n",port);
        pport->con = 0;
        return( -1 );
    }

    if( pports )
        pport->pport = pports;
    pports = pport;

    return( 0 );
}


static int drvGpibConf(const char* port,int addr,int type,int chns)
{
    int i,len;
    Port* pport;
    Instr* pinstr;

    for( pport = pports; pport; pport = pport->pport )
        if( epicsStrCaseCmp(pport->name,port) == 0 )
            break;

    if( pport == NULL )
    {
        printf("drvGpibConf::failure to locate port %s\n",port);
        return( -1 );
    }

    if( addr >= 0 && addr < INSTRMAX )
    {
        pinstr = &pport->instr[addr];
    }
    else
    {
        printf("drvGpibConf::invalid address %d for port %s\n",addr,port);
        return( - 1 );
    }

    switch( type )
    {
    default:
        printf("drvGpibInit::init %s: unknown board type %d\n",port,type);
        free(pport);
        return( - 1 );
    }

    len = sizeof(Chn) * chns;
    pinstr->chn = (Chn*)callocMustSucceed(len,sizeof(char),"drvGpibConf");

    pinstr->con = FALSE;
    pinstr->chns = chns;
    pinstr->addr = addr;

    for( i = 0; i < pinstr->chns; ++i )
    {
        pinstr->chn[i].chn = i;
        pinstr->chn[i].con = FALSE;
    }

    return( 0 );
}


static asynStatus initGpibPort(Port* port,const char* gport,int addr)
{
    asynUser* pasynUser;
    asynInterface* pasynIface;
    Gpibport* pgpib = port->pgpib;

    pasynUser = pasynManager->createAsynUser(NULL,NULL);
    if( pasynUser == NULL )
        return( asynError );

    if( pasynManager->connectDevice(pasynUser,gport,addr) )
    {
        printf("initGpibPort::failure to connect with device %s @ %d\n",gport,addr);
        pasynManager->freeAsynUser(pasynUser);
        return( asynError );
    }

    if( pasynManager->isMultiDevice(pasynUser,gport,&pgpib->multiDevice) )
    {
        printf("initGpibPort::failure to determine if %s @ %d is multi device\n",gport,addr);
        pasynManager->disconnect(pasynUser);
        pasynManager->freeAsynUser(pasynUser);
        return( asynError );
    }

    if( pgpib->multiDevice )
    {
        printf("initGpibPort::%s cannot be configured as multi device\n",gport);
        pasynManager->disconnect(pasynUser);
        pasynManager->freeAsynUser(pasynUser);
        return( asynError );
    }

    if( pasynManager->canBlock(pasynUser,&pgpib->canBlock) )
    {
        printf("initGpibPort::failure to determine if %s @ %d can block\n",gport,addr);
        pasynManager->disconnect(pasynUser);
        pasynManager->freeAsynUser(pasynUser);
        return( asynError );
    }

    if( pasynManager->isAutoConnect(pasynUser,&pgpib->autoConnect) )
    {
        printf("initGpibPort::failure to determine if %s @ %d is autoconnect\n",gport,addr);
        pasynManager->disconnect(pasynUser);
        pasynManager->freeAsynUser(pasynUser);
        return( asynError );
    }

    pasynIface = pasynManager->findInterface(pasynUser,asynOctetType,1);
    if( pasynIface )
    {
        pgpib->pasynOctet = (asynOctet*)pasynIface->pinterface;
        pgpib->pasynOctetPvt = pasynIface->drvPvt;
    }
    else
    {
        printf("initGpibPort::failure to find interface %s\n",asynOctetType);
        pasynManager->disconnect(pasynUser);
        pasynManager->freeAsynUser(pasynUser);
        return( asynError );
    }

    pgpib->con = TRUE;
    pgpib->addr = addr;
    pgpib->pasynUser = pasynUser;
    strcpy(pgpib->name,gport);

    pasynUser->userPvt = port;
    pasynUser->timeout = COMTMO;

    return( asynSuccess );
}


/****************************************************************************
 * Define private interface asynCommon methods
 ****************************************************************************/
static void report(void* ppvt,FILE* fp,int details)
{
    int i;
    Port* pport = (Port*)ppvt;

    fprintf(fp,"    EPICS Brick Weeder Techologies\n");
    fprintf(fp,"        Port %s is connected to %s\n",pport->name,pport->pgpib->name);
    for( i = 0; i < INSTRMAX; ++i )
        if( pport->instr[i].con )
            fprintf(fp,"        Addr %d connected, type %d, %d channels\n",i,pport->instr[i].type,pport->instr[i].chns);

}


static asynStatus connect(void* ppvt,asynUser* pasynUser)
{
    int addr;
    asynStatus sts;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvGpib::connect %s:\n",pport->name);

    if( (sts = pasynManager->getAddr(pasynUser,&addr)) )
        return( sts );

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvGpib::connect %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    if( addr >= INSTRMAX  )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvGpib::connect %s: illegal addr %d\n",pport->name,addr);
        return( asynError );
    }

    if( addr >= 0 )
    {
        if( pport->instr[addr].con == TRUE )
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvGpib::connect %s: already connected to %d\n",pport->name,addr);
            return( asynError );
        }
    
        pport->instr[addr].con = TRUE;
        pport->instr[addr].addr = addr;
        pasynManager->exceptionConnect(pasynUser);
        return( asynSuccess );
    }

    if( pport->con )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvGpib::connect %s: already connected to %d\n",pport->name,addr);
        return( asynError );
    }

    pport->con = TRUE;
    pasynManager->exceptionConnect(pasynUser);
    return( asynSuccess );
}


static asynStatus disconnect(void* ppvt,asynUser* pasynUser)
{
    int addr;
    asynStatus sts;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvGpib::disconnect %s:\n",pport->name);

    if( (sts = pasynManager->getAddr(pasynUser,&addr)) )
        return( sts );

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvGpib::disconnect %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    if( addr >= INSTRMAX )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvGpib::disconnect %s: illegal addr %d\n",pport->name,addr);
        return( asynError );
    }

    if( addr >= 0 )
    {
        if( pport->instr[addr].con == FALSE )
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvGpib::disconnect %s: not connected to %d\n",pport->name,addr);
            return( asynError );
        }

        pport->instr[addr].con = FALSE;
        pasynManager->exceptionDisconnect(pasynUser);
        return( asynSuccess );
    }

    if( pport->con == FALSE )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvGpib::disconnect %s: not connected to %d\n",pport->name,addr);
        return( asynError );
    }

    pport->con = FALSE;
    pasynManager->exceptionDisconnect(pasynUser);
    return( asynSuccess );
}


/****************************************************************************
 * Define private interface asynDrvUser methods
 ****************************************************************************/
static asynStatus create(void* ppvt,asynUser* pasynUser,const char* drvInfo, const char** pptypeName,size_t* psize)
{
    int addr;
    Chn* pchn;
    Instr* pinstr;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvGpib::create %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        pinstr = &pport->instr[addr];
    else
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvGpib::create %s: getAddr failure\n",pport->name);
        return( asynError );
    }

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvGpib::create %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    // Place instrument-specific code here...
    pchn = NULL;

    return( asynSuccess );
}


static asynStatus gettype(void* ppvt,asynUser* pasynUser,const char** pptypeName,size_t* psize)
{
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvGpib::gettype %s:\n",pport->name);

    if( pptypeName )
        *pptypeName = NULL;
    if( psize )
        *psize = 0;

    return( asynSuccess );
}


static asynStatus destroy(void* ppvt,asynUser* pasynUser)
{
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvGpib::destroy %s:\n",pport->name);

    if( pport )
    {
        pasynUser->drvUser = NULL;
        return( asynSuccess );
    }
    else
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvGpib::destroy %s: called before create\n",pport->name);
        return( asynError );
    }

}


/****************************************************************************
 * Define private interface asynLockPort methods
 ****************************************************************************/
static asynStatus lockPort(void* drvPvt,asynUser* pasynUser)
{
    asynStatus sts;
    Port* pport = (Port*)drvPvt;
    Gpibport* pgpib = pport->pgpib;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvGpib::lockPort\n");

    sts = pasynManager->lockPort(pgpib->pasynUser);
    if( ISNOTOK(sts) )
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvGpib::lockPort %s error %s\n",pgpib->name,pgpib->pasynUser->errorMessage);

    return( sts );
}


static asynStatus unlockPort(void* drvPvt,asynUser* pasynUser)
{
    asynStatus sts;
    Port* pport = (Port*)drvPvt;
    Gpibport* pgpib = pport->pgpib;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvGpib::unlockPort\n");

    sts = pasynManager->unlockPort(pgpib->pasynUser);
    if( ISNOTOK(sts) )
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvGpib::unlockPort %s error %s\n",pgpib->name,pgpib->pasynUser->errorMessage);

    return( asynSuccess );
}


/****************************************************************************
 * Define private interface asynInt32 methods
 ****************************************************************************/
static asynStatus writeInt32(void* ppvt,asynUser* pasynUser,epicsInt32 value)
{
    int addr;
    Instr* pinstr;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvGpib::writeUInt32 %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        pinstr = &pport->instr[addr];
    else
        return( asynError );

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvGpib::writeUInt32 %s: asyn - 0x%8.8X, addr - %d, value - 0x%8.8X\n",pport->name,(int)pasynUser,(int)pinstr->addr,(int)value);
    // Process write

    return( asynSuccess );
}


static asynStatus readInt32(void* ppvt,asynUser* pasynUser,epicsInt32* value)
{
    int addr;
    Instr* pinstr;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvGpib::readUInt32 %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        pinstr = &pport->instr[addr];
    else
        return( asynError );

    // Process read
    *value = 0;
    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvGpib::readUInt32 %s: asyn - 0x%8.8X, addr - %d, value - 0x%8.8X\n",pport->name,(int)pasynUser,(int)pinstr->addr,(int)*value);

    return( asynSuccess );
}


/****************************************************************************
 * Define private interface asynUInt32Digital methods
 ****************************************************************************/
static asynStatus writeUInt32(void* ppvt,asynUser* pasynUser,epicsUInt32 value,epicsUInt32 mask)
{
    int addr;
    Instr* pinstr;
    Port* pport = (Port*)ppvt;
    Chn* pchn = (Chn*)pasynUser->drvUser;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvGpib::writeUInt32 %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        pinstr = &pport->instr[addr];
    else
        return( asynError );

    value &= mask;
    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvGpib::writeUInt32 %s: asyn - 0x%8.8X, addr - %d, chan - %d, mask - 0x%8.8X, value - 0x%8.8X\n",pport->name,(int)pasynUser,addr,(int)pchn->chn,(int)mask,(int)value);
    // Process write

    return( asynSuccess );
}


static asynStatus readUInt32(void* ppvt,asynUser* pasynUser,epicsUInt32* value,epicsUInt32 mask)
{
    int addr;
    Instr* pinstr;
    Port* pport = (Port*)ppvt;
    Chn* pchn = (Chn*)pasynUser->drvUser;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvGpib::readUInt32 %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        pinstr = &pport->instr[addr];
    else
        return( asynError );

    // Process read
    *value &= mask;

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvGpib::readUInt32 %s: asyn - 0x%8.8X, addr - %d, chan - %d, mask - 0x%8.8X, value - 0x%8.8X\n",pport->name,(int)pasynUser,addr,(int)pchn->chn,(int)mask,(int)*value);

    return( asynSuccess );
}


/****************************************************************************
 * Define private interface asynFloat64 methods
 ****************************************************************************/
static asynStatus writeFloat64(void* ppvt,asynUser* pasynUser,epicsFloat64 value)
{
    int addr;
    Instr* pinstr;
    Port* pport = (Port*)ppvt;
    Chn* pchn = (Chn*)pasynUser->drvUser;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvGpib::writeFloat64 %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        pinstr = &pport->instr[addr];
    else
        return( asynError );

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvGpib::writeFloat64 %s: asyn - 0x%8.8X, addr - %d, chan - %d, value - %f\n",pport->name,(int)pasynUser,addr,(int)pchn->chn,(int)value);
    // Process write

    return( asynSuccess );
}


static asynStatus readFloat64(void* ppvt,asynUser* pasynUser,epicsFloat64* value)
{
    int addr;
    Instr* pinstr;
    Port* pport = (Port*)ppvt;
    Chn* pchn = (Chn*)pasynUser->drvUser;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvGpib::readFloat64 %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        pinstr = &pport->instr[addr];
    else
        return( asynError );

    // Process read
    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvGpib::readFloat64 %s: asyn - 0x%8.8X, addr - %d, chan - %d, value - %f\n",pport->name,(int)pasynUser,addr,(int)pchn->chn,(int)*value);

    return( asynSuccess );
}


/****************************************************************************
 * Register public methods
 ****************************************************************************/

/* Initialization method definitions */
static const iocshArg drvGpibInitArg0 = {"port",iocshArgString};
static const iocshArg drvGpibInitArg1 = {"gport",iocshArgString};
static const iocshArg* drvGpibInitArgs[]= {&drvGpibInitArg0,&drvGpibInitArg1};
static const iocshFuncDef drvGpibInitFuncDef = {"drvGpibInit",2,drvGpibInitArgs};
static void drvGpibInitCallFunc(const iocshArgBuf* args)
{
    drvGpibInit(args[0].sval,args[1].sval);
}

/* Initialization method definitions */
static const iocshArg drvGpibCfgArg0 = {"port",iocshArgString};
static const iocshArg drvGpibCfgArg1 = {"addr",iocshArgInt};
static const iocshArg drvGpibCfgArg2 = {"type",iocshArgInt};
static const iocshArg drvGpibCfgArg3 = {"chns",iocshArgInt};
static const iocshArg* drvGpibCfgArgs[]= {&drvGpibCfgArg0,&drvGpibCfgArg1,&drvGpibCfgArg2,&drvGpibCfgArg3};
static const iocshFuncDef drvGpibCfgFuncDef = {"drvGpibConf",4,drvGpibCfgArgs};
static void drvGpibCfgCallFunc(const iocshArgBuf* args)
{
    drvGpibConf(args[0].sval,args[1].ival,args[2].ival,args[3].ival);
}

/* Registration method */
static void drvGpibRegister(void)
{
    static int firstTime = 1;

    if( firstTime )
    {
        firstTime = 0;
        iocshRegister( &drvGpibCfgFuncDef, drvGpibCfgCallFunc );
        iocshRegister( &drvGpibInitFuncDef, drvGpibInitCallFunc );
    }
}
epicsExportRegistrar( drvGpibRegister );
