/*

                          Argonne National Laboratory
                            APS Operations Division
                     Beamline Controls and Data Acquisition

                         EPICS Brick Scaler Port Driver



 -----------------------------------------------------------------------------
                                COPYRIGHT NOTICE
 -----------------------------------------------------------------------------
   Copyright (c) 2002 The University of Chicago, as Operator of Argonne
      National Laboratory.
   Copyright (c) 2002 The Regents of the University of California, as
      Operator of Los Alamos National Laboratory.
   Synapps Versions 4-5
   and higher are distributed subject to a Software License Agreement found
   in file LICENSE that is included with this distribution.
 -----------------------------------------------------------------------------

 Description
    This module provides support for a multiple device port driver. To
    initialize the driver, the method drvScalerInit() is called from the
    startup script with the following calling sequence.

        drvScalerInit(myport,ioport,type,inst,addr,ctrs)

        Where:
            myport - Scaler port driver name (i.e. "Scaler")
            ioport - Digital IO port driver name (i.e. "DIO")
            type   - Board type (see boardtypes.h)
            inst   - Board instance
            addr   - Board address, zero for GDIO board
            ctrs   - Number of counters


    The method dbior can be called from the IOC shell to display the current
    status of the driver.


 Developer notes:
    - This requires that either the Athena's or RMM digital IO has been
      configure where registers 0 and 1 are input and register2 is output.

 Source control info:
    Modified by:    $Author: dkline $
                    $Date: 2006/07/24 19:02:29 $
                    $Revision: 1.5 $

 =============================================================================
 History:
 Author: David M. Kline
 -----------------------------------------------------------------------------
 2006-Mar-01  DMK  Derived from the EPICS brick drvDio multidevice driver.
 2006-Apr-28  DMK  Initial development complete.
 2006-May-19  DMK  Added parameters to initialization method.
 2006-Jul-07  DMK  Support asynUInt32Digital for command messages.
 -----------------------------------------------------------------------------

*/


/* EPICS base version-specific definitions (must be performed first) */
#include <epicsVersion.h>
#define LT_EPICSBASE(v,r,l) (EPICS_VERSION<(v)||(EPICS_VERSION==(v)&&(EPICS_REVISION<(r)||(EPICS_REVISION==(r)&&EPICS_MODIFICATION<(l)))))


/* Evaluate EPICS base */
#if LT_EPICSBASE(3,14,6)
    #error "EPICS base must be 3.14.6 or greater"
#endif

/* System related include files */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* EPICS system related include files */
#include <iocsh.h>
#include <epicsStdio.h>
#include <cantProceed.h>
#include <epicsString.h>
#include <epicsExport.h>
#include <epicsThread.h>


/* EPICS synApps/Asyn related include files */
#include <asynDriver.h>
#include <asynDrvUser.h>
#include <asynInt32.h>
#include <asynUInt32Digital.h>


/* DSC related include files */
#include "dscsup.h"
#include "drvScaler.h"


/* Define symbolic constants */
#define TRUE       (1)
#define FALSE      (0)
#define K_COMTMO   (1)

#define ENA        (0x4)
#define CLK        (0x2)
#define DAT        (0x1)
#define DIS        (0x0)

#define CTR_BIT    (6)
#define CMD_BIT    (5)
#define ADR_BIT    (2)
#define CMD_SIZ    ((CMD_BIT+CTR_BIT+ADR_BIT)*2)
#define DAT_BIT    (32)
#define DAT_SIZ    (DAT_BIT*2)


/* Forward struct declarations */
typedef struct Dio Dio;
typedef struct Ctr Ctr;
typedef struct Port Port;
typedef struct Ioport Ioport;


/* Define digital IO register structure */
struct Dio
{
    int                isConn;
    asynUser*          pasynUser;
    asynUInt32Digital* pasynUInt32;
    void*              pasynUInt32Pvt;
};


/* Declare digital IO port structure */
struct Ioport
{
    char*              name;
    int                addr;
    int                isConn;

    /* Interface info */
    asynUser*          pasynUser;
    asynUInt32Digital* pasynUInt32;
    void*              pasynUInt32Pvt;

    /* DIO Interface info */
    Dio                dios[COM_DIO_REGS];
};


/* Declare counter info structure */
struct Ctr
{
    int con;
    int num;
    int cmd;
    int sts;
    epicsInt32 preset;
};


/* Declare port structure */
struct Port
{
    char*         name;
    int           type;
    int           inst;
    int           addr;
    int           isConn;
    epicsMutexId  syncLock;
    asynInterface asynInt32;
    asynInterface asynUInt32;
    asynInterface asynCommon;
    asynInterface asynDrvUser;

    /* Counter info */
    int            ctrs;
    Ctr*           ctr;

    /* Digital IO port info */
    Ioport*        iop;
};


/* Public interface forward references */
static int drvScalerInit(const char* myport,const char* ioport,int type,int inst,int addr,int ctrs);


/* Forward references for asynCommon methods */
static void report(void* ppvt,FILE* fp,int details);
static asynStatus connect(void* ppvt,asynUser* pasynUser);
static asynStatus disconnect(void* ppvt,asynUser* pasynUser);
static asynCommon common = {report,connect,disconnect};


/* Forward references for asynDrvUser methods */
static asynStatus create(void* ppvt,asynUser* pasynUser,const char* drvInfo, const char** pptypeName,size_t* psize);
static asynStatus destroy(void* ppvt,asynUser* pasynUser);
static asynStatus gettype(void* ppvt,asynUser* pasynUser,const char** pptypeName,size_t* psize);
static asynDrvUser drvuser = {create,gettype,destroy};


/* Forward references for asynInt32 methods */
static asynStatus readInt32(void* ppvt,asynUser* pasynUser,epicsInt32* value);
static asynStatus writeInt32(void* ppvt,asynUser* pasynUser,epicsInt32 value);


/* Forward references for asynUInt32Digital methods */
static asynStatus readUInt32(void* ppvt,asynUser* pasynUser,epicsUInt32* value,epicsUInt32 mask);
static asynStatus writeUInt32(void* ppvt,asynUser* pasynUser,epicsUInt32 value,epicsUInt32 mask);


/* Forward references for misc. methods */
static int power(int x,int y);
static asynStatus initIoport(Port* port,const char* name);
static void readScaler(Ioport* ioport,Ctr* pctr,epicsInt32* value);
static void sendCommand(Ioport* ioport,int ctr,epicsInt32 command,epicsInt32 value,int isdata);

/* Define macros */
#define ISOK(s) (asynSuccess==(s))
#define ISNOTOK(s) (!ISOK(s))


/* Define local variants */
static epicsUInt32 maxadr;
static epicsUInt32 maxcmd;
static epicsUInt32 maxctr;


/****************************************************************************
 * Define public interface methods
 ****************************************************************************/
static int drvScalerInit(const char* myport,const char* ioport,int type,int inst,int addr,int ctrs)
{
    int len,ns1,ns2,cs;
    Port* pport;
    asynStatus sts;
    asynInt32* pasynInt32;
    asynUInt32Digital* pasynUInt32;

    maxadr = (epicsUInt32)power(2,ADR_BIT);
    maxcmd = (epicsUInt32)power(2,CMD_BIT);
    maxctr = (epicsUInt32)power(2,CTR_BIT);

    ns1 = strlen(myport)+1;
    ns2 = strlen(ioport)+1;
    cs = (sizeof(Ctr)*ctrs);
    len = ns1+ns2+cs;
    len += sizeof(Port)+sizeof(Ioport)+sizeof(asynInt32)+sizeof(asynUInt32Digital);
    pport = (Port*)callocMustSucceed(len,sizeof(char),"drvScalerInit");

    pasynInt32 = (asynInt32*)(pport + 1);
    pasynUInt32 = (asynUInt32Digital*)(pasynInt32 + 1);
    pport->name = (char*)(pasynUInt32 + 1);
    pport->ctr = (Ctr*)(pport->name + ns1);

    pport->iop = (Ioport*)(pport->ctr + ctrs);
    pport->iop->name = (char*)(pport->iop + 1);

    pport->type = type;
    pport->inst = inst;
    pport->ctrs = ctrs;
    pport->isConn = FALSE;
    strcpy(pport->name,myport);
    pport->syncLock = epicsMutexMustCreate();

    switch( type )
    {
    case SCL_GDIO:
        pport->addr = addr;
        break;

    default:
        printf("drvScalerInit::init %s: unknown board type %d\n",myport,type);
        free(pport);
        return( - 1 );
    }

    sts = initIoport(pport,ioport);
    if( ISNOTOK(sts) )
    {
        printf("drvScalerInit::failure to initialize port %s\n",ioport);
        free(pport);
        return( -1 );
    }

    if( pasynManager->registerPort(myport,ASYN_MULTIDEVICE|ASYN_CANBLOCK,1,0,0) )
    {
        printf("drvScalerInit::init %s: failure to register port\n",myport);
        free(pport);
        return( -1 );
    }

    pport->asynCommon.interfaceType = asynCommonType;
    pport->asynCommon.pinterface = &common;
    pport->asynCommon.drvPvt = pport;

    if( pasynManager->registerInterface(myport,&pport->asynCommon) )
    {
        printf("drvScalerInit::init %s: failure to register asynCommon\n",myport);
        return( -1 );
    }

    pport->asynDrvUser.interfaceType = asynDrvUserType;
    pport->asynDrvUser.pinterface = &drvuser;
    pport->asynDrvUser.drvPvt = pport;

    if( pasynManager->registerInterface(myport,&pport->asynDrvUser) )
    {
        printf("drvScalerInit::init %s: failure to register asynDrvUser\n",myport);
        return( -1 );
    }

    pasynInt32->read = readInt32;
    pasynInt32->write = writeInt32;
    pport->asynInt32.interfaceType = asynInt32Type;
    pport->asynInt32.pinterface = pasynInt32;
    pport->asynInt32.drvPvt = pport;

    if( pasynInt32Base->initialize(myport,&pport->asynInt32) )
    {
        printf("drvScalerInit::init %s: failure to initialize asynInt32Base\n",myport);
        return( -1 );
    }

    pasynUInt32->read = readUInt32;
    pasynUInt32->write = writeUInt32;
    pport->asynUInt32.interfaceType = asynUInt32DigitalType;
    pport->asynUInt32.pinterface = pasynUInt32;
    pport->asynUInt32.drvPvt = pport;

    if( pasynUInt32DigitalBase->initialize(myport,&pport->asynUInt32) )
    {
        printf("drvScalerInit::init %s: failure to initialize asynUInt32DigitalBase\n",myport);
        return( -1 );
    }

    return( 0 );
}


static asynStatus initIoport(Port* pport,const char* name)
{
    asynStatus sts;
    asynUser *pasynUser;
    asynInterface* pasynIface;
    Ioport *ioport = pport->iop;

    pasynUser = pasynManager->createAsynUser(NULL,NULL);
    if( pasynUser == NULL )
        return( asynError );

    sts = pasynManager->connectDevice(pasynUser,name,-1);
    if( ISNOTOK(sts) )
    {
        printf("initIoport::failure to connect with device %s @ %d\n",name,-1);
        pasynManager->freeAsynUser(pasynUser);
        return( asynError );
    }

    pasynIface = pasynManager->findInterface(pasynUser,asynUInt32DigitalType,1);
    if( pasynIface )
    {
        ioport->pasynUInt32 = (asynUInt32Digital*)pasynIface->pinterface;
        ioport->pasynUInt32Pvt = pasynIface->drvPvt;
    }
    else
    {
        printf("initIoport::failure to find interface %s\n",asynUInt32DigitalType);
        pasynManager->disconnect(pasynUser);
        pasynManager->freeAsynUser(pasynUser);
        return( asynError );
    }

    ioport->addr = -1;
    ioport->isConn = TRUE;
    strcpy(ioport->name,name);
    ioport->pasynUser = pasynUser;

    for( int i=0; i<COM_DIO_REGS; ++i )
    {
        pasynUser = pasynManager->createAsynUser(NULL,NULL);
        if( pasynUser == NULL )
        {
            printf("initIoport::failure to create asynUser for device %s @ %d\n",name,i);
            return( asynError );
        }

        sts = pasynManager->connectDevice(pasynUser,name,i);
        if( ISNOTOK(sts) )
        {
            printf("initIoport::failure to connect with device %s @ %d\n",name,-1);
            pasynManager->freeAsynUser(pasynUser);
            return( asynError );
        }

        pasynIface = pasynManager->findInterface(pasynUser,asynUInt32DigitalType,1);
        if( pasynIface )
        {
            pasynUser->userPvt = pport;
            pasynUser->timeout = K_COMTMO;

            ioport->dios[i].isConn = TRUE;
            ioport->dios[i].pasynUser = pasynUser;
            ioport->dios[i].pasynUInt32 = (asynUInt32Digital*)pasynIface->pinterface;
            ioport->dios[i].pasynUInt32Pvt = pasynIface->drvPvt;
        }
        else
        {
            printf("initIoport::failure to find interface %s\n",asynUInt32DigitalType);
            pasynManager->disconnect(pasynUser);
            pasynManager->freeAsynUser(pasynUser);
            return( asynError );
        }

    }

    return( asynSuccess );
}


static int power(int x,int y)
{
    int i,p;
    for(i=0,p=1; i<y; ++i) p *= x;
    return( p );
}


static void readScaler(Ioport* ioport,Ctr* pctr,epicsInt32* value)
{
    epicsUInt32 mask = 0xFF;
    epicsUInt32 dlo,dhi,loh,hih;
    Dio* diol = &ioport->dios[0];
    Dio* dioh = &ioport->dios[1];

    mask = 0xFF;
    sendCommand(ioport,pctr->num,CMD_LO,0,FALSE);
    diol->pasynUInt32->read(diol->pasynUInt32Pvt,diol->pasynUser,&dlo,mask);
    dioh->pasynUInt32->read(dioh->pasynUInt32Pvt,dioh->pasynUser,&dhi,mask);
    loh = (dhi << 8) | dlo;

    sendCommand(ioport,pctr->num,CMD_HI,0,FALSE);
    diol->pasynUInt32->read(diol->pasynUInt32Pvt,diol->pasynUser,&dlo,mask);
    dioh->pasynUInt32->read(dioh->pasynUInt32Pvt,dioh->pasynUser,&dhi,mask);
    hih = (dhi << 8) | dlo;

    *value = (hih << 16) | loh;
}


static void sendCommand(Ioport* ioport,int ctr,epicsInt32 command,epicsInt32 value,int isdata)
{
    int i;
    epicsInt32 dat;
    epicsUInt32 msk;
    Dio* dio = &ioport->dios[2];
    static epicsUInt32 mask = 0xFF;

    dio->pasynUInt32->write(dio->pasynUInt32Pvt,dio->pasynUser,DIS,mask);
    dio->pasynUInt32->write(dio->pasynUInt32Pvt,dio->pasynUser,ENA,mask);

    // Send data
    if(isdata==TRUE)
    {
        msk = 1 << ((DAT_SIZ/2)-1);

        for(i=0; i<DAT_SIZ; ++i)
        {
            dat = ENA;
            if((i&1)==0)
            {
                dat |= (CLK | ((msk & value) ? 1 : 0) );
                value <<= 1;
            }
            dio->pasynUInt32->write(dio->pasynUInt32Pvt,dio->pasynUser,dat,mask);
        }
    }
    dat = ENA;
    dio->pasynUInt32->write(dio->pasynUInt32Pvt,dio->pasynUser,dat,mask);

    // Send command
    msk = 1 << ((CMD_SIZ/2)-1);
    command = (command << CTR_BIT) | (ctr+1);
    // or in the board address at a later time

    for(i=0; i<CMD_SIZ; ++i)
    {
        dat = ENA;
        if((i&1)==0)
        {
            dat |= (CLK | ((msk & command) ? 1 : 0) );
            command <<= 1;
        }
        dio->pasynUInt32->write(dio->pasynUInt32Pvt,dio->pasynUser,dat,mask);
    }

    dio->pasynUInt32->write(dio->pasynUInt32Pvt,dio->pasynUser,DIS,mask);
}


/****************************************************************************
 * Define private interface asynCommon methods
 ****************************************************************************/
static void report(void* ppvt,FILE* fp,int details)
{
    int i;
    Port* pport = (Port*)ppvt;
    Ioport* ioport = pport->iop;

    fprintf(fp,"    EPICS Brick Scaler digital IO registers - %s\n",ioport->name);
    for( i=0; i<COM_DIO_REGS; ++i )
        fprintf(fp,"        reg %i connected %s \n",i,(ioport->dios[i].isConn)?"Yes":"No");
    fprintf(fp,"    EPICS Brick Scaler %d counters connected to %s\n",pport->ctrs,ioport->name);
    for( i=0; i<pport->ctrs; ++i )
        fprintf(fp,"        ctr %2.2d connected %s sts 0x%8.8X\n",i,(pport->ctr[i].con)?"Yes":"No",pport->ctr[i].sts);
}


static asynStatus connect(void* ppvt,asynUser* pasynUser)
{
    int addr;
    asynStatus sts;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvScaler::connect %s:\n",pport->name);

    if( (sts = pasynManager->getAddr(pasynUser,&addr)) )
        return( sts );

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvScaler::connect %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    if( addr > pport->ctrs )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvScaler::connect %s: illegal addr %d\n",pport->name,addr);
        return( asynError );
    }

    if( addr >= 0 )
    {
        if( pport->ctr[addr].con == TRUE )
	{
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvScaler::connect %s: already connected to %d\n",pport->name,addr);
            return( asynError );
	}
    
        pport->ctr[addr].con = TRUE;
        pport->ctr[addr].num = addr;
        pasynManager->exceptionConnect(pasynUser);
        return( asynSuccess );
    }

    if( pport->isConn )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvScaler::connect %s: already connected to %d\n",pport->name,addr);
        return( asynError );
    }

    pport->isConn = TRUE;
    pasynManager->exceptionConnect(pasynUser);
    return( asynSuccess );
}


static asynStatus disconnect(void* ppvt,asynUser* pasynUser)
{
    int addr;
    asynStatus sts;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvScaler::disconnect %s:\n",pport->name);

    if( (sts = pasynManager->getAddr(pasynUser,&addr)) )
        return( sts );

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvScaler::disconnect %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    if( addr > pport->ctrs )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvScaler::disconnect %s: illegal addr %d\n",pport->name,addr);
        return( asynError );
    }

    if( addr >= 0 )
    {
        if( pport->ctr[addr].con == FALSE )
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvScaler::disconnect %s: not connected to %d\n",pport->name,addr);
            return( asynError );
        }

        pport->ctr[addr].con = FALSE;
        pasynManager->exceptionDisconnect(pasynUser);
        return( asynSuccess );
    }

    if( pport->isConn == FALSE )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvScaler::disconnect %s: not connected to %d\n",pport->name,addr);
        return( asynError );
    }

    pport->isConn = FALSE;
    pasynManager->exceptionDisconnect(pasynUser);
    return( asynSuccess );
}


/****************************************************************************
 * Define private interface asynDrvUser methods
 ****************************************************************************/
static asynStatus create(void* ppvt,asynUser* pasynUser,const char* drvInfo, const char** pptypeName,size_t* psize)
{
    int addr;
    Ctr* pctr;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvScaler::create %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        pctr = &pport->ctr[addr];
    else
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvScaler::create %s: getAddr failure\n",pport->name);
        return( asynError );
    }

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvScaler::create %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    pctr->num = addr;
    pasynUser->drvUser = pctr;

    return( asynSuccess );
}


static asynStatus gettype(void* ppvt,asynUser* pasynUser,const char** pptypeName,size_t* psize)
{
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvScaler::gettype %s:\n",pport->name);

    if( pptypeName )
        *pptypeName = NULL;
    if( psize )
        *psize = 0;

    return( asynSuccess );
}


static asynStatus destroy(void* ppvt,asynUser* pasynUser)
{
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvScaler::destroy %s:\n",pport->name);

    if( pport )
    {
        pasynUser->drvUser = NULL;
        return( asynSuccess );
    }
    else
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvScaler::destroy %s: called before create\n",pport->name);
        return( asynError );
    }

}


/****************************************************************************
 * Define private interface asynUInt32Digital methods
 ****************************************************************************/
static asynStatus writeInt32(void* ppvt,asynUser* pasynUser,epicsInt32 value)
{
    int addr;
    Ctr* pctr;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvScaler::writeInt32 %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        pctr = &pport->ctr[addr];
    else
        return( asynError );

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvScaler::writeInt32 %s: asyn - 0x%8.8X, addr - %d, value - %d\n",pport->name,(int)pasynUser,(int)pctr->num,(int)value);

    epicsMutexMustLock(pport->syncLock);
    sendCommand(pport->iop,pctr->num,CMD_PST,value,TRUE);
    epicsMutexUnlock(pport->syncLock);

    return( asynSuccess );
}


static asynStatus readInt32(void* ppvt,asynUser* pasynUser,epicsInt32* value)
{
    int addr;
    Ctr* pctr;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvScaler::readInt32 %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
    {
        pctr = &pport->ctr[addr];
        pctr->preset = *value;
    }
    else
        return( asynError );

    epicsMutexMustLock(pport->syncLock);
    readScaler(pport->iop,pctr,value);
    epicsMutexUnlock(pport->syncLock);

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvScaler::readInt32 %s: asyn - 0x%8.8X, addr - %d, value - %d\n",pport->name,(int)pasynUser,(int)pctr->num,(int)*value);

    return( asynSuccess );
}


/****************************************************************************
 * Define private interface asynUInt32Digital methods
 ****************************************************************************/
static asynStatus writeUInt32(void* ppvt,asynUser* pasynUser,epicsUInt32 value,epicsUInt32 mask)
{
    int addr;
    Ctr* pctr;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvScaler::writeUInt32 %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        pctr = &pport->ctr[addr];
    else
        return( asynError );

    if( value < maxcmd )
    {
        value &= mask;
        pctr->cmd = value;
    }
    else
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvScaler::writeUInt32 %s: illegal command %d\n",pport->name,value);
        return( asynError );
    }

    if( value == CMD_ENA )
    {
        pctr->sts |= power(2,CMD_ENA);
        pctr->sts &= ~power(2,CMD_DIS);

    }
    else if( value == CMD_DIS )
    {
        pctr->sts &= ~power(2,CMD_ENA);
        pctr->sts |= power(2,CMD_DIS);
    }
    else if( value == CMD_UP  )
    {
        pctr->sts |= power(2,CMD_UP);
        pctr->sts &= ~power(2,CMD_DWN);
    }
    else if( value == CMD_DWN )
    {
        pctr->sts &= ~power(2,CMD_UP);
        pctr->sts |= power(2,CMD_DWN);
    }
    else if( value == CMD_IEA )
    {
        pctr->sts |= power(2,CMD_IEA);
        pctr->sts &= ~power(2,CMD_IDS);
    }
    else if( value == CMD_IDS )
    {
        pctr->sts &= ~power(2,CMD_IEA);
        pctr->sts |= power(2,CMD_IDS);
    }
    else if( value == CMD_GEA )
    {
        pctr->sts |= power(2,CMD_GEA);
        pctr->sts &= ~power(2,CMD_GDS);
    }
    else if( value == CMD_GDS )
    {
        pctr->sts &= ~power(2,CMD_GEA);
        pctr->sts |= power(2,CMD_GDS);
    }
    else if( value == CMD_RES )
        pctr->sts = 0;

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvScaler::writeUInt32 %s: asyn - 0x%8.8X, addr - %d, value - %d\n",pport->name,(int)pasynUser,pctr->num,(int)value);

    epicsMutexMustLock(pport->syncLock);
    sendCommand(pport->iop,pctr->num,value,0,FALSE);
    epicsMutexUnlock(pport->syncLock);

    return( asynSuccess );
}


static asynStatus readUInt32(void* ppvt,asynUser* pasynUser,epicsUInt32* value,epicsUInt32 mask)
{
    int addr;
    Ctr* pctr;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvScaler::readUInt32 %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        pctr = &pport->ctr[addr];
    else
        return( asynError );

    epicsMutexMustLock(pport->syncLock);
    *value = pctr->sts;
    epicsMutexUnlock(pport->syncLock);

    *value &= mask;
    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvScaler::readUInt32 %s: asyn - 0x%8.8X, addr - %d, value - %d\n",pport->name,(int)pasynUser,(int)pctr->num,(int)*value);

    return( asynSuccess );
}


/****************************************************************************
 * Register public methods
 ****************************************************************************/
/* Initialization method definitions */
static const iocshArg drvScalerInitArg0 = {"myport",iocshArgString};
static const iocshArg drvScalerInitArg1 = {"ioport",iocshArgString};
static const iocshArg drvScalerInitArg2 = {"type",iocshArgInt};
static const iocshArg drvScalerInitArg3 = {"inst",iocshArgInt};
static const iocshArg drvScalerInitArg4 = {"addr",iocshArgInt};
static const iocshArg drvScalerInitArg5 = {"ctrs",iocshArgInt};
static const iocshArg* drvScalerInitArgs[]= {&drvScalerInitArg0,&drvScalerInitArg1,&drvScalerInitArg2,&drvScalerInitArg3,&drvScalerInitArg4,&drvScalerInitArg5};
static const iocshFuncDef drvScalerInitFuncDef = {"drvScalerInit",6,drvScalerInitArgs};
static void drvScalerInitCallFunc(const iocshArgBuf* args)
{
    drvScalerInit(args[0].sval,args[1].sval,args[2].ival,args[3].ival,args[4].ival,args[5].ival);
}

/* Registration method */
static void drvScalerRegister(void)
{
    static int firstTime = 1;

    if( firstTime )
    {
        firstTime = 0;
        iocshRegister(&drvScalerInitFuncDef,drvScalerInitCallFunc);
    }
}
epicsExportRegistrar( drvScalerRegister );
