/*

                          Argonne National Laboratory
                            APS Operations Division
                     Beamline Controls and Data Acquisition

                  EPICS Brick Sensoray's Smart A/D Port Driver



 -----------------------------------------------------------------------------
                                COPYRIGHT NOTICE
 -----------------------------------------------------------------------------
   Copyright (c) 2002 The University of Chicago, as Operator of Argonne
      National Laboratory.
   Copyright (c) 2002 The Regents of the University of California, as
      Operator of Los Alamos National Laboratory.
   Synapps Versions 4-5
   and higher are distributed subject to a Software License Agreement found
   in file LICENSE that is included with this distribution.
 -----------------------------------------------------------------------------

 Description
    This module provides support for a multiple device port driver. To
    initialize the driver, the method drvSenInit() is called from the
    startup script with the following calling sequence.

        drvSenInit(port,type,inst,addr,freq,chns)

        Where:
            port - Port driver name (i.e. "SEN" )
            type - Board type (see boardtypes.h)
            inst - Board instance
            addr - Board address, zero for Athena
            chns - Channel count
            freq - Sample frequency (1Hz..5Hz)

        drvSenConfig(port,chn,type)

        Where:
            port - Port driver name (i.e. "SEN" )
            chn  - Channel number
            type - Sensor type

    The method dbior can be called from the IOC shell to display the current
    status of the driver.


 Developer notes:

 Source control info:
    Modified by:    $Author: dkline $
                    $Date: 2006/06/21 19:15:29 $
                    $Revision: 1.4 $

 =============================================================================
 History:
 Author: David M. Kline
 -----------------------------------------------------------------------------
 2006-Mar-09  DMK  Development begins. Only thermocouple support initially.
 2006-May-19  DMK  Added 'addr' parameter to initialization method.
 -----------------------------------------------------------------------------

*/


/* EPICS base version-specific definitions (must be performed first) */
#include <epicsVersion.h>
#define LT_EPICSBASE(v,r,l) (EPICS_VERSION<(v)||(EPICS_VERSION==(v)&&(EPICS_REVISION<(r)||(EPICS_REVISION==(r)&&EPICS_MODIFICATION<(l)))))


/* Evaluate EPICS base */
#if LT_EPICSBASE(3,14,6)
    #error "EPICS base must be 3.14.6 or greater"
#endif

/* System related include files */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* EPICS system related include files */
#include <iocsh.h>
#include <epicsStdio.h>
#include <cantProceed.h>
#include <epicsString.h>
#include <epicsExport.h>
#include <epicsThread.h>


/* EPICS synApps/Asyn related include files */
#include <asynDriver.h>
#include <asynInt32.h>
#include <asynDrvUser.h>
#include <asynFloat64.h>


/* Sensoray related include files */
#include "sensup.h"


/* Define symbolic constants */
#define TRUE       (1)
#define FALSE      (0)
#define K_COMTMO   (1)


/* Forward struct declarations */
typedef struct Chn Chn;
typedef struct Port Port;


/* Declare register info structure */
struct Chn
{
    int          isConn;
    int          chnNum;
    int          type;
    asynUser*    pasynUser;

    epicsInt32   intData;
    asynInt32*   pasynInt32;
    void*        pasynInt32Pvt;

    epicsFloat64 fltData;
    asynFloat64* pasynFloat64;
    void*        pasynFloat64Pvt;
};


/* Declare port structure */
struct Port
{
    char*         name;
    int           type;
    int           inst;
    int           addr;
    int           freq;
    int           isConn;
    short         pid;
    float         ver;
    epicsMutexId  syncLock;
    asynInterface asynCommon;
    asynInterface asynDrvUser;
    asynInterface asynInt32;
    asynInterface asynFloat64;

    /* Channels data */
    short chndat[SEN_CHNS];

    /* Channel info */
    int           chns;
    Chn*          chn;
};


/* Public interface forward references */
static void scanSen(Port* pport);
static int drvSenConfig(const char* port,int chn,int type);
static int drvSenInit(const char* port,int type,int inst,int addr,int chns,int freq);


/* Forward references for asynCommon methods */
static void report(void* ppvt,FILE* fp,int details);
static asynStatus connect(void* ppvt,asynUser* pasynUser);
static asynStatus disconnect(void* ppvt,asynUser* pasynUser);
static asynCommon common = {report,connect,disconnect};


/* Forward references for asynDrvUser methods */
static asynStatus create(void* ppvt,asynUser* pasynUser,const char* drvInfo, const char** pptypeName,size_t* psize);
static asynStatus destroy(void* ppvt,asynUser* pasynUser);
static asynStatus gettype(void* ppvt,asynUser* pasynUser,const char** pptypeName,size_t* psize);
static asynDrvUser drvuser = {create,gettype,destroy};


/* Forward references for asynInt32 methods */
static asynStatus readInt32(void* ppvt,asynUser* pasynUser,epicsInt32* value);
static asynStatus writeInt32(void* ppvt,asynUser* pasynUser,epicsInt32 value);


/* Forward references for asynFloat64 methods */
static asynStatus readFloat64(void* ppvt,asynUser* pasynUser,epicsFloat64* value);
static asynStatus writeFloat64(void* ppvt,asynUser* pasynUser,epicsFloat64 value);


/* Forward references for maintence methods */
static void processChannel(Chn* pchn,short iv);


/* Define macros */
#define ISOK(s) (asynSuccess==(s))
#define ISNOTOK(s) (!ISOK(s))


/****************************************************************************
 * Define public interface methods
 ****************************************************************************/
static int drvSenInit(const char* port,int type,int inst,int addr,int chns,int freq)
{
    int len,ns;
    Port* pport;
    unsigned int prio;
    asynInt32* pasynInt32;
    asynFloat64* pasynFloat64;

    ns = strlen(port)+1;
    len = sizeof(Port)+sizeof(asynInt32)+sizeof(asynFloat64)+ns+(sizeof(Chn)*chns);
    pport = (Port*)callocMustSucceed(len,sizeof(char),"drvSenInit");

    pasynInt32 = (asynInt32*)(pport + 1);
    pasynFloat64 = (asynFloat64*)(pasynInt32 + 1);
    pport->name = (char*)(pasynFloat64 + 1);
    pport->chn = (Chn*)(pport->name + ns);

    pport->type = type;
    pport->inst = inst;
    pport->chns = chns;
    pport->freq = freq;
    pport->isConn = FALSE;
    strcpy(pport->name,port);
    pport->syncLock = epicsMutexMustCreate();

    switch( type )
    {
    case SEN_SDA518:
        pport->addr = addr;
        break;

    default:
        printf("drvSenInit::init %s: unknown board type %d\n",port,type);
        free(pport);
        return( - 1 );
    }

    if( freq < 1 || freq > 5 )
    {
        printf("drvSenInit::init %s: invalid sample freqency %d\n",port,freq);
        return( 0 );
    }

    if( pasynManager->registerPort(port,ASYN_MULTIDEVICE|ASYN_CANBLOCK,1,0,0) )
    {
        printf("drvSenInit::init %s: failure to register port\n",port);
        free(pport);
        return( -1 );
    }

    pport->asynCommon.interfaceType = asynCommonType;
    pport->asynCommon.pinterface = &common;
    pport->asynCommon.drvPvt = pport;

    if( pasynManager->registerInterface(port,&pport->asynCommon) )
    {
        printf("drvSenInit::init %s: failure to register asynCommon\n",port);
        return( -1 );
    }

    pport->asynDrvUser.interfaceType = asynDrvUserType;
    pport->asynDrvUser.pinterface = &drvuser;
    pport->asynDrvUser.drvPvt = pport;

    if( pasynManager->registerInterface(port,&pport->asynDrvUser) )
    {
        printf("drvSenInit::init %s: failure to register asynDrvUser\n",port);
        return( -1 );
    }

    pasynInt32->read = readInt32;
    pasynInt32->write = writeInt32;
    pport->asynInt32.interfaceType = asynInt32Type;
    pport->asynInt32.pinterface = pasynInt32;
    pport->asynInt32.drvPvt = pport;

    if( pasynInt32Base->initialize(port,&pport->asynInt32) )
    {
        printf("drvSenInit::init %s: failure to initialize asynInt32Base\n",port);
        return( -1 );
    }

    pasynFloat64->read = readFloat64;
    pasynFloat64->write = writeFloat64;
    pport->asynFloat64.interfaceType = asynFloat64Type;
    pport->asynFloat64.pinterface = pasynFloat64;
    pport->asynFloat64.drvPvt = pport;

    if( pasynFloat64Base->initialize(port,&pport->asynFloat64) )
    {
        printf("drvSenInit::init %s: failure to initialize asynFloat64Base\n",port);
        return( -1 );
    }

    senReset(pport->addr);
    pport->ver = senGetVersion(pport->addr);
    pport->pid = senGetProductID(pport->addr);

    epicsThreadLowestPriorityLevelAbove(epicsThreadPriorityScanHigh,&prio);
    epicsThreadCreate("ScanSen",prio,epicsThreadGetStackSize(epicsThreadStackSmall),(EPICSTHREADFUNC)scanSen,pport);

    return( 0 );
}


static int drvSenConfig(const char* port,int chn,int type)
{
    asynStatus sts;
    asynUser *pasynUser;
    asynInterface* pasynIface;

    if( type < 1 || type > 25 )
    {
        printf("drvSenConfig::invalid sensor type %d\n",type);
        return( -1 );
    }

    pasynUser = pasynManager->createAsynUser(NULL,NULL);
    if( pasynUser == NULL )
        return( -1 );

    sts = pasynManager->connectDevice(pasynUser,port,chn);
    if( ISNOTOK(sts) )
    {
        printf("drvSenConfig::failure to connect with device %s @ %d\n",port,chn);
        pasynManager->freeAsynUser(pasynUser);
        return( -1 );
    }

    pasynIface = pasynManager->findInterface(pasynUser,asynInt32Type,1);
    if( pasynIface )
    {
        Port* pport = (Port*)pasynIface->drvPvt;
        Chn* pchn = &pport->chn[chn];

        if( chn >= pport->chns )
        {
            printf("drvSenConfig::invalid channel number %d\n",chn);
            pasynManager->freeAsynUser(pasynUser);
            return( -1 );
        }

        pchn->type = type;
        pchn->pasynUser = pasynUser;
        pchn->pasynInt32Pvt = pasynIface->drvPvt;
        pchn->pasynInt32 = (asynInt32*)pasynIface->pinterface;

        // Combine data here to create command/data etc...
        pchn->pasynInt32->write(pchn->pasynInt32Pvt,pchn->pasynUser,type);
    }
    else
    {
        printf("drvSenConfig::failure to find interface %s\n",asynInt32Type);
        pasynManager->disconnect(pasynUser);
        pasynManager->freeAsynUser(pasynUser);
        return( -1 );
    }

    return( 0 );
}

static void scanSen(Port* pport)
{
    int i;
    double freq = (0.824 / pport->freq);


    while(0==0)
    {
        epicsThreadSleep(freq);

        senGetChans(pport->addr,pport->chndat);
        for(i=0; i<pport->chns; ++i) processChannel(&pport->chn[i],pport->chndat[i]);
    }
    
}

static void processChannel(Chn* pchn,short iv)
{
    pchn->intData = (epicsInt32)iv;
    pchn->fltData = (epicsInt32)iv * senGetScalar(pchn->type);
}


/****************************************************************************
 * Define private interface asynCommon methods
 ****************************************************************************/
static void report(void* ppvt,FILE* fp,int details)
{
    int i,con;
    Chn* pchn;
    Port* pport = (Port*)ppvt;

    fprintf(fp,"    EPICS Brick Sensoray Smart A/D\n");
    fprintf(fp,"        Base address is 0x%4.4X has %d channels\n",pport->addr,pport->chns);
    fprintf(fp,"        Model %d V%2.2f\n",pport->pid,pport->ver);
    fprintf(fp,"        7409TC temperature id %2.2fF\n",senGetTemp(pport->addr));

    for( i = 0; i < pport->chns; ++i )
    {
        pchn = &pport->chn[i];
        con = pchn->isConn==TRUE;

        fprintf(fp,"        reg %d connected %s type %d\n",i,(con)?"Yes":"No",pchn->type);
    }
}


static asynStatus connect(void* ppvt,asynUser* pasynUser)
{
    int addr;
    asynStatus sts;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvSen::connect %s:\n",pport->name);

    if( (sts = pasynManager->getAddr(pasynUser,&addr)) )
        return( sts );

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvSen::connect %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    if( addr > pport->chns )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvSen::connect %s: illegal addr %d\n",pport->name,addr);
        return( asynError );
    }

    if( addr >= 0 )
    {
        if( pport->chn[addr].isConn == TRUE )
	{
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvSen::connect %s: already connected to %d\n",pport->name,addr);
            return( asynError );
	}
    
        pport->chn[addr].isConn = TRUE;
        pasynManager->exceptionConnect(pasynUser);
        return( asynSuccess );
    }

    if( pport->isConn )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvSen::connect %s: already connected to %d\n",pport->name,addr);
        return( asynError );
    }

    pport->isConn = TRUE;
    pasynManager->exceptionConnect(pasynUser);
    return( asynSuccess );
}


static asynStatus disconnect(void* ppvt,asynUser* pasynUser)
{
    int addr;
    asynStatus sts;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvSen::disconnect %s:\n",pport->name);

    if( (sts = pasynManager->getAddr(pasynUser,&addr)) )
        return( sts );

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvSen::disconnect %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    if( addr > pport->chns )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvSen::disconnect %s: illegal addr %d\n",pport->name,addr);
        return( asynError );
    }

    if( addr >= 0 )
    {
        if( pport->chn[addr].isConn == FALSE )
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvSen::disconnect %s: not connected to %d\n",pport->name,addr);
            return( asynError );
        }

        pport->chn[addr].isConn = FALSE;
        pasynManager->exceptionDisconnect(pasynUser);
        return( asynSuccess );
    }

    if( pport->isConn == FALSE )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvSen::disconnect %s: not connected to %d\n",pport->name,addr);
        return( asynError );
    }

    pport->isConn = FALSE;
    pasynManager->exceptionDisconnect(pasynUser);
    return( asynSuccess );
}


/****************************************************************************
 * Define private interface asynDrvUser methods
 ****************************************************************************/
static asynStatus create(void* ppvt,asynUser* pasynUser,const char* drvInfo, const char** pptypeName,size_t* psize)
{
    int addr;
    Chn* pchn;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvSen::create %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        pchn = &pport->chn[addr];
    else
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvSen::create %s: getAddr failure\n",pport->name);
        return( asynError );
    }

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvSen::create %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    switch( pport->type )
    {
    case SEN_SDA518:
        break;

    }

    pchn->chnNum = addr;
    pasynUser->drvUser = pchn;

    return( asynSuccess );
}


static asynStatus gettype(void* ppvt,asynUser* pasynUser,const char** pptypeName,size_t* psize)
{
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvSen::gettype %s:\n",pport->name);

    if( pptypeName )
        *pptypeName = NULL;
    if( psize )
        *psize = 0;

    return( asynSuccess );
}


static asynStatus destroy(void* ppvt,asynUser* pasynUser)
{
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvSen::destroy %s:\n",pport->name);

    if( pport )
    {
        pasynUser->drvUser = NULL;
        return( asynSuccess );
    }
    else
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvSen::destroy %s: called before create\n",pport->name);
        return( asynError );
    }

}


/****************************************************************************
 * Define private interface asynInt32 methods
 ****************************************************************************/
static asynStatus writeInt32(void* ppvt,asynUser* pasynUser,epicsInt32 value)
{
    int addr;
    Chn* pchn;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvSen::writeUInt32 %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        pchn = &pport->chn[addr];
    else
        return( asynError );

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvSen::writeUInt32 %s: asyn - 0x%8.8X, addr - %d, value - 0x%8.8X\n",pport->name,(int)pasynUser,(int)pchn->chnNum,(int)value);

    epicsMutexMustLock(pport->syncLock);
    senSetType(pport->addr,addr,value);
    epicsMutexUnlock(pport->syncLock);

    return( asynSuccess );
}


static asynStatus readInt32(void* ppvt,asynUser* pasynUser,epicsInt32* value)
{
    int addr;
    Chn* pchn;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvSen::readUInt32 %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        pchn = &pport->chn[addr];
    else
        return( asynError );

    *value = pchn->intData;
    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvSen::readUInt32 %s: asyn - 0x%8.8X, addr - %d, value - 0x%8.8X\n",pport->name,(int)pasynUser,(int)pchn->chnNum,(int)*value);

    return( asynSuccess );
}


/****************************************************************************
 * Define private interface asynFloat64 methods
 ****************************************************************************/
static asynStatus writeFloat64(void* ppvt,asynUser* pasynUser,epicsFloat64 value)
{
    int addr;
    Chn* pchn;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvSen::writeFloat64 %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        pchn = &pport->chn[addr];
    else
        return( asynError );

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvSen::writeFloat64 %s: asyn - 0x%8.8X, addr - %d, value - 0x%8.8X\n",pport->name,(int)pasynUser,(int)pchn->chnNum,(int)value);

    epicsMutexMustLock(pport->syncLock);
    epicsMutexUnlock(pport->syncLock);

    return( asynSuccess );
}


static asynStatus readFloat64(void* ppvt,asynUser* pasynUser,epicsFloat64* value)
{
    int addr;
    Chn* pchn;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvSen::readFloat64 %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        pchn = &pport->chn[addr];
    else
        return( asynError );

    *value = pchn->fltData;
    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvSen::readFloat64 %s: asyn - 0x%8.8X, addr - %d, value - 0x%8.8X\n",pport->name,(int)pasynUser,(int)pchn->chnNum,(int)*value);

    return( asynSuccess );
}


/****************************************************************************
 * Register public methods
 ****************************************************************************/

/* Initialization method definitions */
static const iocshArg drvSenInitArg0 = {"port",iocshArgString};
static const iocshArg drvSenInitArg1 = {"type",iocshArgInt};
static const iocshArg drvSenInitArg2 = {"inst",iocshArgInt};
static const iocshArg drvSenInitArg3 = {"addr",iocshArgInt};
static const iocshArg drvSenInitArg4 = {"chns",iocshArgInt};
static const iocshArg drvSenInitArg5 = {"freq",iocshArgInt};
static const iocshArg* drvSenInitArgs[]= {&drvSenInitArg0,&drvSenInitArg1,&drvSenInitArg2,&drvSenInitArg3,&drvSenInitArg4,&drvSenInitArg5};
static const iocshFuncDef drvSenInitFuncDef = {"drvSenInit",6,drvSenInitArgs};
static void drvSenInitCallFunc(const iocshArgBuf* args)
{
    drvSenInit(args[0].sval,args[1].ival,args[2].ival,args[3].ival,args[4].ival,args[5].ival);
}

/* Initialization method definitions */
static const iocshArg drvSenCfgArg0 = {"port",iocshArgString};
static const iocshArg drvSenCfgArg1 = {"chn",iocshArgInt};
static const iocshArg drvSenCfgArg2 = {"type",iocshArgInt};
static const iocshArg* drvSenCfgArgs[]= {&drvSenCfgArg0,&drvSenCfgArg1,&drvSenCfgArg2};
static const iocshFuncDef drvSenCfgFuncDef = {"drvSenConfig",3,drvSenCfgArgs};
static void drvSenCfgCallFunc(const iocshArgBuf* args)
{
    drvSenConfig(args[0].sval,args[1].ival,args[2].ival);
}

/* Registration method */
static void drvSenRegister(void)
{
    static int firstTime = 1;

    if( firstTime )
    {
        firstTime = 0;
        iocshRegister( &drvSenCfgFuncDef, drvSenCfgCallFunc );
        iocshRegister( &drvSenInitFuncDef, drvSenInitCallFunc );
    }
}
epicsExportRegistrar( drvSenRegister );
