/*

                          Argonne National Laboratory
                            APS Operations Division
                     Beamline Controls and Data Acquisition

                          EPICS Brick WDT Port Driver



 -----------------------------------------------------------------------------
                                COPYRIGHT NOTICE
 -----------------------------------------------------------------------------
   Copyright (c) 2002 The University of Chicago, as Operator of Argonne
      National Laboratory.
   Copyright (c) 2002 The Regents of the University of California, as
      Operator of Los Alamos National Laboratory.
   Synapps Versions 4-5
   and higher are distributed subject to a Software License Agreement found
   in file LICENSE that is included with this distribution.
 -----------------------------------------------------------------------------

 Description
    This module provides support for a multiple device port driver. To
    initialize the driver, the method drvWdtInit() is called from the
    startup script with the following calling sequence.

        drvWdtInit( port )

        Where:
            port - DSC port driver name (i.e. "WDT" )


    The method dbior can be called from the IOC shell to display the current
    status of the driver.


 Developer notes:

 Source control info:
    Modified by:    $Author: dkline $
                    $Date: 2006/05/26 20:36:42 $
                    $Revision: 1.2 $

 =============================================================================
 History:
 Author: David M. Kline
 -----------------------------------------------------------------------------
 2005-Nov-09  DMK  Derived from love multidevice driver.
 2005-Nov-15  DMK  Initial development complete.
 2005-Nov-28  DMK  Added SoftWDT thread processing.
 -----------------------------------------------------------------------------

*/


/* EPICS base version-specific definitions (must be performed first) */
#include <epicsVersion.h>
#define LT_EPICSBASE(v,r,l) (EPICS_VERSION<(v)||(EPICS_VERSION==(v)&&(EPICS_REVISION<(r)||(EPICS_REVISION==(r)&&EPICS_MODIFICATION<(l)))))


/* Evaluate EPICS base */
#if LT_EPICSBASE(3,14,6)
    #error "EPICS base must be 3.14.6 or greater"
#endif

/* System related include files */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* EPICS system related include files */
#include <iocsh.h>
#include <epicsStdio.h>
#include <cantProceed.h>
#include <epicsString.h>
#include <epicsExport.h>
#include <epicsThread.h>


/* EPICS synApps/Asyn related include files */
#include <asynDriver.h>
#include <asynDrvUser.h>
#include <asynUInt32Digital.h>


/* DSC related include files */
#include "dscsup.h"


/* Define symbolic constants */
#define TRUE       (1)
#define FALSE      (0)
#define K_WDTMAX   (4)


/* Forward struct declarations */
typedef struct Wdt Wdt;
typedef struct Port Port;


/* Declare register (A,B,C) info structure */
struct Wdt
{
    int isConn;
    int regNum;
};


/* Declare DSC port structure */
struct Port
{
    char*         name;
    int           isConn;
    epicsMutexId  syncLock;
    asynInterface asynUInt32;
    asynInterface asynCommon;
    asynInterface asynDrvUser;
    Wdt           wdt[K_WDTMAX];
};


/* Define local variants */
static Port* pports = NULL;
static void softWdt(Port* pport);

/* Public interface forward references */
static int drvWdtInit(const char* port);


/* Forward references for asynCommon methods */
static void report(void* ppvt,FILE* fp,int details);
static asynStatus connect(void* ppvt,asynUser* pasynUser);
static asynStatus disconnect(void* ppvt,asynUser* pasynUser);
static asynCommon common = {report,connect,disconnect};


/* Forward references for asynDrvUser methods */
static asynStatus create(void* ppvt,asynUser* pasynUser,const char* drvInfo, const char** pptypeName,size_t* psize);
static asynStatus destroy(void* ppvt,asynUser* pasynUser);
static asynStatus gettype(void* ppvt,asynUser* pasynUser,const char** pptypeName,size_t* psize);
static asynDrvUser drvuser = {create,gettype,destroy};


/* Forward references for asynUInt32Digital methods */
static asynStatus readUInt32(void* ppvt,asynUser* pasynUser,epicsUInt32* value,epicsUInt32 mask);
static asynStatus writeUInt32(void* ppvt,asynUser* pasynUser,epicsUInt32 value,epicsUInt32 mask);


/* Define macros */
#define ISOK(s) (asynSuccess==(s))
#define ISNOTOK(s) (!ISOK(s))


/****************************************************************************
 * Define public interface methods
 ****************************************************************************/
static int drvWdtInit(const char* port)
{
    int len;
    Port* pport;
    asynUInt32Digital* pasynUInt32;

    if( pports )
    {
        printf("drvWdt::init %s: is already initialized\n",port);
        return( 0 );
    }

    len = sizeof(Port)+sizeof(asynUInt32Digital)+strlen(port)+1;
    pport = (Port*)callocMustSucceed(len,sizeof(char),"drvWdtInit");

    pasynUInt32 = (asynUInt32Digital*)(pport + 1);
    pport->name = (char*)(pasynUInt32 + 1);

    pport->isConn = FALSE;
    strcpy(pport->name,port);
    pport->syncLock = epicsMutexMustCreate();

    if( pasynManager->registerPort(port,ASYN_MULTIDEVICE|ASYN_CANBLOCK,1,0,0) )
    {
        printf("drvWdt::init %s: failure to register port\n",port);
        free(pport);
        return( -1 );
    }

    pport->asynCommon.interfaceType = asynCommonType;
    pport->asynCommon.pinterface = &common;
    pport->asynCommon.drvPvt = pport;

    if( pasynManager->registerInterface(port,&pport->asynCommon) )
    {
        printf("drvWdt::init %s: failure to register asynCommon\n",port);
        return( -1 );
    }

    pport->asynDrvUser.interfaceType = asynDrvUserType;
    pport->asynDrvUser.pinterface = &drvuser;
    pport->asynDrvUser.drvPvt = pport;

    if( pasynManager->registerInterface(port,&pport->asynDrvUser) )
    {
        printf("drvWdt::init %s: failure to register asynDrvUser\n",port);
        return( -1 );
    }

    pasynUInt32->read = readUInt32;
    pasynUInt32->write = writeUInt32;
    pport->asynUInt32.interfaceType = asynUInt32DigitalType;
    pport->asynUInt32.pinterface = pasynUInt32;
    pport->asynUInt32.drvPvt = pport;

    if( pasynUInt32DigitalBase->initialize(port,&pport->asynUInt32) )
    {
        printf("drvWdt::init %s: failure to initialize asynUInt32DigitalBase\n",port);
        return( -1 );
    }

    pports = pport;
    epicsThreadCreate("SoftWDT",epicsThreadPriorityLow,epicsThreadGetStackSize(epicsThreadStackSmall),(EPICSTHREADFUNC)softWdt,pport);

    return( 0 );
}


static void softWdt(Port* pport)
{
    unsigned char selReg,ctlReg,cntReg;

    while(0==0)
    {
        epicsThreadSleep(0.1);

        epicsMutexMustLock(pport->syncLock);
        selReg = wdtRead(ATH_INPB_WDT_SELECT);
        ctlReg = wdtRead(ATH_INPB_WDT_CONTROL);
        cntReg = wdtRead(ATH_INPB_WDT_COUNTER);
        epicsMutexUnlock(pport->syncLock);

        if((selReg & ATH_WDT_SELECT_WDEN)==0) continue;
        if((ctlReg & (ATH_WDT_CONTROL_WDOEN|ATH_WDT_CONTROL_WDIEN))) continue;
        if(cntReg) wdtWrite(ATH_OUTB_WDT_TRIGGER,ATH_WDT_TRIGGER);
    }
    
}


/****************************************************************************
 * Define private interface asynCommon methods
 ****************************************************************************/
static void report(void* ppvt,FILE* fp,int details)
{
    int i;
    Port* pport = (Port*)ppvt;

    fprintf(fp,"    EPICS Brick watchdog @ 0x%4.4X\n",ATH_DAQ_BASE);

    for( i = 0; i < K_WDTMAX; ++i )
        fprintf(fp,"        reg %i connected %s\n",i,(pport->wdt[i].isConn==TRUE)?"Yes":"No");
}


static asynStatus connect(void* ppvt,asynUser* pasynUser)
{
    int addr;
    asynStatus sts;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWdt::connect %s:\n",pport->name);

    if( (sts = pasynManager->getAddr(pasynUser,&addr)) )
        return( sts );

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvWdt::connect %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    if( addr > K_WDTMAX )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWdt::connect %s: illegal addr %d\n",pport->name,addr);
        return( asynError );
    }

    if( addr >= 0 )
    {
        if( pport->wdt[addr].isConn == TRUE )
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWdt::connect %s: already connected to %d\n",pport->name,addr);
            return( asynError );
        }
    
        pport->wdt[addr].isConn = TRUE;
        pasynManager->exceptionConnect(pasynUser);
        return( asynSuccess );
    }

    if( pport->isConn )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWdt::connect %s: already connected to %d\n",pport->name,addr);
        return( asynError );
    }

    pport->isConn = TRUE;
    pasynManager->exceptionConnect(pasynUser);
    return( asynSuccess );
}


static asynStatus disconnect(void* ppvt,asynUser* pasynUser)
{
    int addr;
    asynStatus sts;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWdt::disconnect %s:\n",pport->name);

    if( (sts = pasynManager->getAddr(pasynUser,&addr)) )
        return( sts );

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvWdt::disconnect %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    if( addr > K_WDTMAX )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWdt::disconnect %s: illegal addr %d\n",pport->name,addr);
        return( asynError );
    }

    if( addr >= 0 )
    {
        if( pport->wdt[addr].isConn == FALSE )
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWdt::disconnect %s: not connected to %d\n",pport->name,addr);
            return( asynError );
        }

        pport->wdt[addr].isConn = FALSE;
        pasynManager->exceptionDisconnect(pasynUser);
        return( asynSuccess );
    }

    if( pport->isConn == FALSE )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWdt::disconnect %s: not connected to %d\n",pport->name,addr);
        return( asynError );
    }

    pport->isConn = FALSE;
    pasynManager->exceptionDisconnect(pasynUser);
    return( asynSuccess );
}


/****************************************************************************
 * Define private interface asynDrvUser methods
 ****************************************************************************/
static asynStatus create(void* ppvt,asynUser* pasynUser,const char* drvInfo, const char** pptypeName,size_t* psize)
{
    int addr;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWdt::create %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWdt::create %s: getAddr failure\n",pport->name);
        return( asynError );
    }

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvWdt::create %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    if( epicsStrCaseCmp("WDT",drvInfo) )
    {
        asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvWdt::create %s: failure to determine register %d direction\n",pport->name,addr);
        return( asynError );
    }

    pport->wdt[addr].regNum = addr;
    pasynUser->drvUser = &pport->wdt[addr];

    return( asynSuccess );
}


static asynStatus gettype(void* ppvt,asynUser* pasynUser,const char** pptypeName,size_t* psize)
{
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWdt::gettype %s:\n",pport->name);

    if( pptypeName )
        *pptypeName = NULL;
    if( psize )
        *psize = 0;

    return( asynSuccess );
}


static asynStatus destroy(void* ppvt,asynUser* pasynUser)
{
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWdt::destroy %s:\n",pport->name);

    if( pport )
    {
        pasynUser->drvUser = NULL;
        return( asynSuccess );
    }
    else
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWdt::destroy %s: called before create\n",pport->name);
        return( asynError );
    }

}


/****************************************************************************
 * Define private interface asynUInt32Digital methods
 ****************************************************************************/
static asynStatus writeUInt32(void* ppvt,asynUser* pasynUser,epicsUInt32 value,epicsUInt32 mask)
{
    Port* pport = (Port*)ppvt;
    Wdt* pwdt = (Wdt*)pasynUser->drvUser;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWdt::writeUInt32 %s:\n",pport->name);

    value &= mask;
    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvWdt::writeUInt32 %s: asyn - 0x%8.8X, addr - %d, mask - 0x%8.8X, value - 0x%8.8X\n",pport->name,(int)pasynUser,(int)pwdt->regNum,(int)mask,(int)value);

    epicsMutexMustLock(pport->syncLock);
    wdtWrite(pwdt->regNum,value);
    epicsMutexUnlock(pport->syncLock);

    return( asynSuccess );
}


static asynStatus readUInt32(void* ppvt,asynUser* pasynUser,epicsUInt32* value,epicsUInt32 mask)
{
    Port* pport = (Port*)ppvt;
    Wdt* pwdt = (Wdt*)pasynUser->drvUser;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWdt::readUInt32 %s:\n",pport->name);

    epicsMutexMustLock(pport->syncLock);
    *value = (epicsUInt32)wdtRead(pwdt->regNum);
    epicsMutexUnlock(pport->syncLock);

    *value &= mask;
    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvWdt::readUInt32 %s: asyn - 0x%8.8X, addr - %d, mask - 0x%8.8X, value - 0x%8.8X\n",pport->name,(int)pasynUser,(int)pwdt->regNum,(int)mask,(int)*value);

    return( asynSuccess );
}


/****************************************************************************
 * Register public methods
 ****************************************************************************/

/* Initialization method definitions */
static const iocshArg drvWdtInitArg0 = {"port",iocshArgString};
static const iocshArg* drvWdtInitArgs[]= {&drvWdtInitArg0};
static const iocshFuncDef drvWdtInitFuncDef = {"drvWdtInit",1,drvWdtInitArgs};
static void drvWdtInitCallFunc(const iocshArgBuf* args)
{
    drvWdtInit(args[0].sval);
}

/* Registration method */
static void drvWdtRegister(void)
{
    static int firstTime = 1;

    if( firstTime )
    {
        firstTime = 0;
        iocshRegister( &drvWdtInitFuncDef, drvWdtInitCallFunc );
    }
}
epicsExportRegistrar( drvWdtRegister );
