/*

                          Argonne National Laboratory
                            APS Operations Division
                     Beamline Controls and Data Acquisition

                   EPICS Brick Weeder Techologies Port Driver



 -----------------------------------------------------------------------------
                                COPYRIGHT NOTICE
 -----------------------------------------------------------------------------
   Copyright (c) 2002 The University of Chicago, as Operator of Argonne
      National Laboratory.
   Copyright (c) 2002 The Regents of the University of California, as
      Operator of Los Alamos National Laboratory.
   Synapps Versions 4-5
   and higher are distributed subject to a Software License Agreement found
   in file LICENSE that is included with this distribution.
 -----------------------------------------------------------------------------

 Description
    This module provides support for a multiple device port driver. To
    initialize the driver, the method drvWeedInit() is called from the
    startup script with the following calling sequence.

        drvWeedInit(port,serport,seraddr)
        Where:
            port    - Port driver name (i.e. "WTTCI" )
            serport - Serial port driver name (i.e. "S0" )
            seraddr - Serial port driver address

        drvWeedConfig(port,addr,type,chns)
        Where:
            port - Port driver name (i.e. "WTTCI" )
            addr - Controller address
            type - Controller type (boardtypes.h)
            chns - Channel count

    The method dbior can be called from the IOC shell to display the current
    status of the driver.


 Developer notes:

 Source control info:
    Modified by:    $Author: dkline $
                    $Date: 2006/06/28 20:35:14 $
                    $Revision: 1.3 $

 =============================================================================
 History:
 Author: David M. Kline
 -----------------------------------------------------------------------------
 2006-Mar-17  DMK  Development begins. Only thermocouple support initially.
 2006-Jun-28  DMK  Added support for WTDIO, WTTCI, and WTSSR modules.
 -----------------------------------------------------------------------------

*/


/* EPICS base version-specific definitions (must be performed first) */
#include <epicsVersion.h>
#define LT_EPICSBASE(v,r,l) (EPICS_VERSION<(v)||(EPICS_VERSION==(v)&&(EPICS_REVISION<(r)||(EPICS_REVISION==(r)&&EPICS_MODIFICATION<(l)))))


/* Evaluate EPICS base */
#if LT_EPICSBASE(3,14,6)
    #error "EPICS base must be 3.14.6 or greater"
#endif

/* System related include files */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* EPICS system related include files */
#include <iocsh.h>
#include <epicsStdio.h>
#include <cantProceed.h>
#include <epicsString.h>
#include <epicsExport.h>
#include <epicsThread.h>


/* EPICS synApps/Asyn related include files */
#include <asynDriver.h>
#include <asynInt32.h>
#include <asynOctet.h>
#include <asynDrvUser.h>
#include <asynFloat64.h>
#include <asynUInt32Digital.h>


/* Boardtypes related include files */
#include "boardtypes.h"


/* Define symbolic constants */
#define OUT      (0)
#define INP      (1)
#define TRUE     (1)
#define FALSE    (0)
#define COMTMO   (1)
#define INSTRMAX (32)


/* Forward struct declarations */
typedef struct Chn Chn;
typedef struct Port Port;
typedef struct Instr Instr;
typedef struct Serport Serport;


/* Define controller type enum */
typedef enum {DIO=200,TCI,SSR,SMD} Type;


/* Declare register info structure */
struct Chn
{
    int    isConn;
    int    chn;

    union
    {
        struct
        {
            int dir;
            int time;
        } DIO;

        struct
        {
            char type;
            char unit;
        } TCI;

        struct
        {
            int time;
            int init;
        } SSR;

    } Uconf;
};


/* Declare instrument info structure */
struct Instr
{
    int  isConn;
    Type type;
    int  addr;
    int  chns;
    Chn* chn;
};


/* Declare port structure */
struct Port
{
    Port*         pport;

    char*         name;
    int           isConn;
    Serport*      pserial;
    asynUser*     pasynUser;
    asynInterface asynInt32;
    asynInterface asynUInt32;
    asynInterface asynCommon;
    asynInterface asynDrvUser;
    asynInterface asynFloat64;
    asynInterface asynLockPort;

    Instr instr[INSTRMAX];
};


/* Declare serial port structure */
struct Serport
{
    char*       name;
    int         addr;
    int         isConn;
    int         canBlock;
    int         multiDevice;
    int         autoConnect;
    asynUser*   pasynUser;
    asynOctet*  pasynOctet;
    void*       pasynOctetPvt;
};


/* Defaine local variants */
static Port* pports;
static char tciChns[] =
    {'A','B','C','D'};
static char ssrChns[] =
    {'A','B','C','D','E'};
static char dioChns[] =
    {'A','B','C','D','E','F','G','H','I','J','K','L','M','N'};
static char instrAddrs[] =
    {'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P',
     'a','b','c','d','e','f','g','h','i','j','k','l','m','n','o','p'};

/* Public interface forward references */
static int drvWeedConfig(const char* port,int addr,int type,int chns);
static int drvWeedInit(const char* port,const char* serport,int seraddr);

static asynStatus processRead(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn,void* value);
static asynStatus processWrite(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn,void* value);
static asynStatus sendCommand(asynUser* pasynUser,Serport* pserial,char* send,char* reply,size_t max);

static void setdioInp(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn);
static Chn* createdio(asynUser* pasynUser,Port* pport,Instr* pinstr,const char* drvInfo);
static asynStatus processdioRead(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn,epicsUInt32* value);
static asynStatus processdioWrite(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn,epicsUInt32* value);

static void settciType(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn);
static void settciUnit(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn);
static Chn* createtci(asynUser* pasynUser,Port* pport,Instr* pinstr,const char* drvInfo);
static asynStatus processtciRead(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn,epicsFloat64* value);
static asynStatus processtciWrite(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn,epicsFloat64* value);

static void setssrInit(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn);
static Chn* createssr(asynUser* pasynUser,Port* pport,Instr* pinstr,const char* drvInfo);
static asynStatus processssrRead(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn,epicsUInt32* value);
static asynStatus processssrWrite(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn,epicsUInt32* value);

static asynStatus setEos(Port* pport);
static asynStatus initSerialPort(Port* port,const char* serport,int addr);


/* Forward references for asynCommon methods */
static void report(void* ppvt,FILE* fp,int details);
static asynStatus connect(void* ppvt,asynUser* pasynUser);
static asynStatus disconnect(void* ppvt,asynUser* pasynUser);
static asynCommon common = {report,connect,disconnect};


/* Forward references for asynDrvUser methods */
static asynStatus create(void* ppvt,asynUser* pasynUser,const char* drvInfo, const char** pptypeName,size_t* psize);
static asynStatus destroy(void* ppvt,asynUser* pasynUser);
static asynStatus gettype(void* ppvt,asynUser* pasynUser,const char** pptypeName,size_t* psize);
static asynDrvUser drvuser = {create,gettype,destroy};


/* asynLockPortNotify methods */
static asynStatus lockPort(void *drvPvt,asynUser *pasynUser);
static asynStatus unlockPort(void *drvPvt,asynUser *pasynUser);
static asynLockPortNotify lockport = {lockPort,unlockPort};


/* Forward references for asynInt32Digital methods */
static asynStatus readInt32(void* ppvt,asynUser* pasynUser,epicsInt32* value);
static asynStatus writeInt32(void* ppvt,asynUser* pasynUser,epicsInt32 value);


/* Forward references for asynUInt32Digital methods */
static asynStatus readUInt32(void* ppvt,asynUser* pasynUser,epicsUInt32* value,epicsUInt32 mask);
static asynStatus writeUInt32(void* ppvt,asynUser* pasynUser,epicsUInt32 value,epicsUInt32 mask);


/* Forward references for asynFloat64 methods */
static asynStatus readFloat64(void* ppvt,asynUser* pasynUser,epicsFloat64* value);
static asynStatus writeFloat64(void* ppvt,asynUser* pasynUser,epicsFloat64 value);


/* Define macros */
#define ISOK(s) (asynSuccess==(s))
#define ISNOTOK(s) (!ISOK(s))


/****************************************************************************
 * Define public interface methods
 ****************************************************************************/
static int drvWeedInit(const char* port,const char* serport,int seraddr)
{
    int len,attr;
    Port* pport;
    Serport* pserial;
    asynUser* pasynUser;
    asynInt32* pasynInt32;
    asynFloat64* pasynFloat64;
    asynUInt32Digital* pasynUInt32;

    len = strlen(port)+strlen(serport)+2;
    len += sizeof(Port)+sizeof(Serport)+sizeof(asynInt32)+sizeof(asynFloat64)+sizeof(asynUInt32Digital);
    pport = (Port*)callocMustSucceed(len,sizeof(char),"drvWeedInit");

    pserial = (Serport*)(pport + 1);
    pasynInt32 = (asynInt32*)(pserial + 1);
    pasynUInt32 = (asynUInt32Digital*)(pasynInt32 + 1);
    pasynFloat64 = (asynFloat64*)(pasynUInt32 + 1);

    pport->name = (char*)(pasynFloat64 + 1);
    pserial->name = (pport->name + strlen(port) + 1);

    pport->isConn = FALSE;
    pport->pserial = pserial;
    strcpy(pport->name,port);

    if( initSerialPort(pport,serport,-1) )
    {
        printf("drvWeedInit::failure to initialize serial port %s\n",serport);
        free(pport);
        return( -1 );
    }

    attr = (pserial->canBlock)?(ASYN_MULTIDEVICE|ASYN_CANBLOCK):ASYN_MULTIDEVICE;
    if( pasynManager->registerPort(port,attr,pserial->autoConnect,0,0) )
    {
        printf("drvWeedInit::init %s: failure to register port\n",port);
        free(pport);
        return( -1 );
    }

    pport->asynCommon.interfaceType = asynCommonType;
    pport->asynCommon.pinterface = &common;
    pport->asynCommon.drvPvt = pport;

    if( pasynManager->registerInterface(port,&pport->asynCommon) )
    {
        printf("drvWeedInit::init %s: failure to register asynCommon\n",port);
        return( -1 );
    }

    pport->asynDrvUser.interfaceType = asynDrvUserType;
    pport->asynDrvUser.pinterface = &drvuser;
    pport->asynDrvUser.drvPvt = pport;

    if( pasynManager->registerInterface(port,&pport->asynDrvUser) )
    {
        printf("drvWeedInit::init %s: failure to register asynDrvUser\n",port);
        return( -1 );
    }

    pport->asynLockPort.interfaceType = asynLockPortNotifyType;
    pport->asynLockPort.pinterface = &lockport;
    pport->asynLockPort.drvPvt = pport;

    if( pasynManager->registerInterface(port,&pport->asynLockPort) )
    {
        printf("drvWeedInit::failure to register asynLockPort\n");
        return( -1 );
    }

    pasynInt32->read = readInt32;
    pasynInt32->write = writeInt32;
    pport->asynInt32.interfaceType = asynInt32Type;
    pport->asynInt32.pinterface = pasynInt32;
    pport->asynInt32.drvPvt = pport;

    if( pasynInt32Base->initialize(port,&pport->asynInt32) )
    {
        printf("drvWeedInit::init %s: failure to initialize asynInt32Base\n",port);
        return( -1 );
    }

    pasynUInt32->read = readUInt32;
    pasynUInt32->write = writeUInt32;
    pport->asynUInt32.interfaceType = asynUInt32DigitalType;
    pport->asynUInt32.pinterface = pasynUInt32;
    pport->asynUInt32.drvPvt = pport;

    if( pasynUInt32DigitalBase->initialize(port,&pport->asynUInt32) )
    {
        printf("drvDioInit::init %s: failure to initialize asynUInt32DigitalBase\n",port);
        return( -1 );
    }

    pasynFloat64->read = readFloat64;
    pasynFloat64->write = writeFloat64;
    pport->asynFloat64.interfaceType = asynFloat64Type;
    pport->asynFloat64.pinterface = pasynFloat64;
    pport->asynFloat64.drvPvt = pport;

    if( pasynFloat64Base->initialize(port,&pport->asynFloat64) )
    {
        printf("drvSenInit::init %s: failure to initialize asynFloat64Base\n",port);
        return( -1 );
    }

    if( (pasynUser = pasynManager->createAsynUser(NULL,NULL)) )
    {
        pport->pasynUser = pasynUser;
        pasynUser->userPvt = pport;
        pasynUser->timeout = COMTMO;
    }
    else
    {
        printf("drvWeedInit::create asynUser failure\n");
        return( -1 );
    }

    if( pasynManager->connectDevice(pasynUser,port,-1) )
    {
        printf("drvWeedInit::failure to connect with device %s\n",port);
        pport->isConn = 0;
        return( -1 );
    }

    if( setEos(pport) )
    {
        printf("drvWeedInit::failure to set %s EOS\n",port);
        return( -1 );
    }

    if( pports )
        pport->pport = pports;
    pports = pport;

    return( 0 );
}


static int drvWeedConfig(const char* port,int addr,int type,int chns)
{
    int i,len;
    Port* pport;
    Instr* pinstr;

    for( pport = pports; pport; pport = pport->pport )
        if( epicsStrCaseCmp(pport->name,port) == 0 )
            break;

    if( pport == NULL )
    {
        printf("drvWeedConfig::failure to locate port %s\n",port);
        return( -1 );
    }

    if( addr >= 0 && addr < INSTRMAX )
    {
        pinstr = &pport->instr[addr];
    }
    else
    {
        printf("drvWeedConfig::invalid address %d for port %s\n",addr,port);
        return( - 1 );
    }

    if( type == WED_DIO )
        pinstr->type = DIO;
    else if( type == WED_TCI )
        pinstr->type = TCI;
    else if( type == WED_SSR )
        pinstr->type = SSR;
    else if( type == WED_SMD )
        pinstr->type = SMD;
    else
    {
        printf("drvWeedConfig::unknown board type %d for port %s\n",type,port);
        return( - 1 );
    }

    len = sizeof(Chn) * chns;
    pinstr->chn = (Chn*)callocMustSucceed(len,sizeof(char),"drvWeedConfig");

    pinstr->isConn = FALSE;
    pinstr->chns = chns;
    pinstr->addr = addr;

    for( i = 0; i < pinstr->chns; ++i )
    {
        pinstr->chn[i].chn = i;
        pinstr->chn[i].isConn = FALSE;
    }

    return( 0 );
}


static asynStatus initSerialPort(Port* port,const char* serport,int addr)
{
    asynUser* pasynUser;
    asynInterface* pasynIface;
    Serport* pserial = port->pserial;

    pasynUser = pasynManager->createAsynUser(NULL,NULL);
    if( pasynUser == NULL )
        return( asynError );

    if( pasynManager->connectDevice(pasynUser,serport,addr) )
    {
        printf("initSerialPort::failure to connect with device %s @ %d\n",serport,addr);
        pasynManager->freeAsynUser(pasynUser);
        return( asynError );
    }

    if( pasynManager->isMultiDevice(pasynUser,serport,&pserial->multiDevice) )
    {
        printf("initSerialPort::failure to determine if %s @ %d is multi device\n",serport,addr);
        pasynManager->disconnect(pasynUser);
        pasynManager->freeAsynUser(pasynUser);
        return( asynError );
    }

    if( pserial->multiDevice )
    {
        printf("initSerialPort::%s cannot be configured as multi device\n",serport);
        pasynManager->disconnect(pasynUser);
        pasynManager->freeAsynUser(pasynUser);
        return( asynError );
    }

    if( pasynManager->canBlock(pasynUser,&pserial->canBlock) )
    {
        printf("initSerialPort::failure to determine if %s @ %d can block\n",serport,addr);
        pasynManager->disconnect(pasynUser);
        pasynManager->freeAsynUser(pasynUser);
        return( asynError );
    }

    if( pasynManager->isAutoConnect(pasynUser,&pserial->autoConnect) )
    {
        printf("initSerialPort::failure to determine if %s @ %d is autoconnect\n",serport,addr);
        pasynManager->disconnect(pasynUser);
        pasynManager->freeAsynUser(pasynUser);
        return( asynError );
    }

    pasynIface = pasynManager->findInterface(pasynUser,asynOctetType,1);
    if( pasynIface )
    {
        pserial->pasynOctet = (asynOctet*)pasynIface->pinterface;
        pserial->pasynOctetPvt = pasynIface->drvPvt;
    }
    else
    {
        printf("initSerialPort::failure to find interface %s\n",asynOctetType);
        pasynManager->disconnect(pasynUser);
        pasynManager->freeAsynUser(pasynUser);
        return( asynError );
    }

    pserial->addr = addr;
    pserial->isConn = TRUE;
    pserial->pasynUser = pasynUser;
    strcpy(pserial->name,serport);

    pasynUser->userPvt = port;
    pasynUser->timeout = COMTMO;

    return( asynSuccess );
}


static asynStatus setEos(Port* pport)
{
    asynStatus sts;
    char eos = '\r';
    Serport* pserial = pport->pserial;

    sts = pserial->pasynOctet->setInputEos(pserial->pasynOctetPvt,pserial->pasynUser,&eos,1);
    if( ISNOTOK(sts) )
        printf("drvWeedInit::setEos Input EOS set failed to \\0%d\n",eos);

    sts = pserial->pasynOctet->setOutputEos(pserial->pasynOctetPvt,pserial->pasynUser,&eos,1);
    if( ISNOTOK(sts) )
        printf("drvWeedInit::setEos Output EOS set failed to \\0%d\n",eos);

    return( sts );
}


static asynStatus sendCommand(asynUser* pasynUser,Serport* pserial,char* send,char* reply,size_t max)
{
    int eom;
    asynStatus sts;
    size_t len,bytes;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::sendCommand\n");

    len = strlen(send);
    pasynUser->timeout = COMTMO;
    sts = pserial->pasynOctet->write(pserial->pasynOctetPvt,pserial->pasynUser,send,len,&bytes);
    if( ISNOTOK(sts) )
    {
        if( sts == asynTimeout )
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::sendCommand - send asynTimeout\n");
        else if( sts == asynOverflow )
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::sendCommand - send asynOverflow\n");
        else if( sts == asynError )
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::sendCommand - send asynError\n");
        else
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::sendCommand - send unknown error\n");

        return( sts );
    }

    sts = pserial->pasynOctet->read(pserial->pasynOctetPvt,pserial->pasynUser,reply,max,&bytes,&eom);
    if( ISOK(sts) )
    {
        if( eom != ASYN_EOM_EOS )
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::sendCommand - reply eos (0x%2.2X) error\n",eom);
            return( asynError );
        }
    }
    else
    {
        if( sts == asynTimeout )
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::sendCommand - reply asynTimeout\n");
        else if( sts == asynOverflow )
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::sendCommand - reply asynOverflow\n");
        else if( sts == asynError )
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::sendCommand - reply asynError\n");
        else
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::sendCommand - reply unknown error\n");

        return( sts );
    }
    epicsThreadSleep(0.001);

    return( asynSuccess );
}


static asynStatus processRead(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn,void* value)
{
    asynStatus sts;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::processRead\n");

    if( pinstr->type == DIO )
        sts = processdioRead(pasynUser,pport,pinstr,pchn,(epicsUInt32*)value);
    else if( pinstr->type == TCI )
        sts = processtciRead(pasynUser,pport,pinstr,pchn,(epicsFloat64*)value);
    else if( pinstr->type == SSR )
        sts = processssrRead(pasynUser,pport,pinstr,pchn,(epicsUInt32*)value);
    else if( pinstr->type == SMD )
        sts = asynError;
    else
        sts = asynError;

    return( sts );
}


static asynStatus processWrite(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn,void* value)
{
    asynStatus sts;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::processWrite\n");

    if( pinstr->type == DIO )
        sts = processdioWrite(pasynUser,pport,pinstr,pchn,(epicsUInt32*)value);
    else if( pinstr->type == TCI )
        sts = processtciWrite(pasynUser,pport,pinstr,pchn,(epicsFloat64*)value);
    else if( pinstr->type == SSR )
        sts = processssrWrite(pasynUser,pport,pinstr,pchn,(epicsUInt32*)value);
    else if( pinstr->type == SMD )
        sts = asynError;
    else
        sts = asynError;

    return( sts );
}


/****************************************************************************
 * Define private interface WTDIO methods
 ****************************************************************************/
static void setdioInp(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn)
{
    asynStatus sts;
    char send[12],reply[12];

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::setdioInp\n");

    send[0] = instrAddrs[pinstr->addr];
    send[1] = 'I';
    send[2] = dioChns[pchn->chn];
    send[3] = '\0';

    sts = sendCommand(pasynUser,pport->pserial,send,reply,sizeof(reply));
    if( ISOK(sts) && epicsStrCaseCmp(send,reply) )
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::setdioInp - invalid data\n");
}


static Chn* createdio(asynUser* pasynUser,Port* pport,Instr* pinstr,const char* drvInfo)
{
    char dir[12];
    int chn,time;

    sscanf(drvInfo,"%s %d %d",dir,&chn,&time);
    if( chn < 0 || chn > 13 )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::createdio %s: invalid channel number %c\n",pport->name,chn);
        return( NULL );
    }

    if( epicsStrCaseCmp("Input",dir) == 0 )
    {
        pinstr->chn[chn].Uconf.DIO.dir = INP;
        pinstr->chn[chn].Uconf.DIO.time = 0;
        setdioInp(pasynUser,pport,pinstr,&pinstr->chn[chn]);
    }
    else if( epicsStrCaseCmp("Output",dir) == 0 )
    {
        pinstr->chn[chn].Uconf.DIO.dir = OUT;
        pinstr->chn[chn].Uconf.DIO.time = time;
    }
    else
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::createdio %s: invalid direction %s\n",pport->name,dir);
        return( NULL );
    }
    pinstr->chn[chn].isConn = TRUE;

    return( &pinstr->chn[chn] );
}


static asynStatus processdioRead(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn,epicsUInt32* value)
{
    asynStatus sts;
    char send[12],reply[12];

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::processdioRead\n");

    send[0] = instrAddrs[pinstr->addr];
    send[1] = 'R';
    send[2] = dioChns[pchn->chn];
    send[3] = '\0';

    sts = sendCommand(pasynUser,pport->pserial,send,reply,sizeof(reply));
    if( ISOK(sts) )
    {
        if( send[0] != reply[0] )
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::processdioRead - addr failure\n");
            return( asynError );
        }

        if( reply[1] == '?' )
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::processdioRead - invalid command\n");
            return( asynError );
        }

        if( send[2] != reply[1] )
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::processdioRead - chan failure\n");
            return( asynError );
        }

        if( reply[2] == 'H' )
            *value = 1;
        else if( reply[2] == 'L' )
            *value = 0;
        else
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::processdioRead - invalid data\n");
            return( asynError );
        }

    }

    return( sts );
}


static asynStatus processdioWrite(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn,epicsUInt32* value)
{
    asynStatus sts;
    char send[12],reply[12];

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::processdioWrite\n");

    send[0] = instrAddrs[pinstr->addr];
    send[1] = (*value) ? 'H' : 'L';
    send[2] = dioChns[pchn->chn];
    if( pchn->Uconf.DIO.time )
        sprintf(&send[3],"%d",pchn->Uconf.DIO.time);
    else
        send[3] = '\0';

    sts = sendCommand(pasynUser,pport->pserial,send,reply,sizeof(reply));
    if( ISOK(sts) )
    {
        if( epicsStrCaseCmp(send,reply) )
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::processdioWrite - invalid data\n");
            return( asynError );
        }
    }

    return( sts );
}


/****************************************************************************
 * Define private interface WTTCI methods
 ****************************************************************************/
static void settciType(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn)
{
    asynStatus sts;
    char send[12],reply[12];

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::settciType\n");

    send[0] = instrAddrs[pinstr->addr];
    send[1] = 'T';
    send[2] = tciChns[pchn->chn];
    send[3] = pchn->Uconf.TCI.type;
    send[4] = '\0';

    sts = sendCommand(pasynUser,pport->pserial,send,reply,sizeof(reply));
    if( ISOK(sts) && epicsStrCaseCmp(send,reply) )
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::settciType - invalid data\n");
}


static void settciUnit(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn)
{
    asynStatus sts;
    char send[12],reply[12];

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::settciUnits\n");

    send[0] = instrAddrs[pinstr->addr];
    send[1] = 'U';
    send[2] = tciChns[pchn->chn];
    send[3] = pchn->Uconf.TCI.unit;
    send[4] = '\0';

    sts = sendCommand(pasynUser,pport->pserial,send,reply,sizeof(reply));
    if( ISOK(sts) && epicsStrCaseCmp(send,reply) )
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::settciUnit - invalid data\n");
}


static Chn* createtci(asynUser* pasynUser,Port* pport,Instr* pinstr,const char* drvInfo)
{
    int chn;
    char dir[12],type,unit;

    sscanf(drvInfo,"%s %d %c %c",dir,&chn,&type,&unit);

    if( epicsStrCaseCmp("Input",dir) )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::createtci %s: invalid direction %s\n",pport->name,dir);
        return( NULL );
    }

    if( chn < 0 || chn > 3 )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::createtci %s: invalid channel number %c\n",pport->name,chn);
        return( NULL );
    }

    if( type != 'J' && type != 'K' && type != 'T' && type != 'E' )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::createtci %s: invalid thermocouple type %c\n",pport->name,type);
        return( NULL );
    }

    if( unit != 'F' && unit != 'C' )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::createtci %s: invalid units %c\n",pport->name,unit);
        return( NULL );
    }

    if( pinstr->chn[chn].isConn == FALSE )
    {
        pinstr->chn[chn].isConn = TRUE;
        pinstr->chn[chn].Uconf.TCI.type = type;
        pinstr->chn[chn].Uconf.TCI.unit = unit;

        settciUnit(pasynUser,pport,pinstr,&pinstr->chn[chn]);
        settciType(pasynUser,pport,pinstr,&pinstr->chn[chn]);
    }

    return( &pinstr->chn[chn] );
}


static asynStatus processtciRead(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn,epicsFloat64* value)
{
    asynStatus sts;
    char send[12],reply[12];

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::processtciRead\n");

    send[0] = instrAddrs[pinstr->addr];
    send[1] = 'R';
    send[2] = tciChns[pchn->chn];
    send[3] = '\0';

    sts = sendCommand(pasynUser,pport->pserial,send,reply,sizeof(reply));
    if( ISOK(sts) )
    {
        if( send[0] != reply[0] )
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::processtciRead - addr failure\n");
            return( asynError );
        }

        if( reply[1] == '?' )
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::processtciRead - invalid command\n");
            return( asynError );
        }

        sscanf(&reply[1],"%lf",value);
    }

    return( sts );
}


static asynStatus processtciWrite(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn,epicsFloat64* value)
{
    return( asynError );
}


/****************************************************************************
 * Define private interface WTSSR methods
 ****************************************************************************/
static void setssrInit(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn)
{
    asynStatus sts;
    char send[12],reply[12];

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::setssrInit\n");

    send[0] = instrAddrs[pinstr->addr];
    send[1] = 'D';
    send[2] = ssrChns[pchn->chn];
    send[3] = '\0';

    sts = sendCommand(pasynUser,pport->pserial,send,reply,sizeof(reply));
    if( ISOK(sts) && epicsStrCaseCmp(send,reply) )
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::setssrIinit - invalid data\n");
}


static Chn* createssr(asynUser* pasynUser,Port* pport,Instr* pinstr,const char* drvInfo)
{
    char dir[12];
    int chn,time;

    sscanf(drvInfo,"%s %d %d",dir,&chn,&time);

    if( chn < 0 || chn > 4 )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::createssr %s: invalid channel number %c\n",pport->name,chn);
        return( NULL );
    }

    if( epicsStrCaseCmp("Output",dir) )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::createssr %s: invalid direction %s\n",pport->name,dir);
        return( NULL );
    }

    pinstr->chn[chn].Uconf.SSR.init = 0;
    pinstr->chn[chn].Uconf.SSR.time = time;
    pinstr->chn[chn].isConn = TRUE;

    return( &pinstr->chn[chn] );
}


static asynStatus processssrRead(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn,epicsUInt32* value)
{
    asynStatus sts;
    char send[12],reply[12];

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::processssrRead\n");

    send[0] = instrAddrs[pinstr->addr];
    send[1] = 'R';
    send[2] = ssrChns[pchn->chn];
    send[3] = '\0';

    sts = sendCommand(pasynUser,pport->pserial,send,reply,sizeof(reply));
    if( ISOK(sts) )
    {
        if( send[0] != reply[0] )
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::processssrRead - addr failure\n");
            return( asynError );
        }

        if( reply[1] == '?' )
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::processssrRead - invalid command\n");
            return( asynError );
        }

        if( send[2] != reply[1] )
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::processssrRead - chan failure\n");
            return( asynError );
        }

        if( reply[2] == 'C' )
            *value = 1;
        else if( reply[2] == 'O' )
            *value = 0;
        else
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::processssrRead - invalid data\n");
            return( asynError );
        }

    }

    return( sts );
}


static asynStatus processssrWrite(asynUser* pasynUser,Port* pport,Instr* pinstr,Chn* pchn,epicsUInt32* value)
{
    asynStatus sts;
    char send[12],reply[12];

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::processssrWrite\n");

    send[0] = instrAddrs[pinstr->addr];
    send[1] = (*value) ? 'C' : 'O';
    send[2] = ssrChns[pchn->chn];
    if( pchn->Uconf.DIO.time )
        sprintf(&send[3],"%d",pchn->Uconf.SSR.time);
    else
        send[3] = '\0';

    sts = sendCommand(pasynUser,pport->pserial,send,reply,sizeof(reply));
    if( ISOK(sts) )
    {
        if( epicsStrCaseCmp(send,reply) )
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::processssrWrite - invalid data\n");
            return( asynError );
        }
    }

    return( sts );
}


/****************************************************************************
 * Define private interface asynCommon methods
 ****************************************************************************/
static void report(void* ppvt,FILE* fp,int details)
{
    int i;
    Port* pport = (Port*)ppvt;

    fprintf(fp,"    EPICS Brick Weeder Techologies\n");
    fprintf(fp,"        Port %s is connected to %s\n",pport->name,pport->pserial->name);
    for( i = 0; i < INSTRMAX; ++i )
        if( pport->instr[i].isConn )
            fprintf(fp,"        Addr %d connected, type %d, %d channels\n",i,pport->instr[i].type,pport->instr[i].chns);

}


static asynStatus connect(void* ppvt,asynUser* pasynUser)
{
    int addr;
    asynStatus sts;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::connect %s:\n",pport->name);

    if( (sts = pasynManager->getAddr(pasynUser,&addr)) )
        return( sts );

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvWeed::connect %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    if( addr >= INSTRMAX  )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::connect %s: illegal addr %d\n",pport->name,addr);
        return( asynError );
    }

    if( addr >= 0 )
    {
        if( pport->instr[addr].isConn == TRUE )
	{
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::connect %s: already connected to %d\n",pport->name,addr);
            return( asynError );
	}
    
        pport->instr[addr].isConn = TRUE;
        pasynManager->exceptionConnect(pasynUser);
        return( asynSuccess );
    }

    if( pport->isConn )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::connect %s: already connected to %d\n",pport->name,addr);
        return( asynError );
    }

    pport->isConn = TRUE;
    pasynManager->exceptionConnect(pasynUser);
    return( asynSuccess );
}


static asynStatus disconnect(void* ppvt,asynUser* pasynUser)
{
    int addr;
    asynStatus sts;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::disconnect %s:\n",pport->name);

    if( (sts = pasynManager->getAddr(pasynUser,&addr)) )
        return( sts );

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvWeed::disconnect %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    if( addr >= INSTRMAX )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::disconnect %s: illegal addr %d\n",pport->name,addr);
        return( asynError );
    }

    if( addr >= 0 )
    {
        if( pport->instr[addr].isConn == FALSE )
        {
            asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::disconnect %s: not connected to %d\n",pport->name,addr);
            return( asynError );
        }

        pport->instr[addr].isConn = FALSE;
        pasynManager->exceptionDisconnect(pasynUser);
        return( asynSuccess );
    }

    if( pport->isConn == FALSE )
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::disconnect %s: not connected to %d\n",pport->name,addr);
        return( asynError );
    }

    pport->isConn = FALSE;
    pasynManager->exceptionDisconnect(pasynUser);
    return( asynSuccess );
}


/****************************************************************************
 * Define private interface asynDrvUser methods
 ****************************************************************************/
static asynStatus create(void* ppvt,asynUser* pasynUser,const char* drvInfo, const char** pptypeName,size_t* psize)
{
    int addr;
    Chn* pchn;
    Instr* pinstr;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::create %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        pinstr = &pport->instr[addr];
    else
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::create %s: getAddr failure\n",pport->name);
        return( asynError );
    }

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvWeed::create %s: asyn - 0x%8.8X, addr - %d\n",pport->name,(int)pasynUser,(int)addr);

    if( pinstr->type == DIO )
        pchn = createdio(pasynUser,pport,pinstr,drvInfo);
    else if( pinstr->type == TCI )
        pchn = createtci(pasynUser,pport,pinstr,drvInfo);
    else if( pinstr->type == SSR )
        pchn = createssr(pasynUser,pport,pinstr,drvInfo);
    else if( pinstr->type == SMD )
        pchn = NULL;
    else
        pchn = NULL;

    if( pchn )
    {
        pasynUser->drvUser = pchn;
        pasynUser->timeout = COMTMO;
        return( asynSuccess );
    }
    else
        return( asynError );

}


static asynStatus gettype(void* ppvt,asynUser* pasynUser,const char** pptypeName,size_t* psize)
{
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::gettype %s:\n",pport->name);

    if( pptypeName )
        *pptypeName = NULL;
    if( psize )
        *psize = 0;

    return( asynSuccess );
}


static asynStatus destroy(void* ppvt,asynUser* pasynUser)
{
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::destroy %s:\n",pport->name);

    if( pport )
    {
        pasynUser->drvUser = NULL;
        return( asynSuccess );
    }
    else
    {
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::destroy %s: called before create\n",pport->name);
        return( asynError );
    }

}


/****************************************************************************
 * Define private interface asynLockPort methods
 ****************************************************************************/
static asynStatus lockPort(void* drvPvt,asynUser* pasynUser)
{
    asynStatus sts;
    Port* pport = (Port*)drvPvt;
    Serport* pserial = pport->pserial;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::lockPort\n");

    sts = pasynManager->lockPort(pserial->pasynUser);
    if( ISNOTOK(sts) )
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::lockPort %s error %s\n",pserial->name,pserial->pasynUser->errorMessage);

    return( sts );
}


static asynStatus unlockPort(void* drvPvt,asynUser* pasynUser)
{
    asynStatus sts;
    Port* pport = (Port*)drvPvt;
    Serport* pserial = pport->pserial;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::unlockPort\n");

    sts = pasynManager->unlockPort(pserial->pasynUser);
    if( ISNOTOK(sts) )
        asynPrint(pasynUser,ASYN_TRACE_ERROR,"drvWeed::unlockPort %s error %s\n",pserial->name,pserial->pasynUser->errorMessage);

    return( asynSuccess );
}


/****************************************************************************
 * Define private interface asynInt32 methods
 ****************************************************************************/
static asynStatus writeInt32(void* ppvt,asynUser* pasynUser,epicsInt32 value)
{
    int addr;
    Instr* pinstr;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::writeUInt32 %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        pinstr = &pport->instr[addr];
    else
        return( asynError );

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvWeed::writeUInt32 %s: asyn - 0x%8.8X, addr - %d, value - 0x%8.8X\n",pport->name,(int)pasynUser,(int)pinstr->addr,(int)value);
    // Send command

    return( asynSuccess );
}


static asynStatus readInt32(void* ppvt,asynUser* pasynUser,epicsInt32* value)
{
    int addr;
    Instr* pinstr;
    Port* pport = (Port*)ppvt;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::readUInt32 %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        pinstr = &pport->instr[addr];
    else
        return( asynError );

    *value = 0;
    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvWeed::readUInt32 %s: asyn - 0x%8.8X, addr - %d, value - 0x%8.8X\n",pport->name,(int)pasynUser,(int)pinstr->addr,(int)*value);

    return( asynSuccess );
}


/****************************************************************************
 * Define private interface asynUInt32Digital methods
 ****************************************************************************/
static asynStatus writeUInt32(void* ppvt,asynUser* pasynUser,epicsUInt32 value,epicsUInt32 mask)
{
    int addr;
    Instr* pinstr;
    Port* pport = (Port*)ppvt;
    Chn* pchn = (Chn*)pasynUser->drvUser;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::writeUInt32 %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        pinstr = &pport->instr[addr];
    else
        return( asynError );

    value &= mask;
    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvWeed::writeUInt32 %s: asyn - 0x%8.8X, addr - %d, chan - %d, mask - 0x%8.8X, value - 0x%8.8X\n",pport->name,(int)pasynUser,addr,(int)pchn->chn,(int)mask,(int)value);

    return( processWrite(pasynUser,pport,pinstr,pchn,&value) );
}


static asynStatus readUInt32(void* ppvt,asynUser* pasynUser,epicsUInt32* value,epicsUInt32 mask)
{
    int addr;
    asynStatus sts;
    Instr* pinstr;
    Port* pport = (Port*)ppvt;
    Chn* pchn = (Chn*)pasynUser->drvUser;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::readUInt32 %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        pinstr = &pport->instr[addr];
    else
        return( asynError );

    sts = processRead(pasynUser,pport,pinstr,pchn,value);
    *value &= mask;

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvWeed::readUInt32 %s: asyn - 0x%8.8X, addr - %d, chan - %d, mask - 0x%8.8X, value - 0x%8.8X\n",pport->name,(int)pasynUser,addr,(int)pchn->chn,(int)mask,(int)*value);

    return( sts );
}


/****************************************************************************
 * Define private interface asynFloat64 methods
 ****************************************************************************/
static asynStatus writeFloat64(void* ppvt,asynUser* pasynUser,epicsFloat64 value)
{
    int addr;
    asynStatus sts;
    Instr* pinstr;
    Port* pport = (Port*)ppvt;
    Chn* pchn = (Chn*)pasynUser->drvUser;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::writeFloat64 %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        pinstr = &pport->instr[addr];
    else
        return( asynError );

    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvWeed::writeFloat64 %s: asyn - 0x%8.8X, addr - %d, chan - %d, value - %f\n",pport->name,(int)pasynUser,addr,(int)pchn->chn,(int)value);
    sts = processWrite(pasynUser,pport,pinstr,pchn,&value);

    return( sts );
}


static asynStatus readFloat64(void* ppvt,asynUser* pasynUser,epicsFloat64* value)
{
    int addr;
    asynStatus sts;
    Instr* pinstr;
    Port* pport = (Port*)ppvt;
    Chn* pchn = (Chn*)pasynUser->drvUser;

    asynPrint(pasynUser,ASYN_TRACE_FLOW,"drvWeed::readFloat64 %s:\n",pport->name);

    if( pasynManager->getAddr(pasynUser,&addr) == asynSuccess )
        pinstr = &pport->instr[addr];
    else
        return( asynError );

    sts = processRead(pasynUser,pport,pinstr,pchn,value);
    asynPrint(pasynUser,ASYN_TRACEIO_FILTER,"drvWeed::readFloat64 %s: asyn - 0x%8.8X, addr - %d, chan - %d, value - %f\n",pport->name,(int)pasynUser,addr,(int)pchn->chn,(int)*value);

    return( sts );
}


/****************************************************************************
 * Register public methods
 ****************************************************************************/

/* Initialization method definitions */
static const iocshArg drvWeedInitArg0 = {"port",iocshArgString};
static const iocshArg drvWeedInitArg1 = {"serport",iocshArgString};
static const iocshArg drvWeedInitArg2 = {"seraddr",iocshArgInt};
static const iocshArg* drvWeedInitArgs[]= {&drvWeedInitArg0,&drvWeedInitArg1,&drvWeedInitArg2};
static const iocshFuncDef drvWeedInitFuncDef = {"drvWeedInit",3,drvWeedInitArgs};
static void drvWeedInitCallFunc(const iocshArgBuf* args)
{
    drvWeedInit(args[0].sval,args[1].sval,args[2].ival);
}

/* Initialization method definitions */
static const iocshArg drvWeedCfgArg0 = {"port",iocshArgString};
static const iocshArg drvWeedCfgArg1 = {"addr",iocshArgInt};
static const iocshArg drvWeedCfgArg2 = {"type",iocshArgInt};
static const iocshArg drvWeedCfgArg3 = {"chns",iocshArgInt};
static const iocshArg* drvWeedCfgArgs[]= {&drvWeedCfgArg0,&drvWeedCfgArg1,&drvWeedCfgArg2,&drvWeedCfgArg3};
static const iocshFuncDef drvWeedCfgFuncDef = {"drvWeedConfig",4,drvWeedCfgArgs};
static void drvWeedCfgCallFunc(const iocshArgBuf* args)
{
    drvWeedConfig(args[0].sval,args[1].ival,args[2].ival,args[3].ival);
}

/* Registration method */
static void drvWeedRegister(void)
{
    static int firstTime = 1;

    if( firstTime )
    {
        firstTime = 0;
        iocshRegister( &drvWeedCfgFuncDef, drvWeedCfgCallFunc );
        iocshRegister( &drvWeedInitFuncDef, drvWeedInitCallFunc );
    }
}
epicsExportRegistrar( drvWeedRegister );
