/*

                          Argonne National Laboratory
                            APS Operations Division
                     Beamline Controls and Data Acquisition

                    Support for Diamond Systems Incorporated
                         Athena Single-Board Computers



 -----------------------------------------------------------------------------
                                COPYRIGHT NOTICE
 -----------------------------------------------------------------------------
   Copyright (c) 2002 The University of Chicago, as Operator of Argonne
      National Laboratory.
   Copyright (c) 2002 The Regents of the University of California, as
      Operator of Los Alamos National Laboratory.
   Synapps Versions 4-5
   and higher are distributed subject to a Software License Agreement found
   in file LICENSE that is included with this distribution.
 -----------------------------------------------------------------------------

 Description
    This module provides support for the Athena single board computers provided
    by Diamond Systems Incorporated.

 Developer notes:

 Source control info:
    Modified by:    $Author: dkline $
                    $Date: 2006/07/31 14:57:37 $
                    $Revision: 1.5 $

 =============================================================================
 History:
 Author: David M. Kline
 -----------------------------------------------------------------------------
 2005-Nov-11  DMK  Development started.
 2005-Nov-15  DMK  Initial development complete.
 2005-Nov-17  DMK  Including functions for WDT.
 2006-Feb-28  DMK  Added support for RMM416 and made methods more generic.
 2006-May-15  DMK  Removed dscInitBoard().
 2006-May-20  DMK  Added support in dioSetDir() for no configuration register.
 2006-Jul-31  DMK  Added support for IR-104.
 -----------------------------------------------------------------------------

*/

/* System related include files */
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>


/* DSC related include files */
#include "dscsup.h"


/* Define local variants */
static unsigned char athena_wdt[4];


/****************************************************************************
 * Define DIO public interface methods
 ****************************************************************************/
void dioSetDir(unsigned short base,char ro,char co,unsigned char* cd,char dir)
{
unsigned char dat,mask=0;

    if( co < 0 )
        return;

    switch( ro )
    {
    case ATH_OUTB_DPORT_A:
    case RMM_OUTB_DPORT_A:
    case OMM_OUTB_DPORT1_A:
    case OMM_OUTB_DPORT2_A:
        mask = COM_DIOCR_DIRA;
        break;

    case ATH_OUTB_DPORT_B:
    case RMM_OUTB_DPORT_B:
    case OMM_OUTB_DPORT1_B:
    case OMM_OUTB_DPORT2_B:
        mask = COM_DIOCR_DIRB;
        break;

    case ATH_OUTB_DPORT_C:
    case RMM_OUTB_DPORT_C:
    case OMM_OUTB_DPORT1_C:
    case OMM_OUTB_DPORT2_C:
        mask = COM_DIOCR_DIRC;
        break;

    }

    dat = *cd;
    if( dir == 0 )
        dat &= ~mask;
    else if( dir == 1 )
        dat |= mask;
    else
    {
        printf("dioSetCfg:unknown direction %d\n",dir);
        return;
    }

    dat |= COM_DIOCR_DIOCTL;
    *cd = dat;
    putdaq(base,co,dat);
}

unsigned char dioGetDir(unsigned short base,char ro,char dat)
{
unsigned char mask=0;

    switch( ro )
    {
    case ATH_OUTB_DPORT_A:
    case RMM_OUTB_DPORT_A:
    case OMM_OUTB_DPORT1_A:
    case OMM_OUTB_DPORT2_A:
        mask = COM_DIOCR_DIRA;
        break;

    case ATH_OUTB_DPORT_B:
    case RMM_OUTB_DPORT_B:
    case OMM_OUTB_DPORT1_B:
    case OMM_OUTB_DPORT2_B:
        mask = COM_DIOCR_DIRB;
        break;

    case ATH_OUTB_DPORT_C:
    case RMM_OUTB_DPORT_C:
    case OMM_OUTB_DPORT1_C:
    case OMM_OUTB_DPORT2_C:
        mask = COM_DIOCR_DIRC;
        break;
    }

    return( (dat & mask)?1:0);
}

void dioWrite(unsigned short base,char ro,unsigned char dat)
{
    putdaq(base,ro,dat);
}

void dioRead(unsigned short base,char ro,unsigned char* dat)
{
    *dat = getdaq(base,ro);
}


/****************************************************************************
 * Define Analog Output public interface methods
 ****************************************************************************/
void dacWrite(unsigned short base,int type,int chan,int dat)
{
short val;
unsigned char lsb,msb,off;


    switch( type )
    {
    case DSC_ATHENA:
        val = ((float)dat / COM_ANA_VAL / 10.0 * 2048.0 + 2048.5);
        lsb = val & 0xFF;
        msb = (chan << 6) | ((val >> 8) & 0xFF);

        putdaq(base,ATH_OUTB_DAC_LSB,lsb);
        putdaq(base,ATH_OUTB_DAC_MSB,msb);
        while(getdaq(ATH_DAQ_BASE,ATH_INB_ANALOG_INPUT_STATUS) & ATH_ANALOG_INPUT_STATUS_DACBSY);
        break;

    case DSC_RMM416:
        val = ((float)dat / COM_ANA_VAL / 10.0 * 32768);
        lsb = val & 0xFF;
        msb = (val & 0xFF00) / 256;

        off = (chan * 2);
        putdaq(base,off,lsb);
        putdaq(base,off+1,msb);
        off = getdaq(base,0);
        break;
    }

}


/****************************************************************************
 * Define Watchdog Timer public interface methods
 ****************************************************************************/
void wdtWrite(unsigned char reg,unsigned char dat)
{
    athena_wdt[reg] = dat;
    putdaq(ATH_WDT_BASE,reg,dat);
}

unsigned char wdtRead(unsigned char reg)
{
unsigned char data;

    switch(reg)
    {
    case ATH_INPB_WDT_TRIGGER:
    case ATH_INPB_WDT_COUNTER:
        data = athena_wdt[reg];
        break;

    default:
        data = getdaq(ATH_WDT_BASE,reg);
    }

    return(data);
}
