#ifndef _dscsuph
#define _dscsuph
/*

                          Argonne National Laboratory
                            APS Operations Division
                     Beamline Controls and Data Acquisition

                    Support for Diamond Systems Incorporated
		         Athena Single-Board Computers



 -----------------------------------------------------------------------------
                                COPYRIGHT NOTICE
 -----------------------------------------------------------------------------
   Copyright (c) 2002 The University of Chicago, as Operator of Argonne
      National Laboratory.
   Copyright (c) 2002 The Regents of the University of California, as
      Operator of Los Alamos National Laboratory.
   Synapps Versions 4-5
   and higher are distributed subject to a Software License Agreement found
   in file LICENSE that is included with this distribution.
 -----------------------------------------------------------------------------

 Description
    This module provides support for the Athena single board computers
    provided by Diamond Systems Incorporated.

 Developer notes:

 Source control info:
    Modified by:    $Author: dkline $
                    $Date: 2006/07/31 14:57:37 $
                    $Revision: 1.6 $

 =============================================================================
 History:
 Author: David M. Kline
 -----------------------------------------------------------------------------
 2005-Nov-14  DMK  Development started.
 2005-Nov-15  DMK  Initial development complete.
 2005-Nov-17  DMK  Including functions for WDT.
 2006-Feb-28  DMK  Added support for Ruby-MM-416.
                   Made methods more generic.
 2006-May-15  DMK  Removed dscInitBoard().
 2006-May-20  DMK  Added support for Pearl-MM.
 2006-Jun-03  DMK  Added support for Onyx-MM.
 2006-Jul-31  DMK  Added support for IR-104.
 -----------------------------------------------------------------------------

*/


/* System related include files */
#include <sys/io.h>

/* Board definition include files */
#include "boardtypes.h"


/******************************************************************************
 * Define common symbolic constants between boards
 ******************************************************************************/

/* Analog I/O conversion factors */
#define COM_ANA_VAL         (1000.0)

/* Digital I/O */
#define COM_DIO_REGS        (3)
#define COM_DIOCR_DIRA      (0x10)
#define COM_DIOCR_DIRB      (0x02)
#define COM_DIOCR_DIRC      (COM_DIOCR_DIRCH|COM_DIOCR_DIRCL)
#define COM_DIOCR_DIRCL     (0x01)
#define COM_DIOCR_DIRCH     (0x08)
#define COM_DIOCR_DIOCTL    (0x80)


/******************************************************************************
 * Define symbolic constants for Athena's DAQ
 ******************************************************************************/

#define ATH_ADC_CHANS       (16)
#define ATH_ADC_FIFOMAX     (48)

/* Define IO register map symbolic constants */
#define ATH_INB_ADC_LSB                        (0)
#define ATH_INB_ADC_MSB                        (1)
#define ATH_INB_ADC_CHANNEL                    (2)
#define ATH_INB_ANALOG_INPUT_STATUS            (3)
#define ATH_INB_INTERRUPT_DMA_COUNTER_CONTROL  (4)
#define ATH_INB_FIFO_THRESHOLD                 (5)
#define ATH_INB_FIFO_DEPTH                     (6)
#define ATH_INB_ANALOG_OPERATION_STATUS        (7)
#define ATH_INB_DPORT_A                        (8)
#define ATH_INB_DPORT_B                        (9)
#define ATH_INB_DPORT_C                        (10)
#define ATH_INB_DIOCR                          (11)
#define ATH_INB_COUNTER_TIMER_LOW              (12)
#define ATH_INB_COUNTER_TIMER_MIDDLE           (13)
#define ATH_INB_COUNTER_TIMER_HIGH             (14)

#define ATH_OUTB_COMMAND                       (0)
#define ATH_OUTB_ADC_CHANNEL                   (2)
#define ATH_OUTB_ANALOG_INPUT_GAIN             (3)
#define ATH_OUTB_INTERRUPT_DMA_COUNTER_CONTROL (4)
#define ATH_OUTB_FIFO_THRESHOLD                (5)
#define ATH_OUTB_DAC_LSB                       (6)
#define ATH_OUTB_DAC_MSB                       (7)
#define ATH_OUTB_DPORT_A                       (8)
#define ATH_OUTB_DPORT_B                       (9)
#define ATH_OUTB_DPORT_C                       (10)
#define ATH_OUTB_DIOCR                         (11)
#define ATH_OUTB_COUNTER_TIMER_LOW             (12)
#define ATH_OUTB_COUNTER_TIMER_MIDDLE          (13)
#define ATH_OUTB_COUNTER_TIMER_HIGH            (14)
#define ATH_OUTB_COUNTER_CONTROL               (15)

/* Define command register symbolic constants */
#define ATH_CMD_CLRA     (0x01)
#define ATH_CMD_CLRD     (0x02)
#define ATH_CMD_CLRT     (0x04)
#define ATH_CMD_CLRDMA   (0x08)
#define ATH_CMD_RSTFIFO  (0x10)
#define ATH_CMD_RSTDA    (0x20)
#define ATH_CMD_RSTBRD   (0x40)
#define ATH_CMD_STRTAD   (0x80)
#define ATH_CMD_INIT     (ATH_CMD_RSTBRD|ATH_CMD_RSTFIFO|ATH_CMD_RSTDA)

/* Define analog input gain/status register symbolic constants */
#define ATH_ANALOG_INPUT_GAIN_G0           (0x01)
#define ATH_ANALOG_INPUT_GAIN_G1           (0x02)
#define ATH_ANALOG_INPUT_GAIN_SCANEN       (0x04)

#define ATH_ANALOG_INPUT_STATUS_G0         (0x01)
#define ATH_ANALOG_INPUT_STATUS_G1         (0x02)
#define ATH_ANALOG_INPUT_STATUS_SCANEN     (0x04)
#define ATH_ANALOG_INPUT_STATUS_OVF        (0x08)
#define ATH_ANALOG_INPUT_STATUS_DACBSY     (0x10)
#define ATH_ANALOG_INPUT_STATUS_WAIT       (0x20)
#define ATH_ANALOG_INPUT_STATUS_SD         (0x40)
#define ATH_ANALOG_INPUT_STATUS_STS        (0x80)

/* Define interrupt, DMA, counter control register symbolic constants */
#define ATH_IDC_CONTROL_AINTE              (0x01)
#define ATH_IDC_CONTROL_DINTE              (0x02)
#define ATH_IDC_CONTROL_TINTE              (0x04)
#define ATH_IDC_CONTROL_DMAEN              (0x08)
#define ATH_IDC_CONTROL_ADCLK              (0x10)
#define ATH_IDC_CONTROL_CKFRQ0             (0x20)
#define ATH_IDC_CONTROL_CKFRQ1             (0x40)
#define ATH_IDC_CONTROL_CKSEL1             (0x80)

/* Define Watchdog Timer register symbolic constants */
#define ATH_INPB_WDT_TRIGGER        (0)
#define ATH_INPB_WDT_COUNTER        (1)
#define ATH_INPB_WDT_CONTROL        (2)
#define ATH_INPB_WDT_SELECT         (3)

#define ATH_OUTB_WDT_TRIGGER        (0)
#define ATH_OUTB_WDT_COUNTER        (1)
#define ATH_OUTB_WDT_CONTROL        (2)
#define ATH_OUTB_WDT_SELECT         (3)

/* Define trigger register symbolic constants */
#define ATH_WDT_TRIGGER             (0x10)

/* Define counter register symbolic constants */
#define ATH_WDT_COUNTER             (0xF0)

/* Define control register symbolic constants */
#define ATH_WDT_CONTROL             (0xF0)
#define ATH_WDT_CONTROL_WDEDGE      (0x10)
#define ATH_WDT_CONTROL_WDSMI       (0x20)
#define ATH_WDT_CONTROL_WDOEN       (0x40)
#define ATH_WDT_CONTROL_WDIEN       (0x80)

/* Define chip select register symbolic constants */
#define ATH_WDT_SELECT              (0xF0)
#define ATH_WDT_SELECT_WDEN         (0x10)
#define ATH_WDT_SELECT_FPGAEN       (0x20)
#define ATH_WDT_SELECT_COM3EN       (0x40)
#define ATH_WDT_SELECT_COM4EN       (0x80)


/******************************************************************************
 * Define symbolic constants for Ruby-416 (RMM) DAQ
 ******************************************************************************/

/* Define IO register map symbolic constants */
#define RMM_OUTB_DAC0_LSB           (0)
#define RMM_OUTB_DAC0_MSB           (1)
#define RMM_OUTB_DAC1_LSB           (2)
#define RMM_OUTB_DAC1_MSB           (3)
#define RMM_OUTB_DAC2_LSB           (4)
#define RMM_OUTB_DAC2_MSB           (5)
#define RMM_OUTB_DAC3_LSB           (6)
#define RMM_OUTB_DAC3_MSB           (7)
#define RMM_OUTB_RESET              (8)
#define RMM_OUTB_EXTENA             (9)
#define RMM_OUTB_DPORT_A            (12)
#define RMM_OUTB_DPORT_B            (13)
#define RMM_OUTB_DPORT_C            (14)
#define RMM_OUTB_DIOCR              (15)

#define RMM_INB_DPORT_A             (12)
#define RMM_INB_DPORT_B             (13)
#define RMM_INB_DPORT_C             (14)
#define RMM_INB_DIOCR               (15)


/******************************************************************************
 * Define symbolic constants for Pearl-MM (PMM) DAQ
 ******************************************************************************/

/* Define IO register map symbolic constants */
#define PMM_OUTB_DPORT_A            (0)
#define PMM_OUTB_DPORT_B            (1)
#define PMM_OUTB_DIOCR              (-1)


/******************************************************************************
 * Define symbolic constants for Onyx-MM (PMM) Digital IO
 ******************************************************************************/

/* Define IO register map symbolic constants */
#define OMM_OUTB_DPORT1_A           (0)
#define OMM_OUTB_DPORT1_B           (1)
#define OMM_OUTB_DPORT1_C           (2)
#define OMM_OUTB_DIOCR1             (3)
#define OMM_OUTB_DPORT2_A           (4)
#define OMM_OUTB_DPORT2_B           (5)
#define OMM_OUTB_DPORT2_C           (6)
#define OMM_OUTB_DIOCR2             (7)

/******************************************************************************
 * Define symbolic constants for GPIO-MM (GMM) Digital IO
 ******************************************************************************/

/* Define IO register map symbolic constants */
#define GMM_PIO_DPORT1_A            (0)
#define GMM_PIO_DPORT1_B            (1)
#define GMM_PIO_DPORT1_C            (2)
#define GMM_PIO_DIOCR1              (3)
#define GMM_PIO_DPORT2_A            (4)
#define GMM_PIO_DPORT2_B            (5)
#define GMM_PIO_DPORT2_C            (6)
#define GMM_PIO_DIOCR2              (7)

#define GMM_FIX_DPORT1_A            (2)
#define GMM_FIX_DPORT1_B            (2)
#define GMM_FIX_DIOCR               (-1)

#define GMM_AUX_DPORT               (14)
#define GMM_AUX_DIOCR               (14)
#define GMM_AUX_DIR1                (0x10)
#define GMM_AUX_DIR2                (0x20)
#define GMM_AUX_DIR3                (0x40)
#define GMM_AUX_DIR4                (0x80)

/******************************************************************************
 * Define symbolic constants for IR-104 (IR4) Digital IO
 ******************************************************************************/

/* Define IO register map symbolic constants */
#define IR4_OUTB_DPORT_A            (0)
#define IR4_OUTB_DPORT_B            (1)
#define IR4_OUTB_DPORT_C            (2)
#define IR4_OUTB_DIOCR              (-1)

#define IR4_INPB_DPORT_A            (4)
#define IR4_INPB_DPORT_B            (5)
#define IR4_INPB_DPORT_C            (6)
#define IR4_INPB_DIOCR              (-1)


/******************************************************************************
 * Define access macros
 ******************************************************************************/
#define getdaq(b,o) inb((b)+(o))
#define putdaq(b,o,v) outb((v),(b)+(o))


#ifdef __cplusplus
extern "C" {
#endif

extern void dioSetDir(unsigned short base,char ro,char co,unsigned char* cd,char dir);
extern unsigned char dioGetDir(unsigned short base,char ro,char dat);
extern void dioWrite(unsigned short base,char ro,unsigned char dat);
extern void dioRead(unsigned short base,char ro,unsigned char* dat);
extern void dacWrite(unsigned short base,int type,int chan,int data);
extern void wdtWrite(unsigned char reg,unsigned char dat);
extern unsigned char wdtRead(unsigned char reg);

#ifdef __cplusplus
}
#endif

#endif /* _dscsuph */
