/*

                          Argonne National Laboratory
                            APS Operations Division
                     Beamline Controls and Data Acquisition

                   Support for Sensoray's Smart A/D Model 518



 -----------------------------------------------------------------------------
                                COPYRIGHT NOTICE
 -----------------------------------------------------------------------------
   Copyright (c) 2002 The University of Chicago, as Operator of Argonne
      National Laboratory.
   Copyright (c) 2002 The Regents of the University of California, as
      Operator of Los Alamos National Laboratory.
   Synapps Versions 4-5
   and higher are distributed subject to a Software License Agreement found
   in file LICENSE that is included with this distribution.
 -----------------------------------------------------------------------------

 Description
    This module provides the low-level support for Sensoray's PC104-based
    Smart A/D Model 518. 

 Developer notes:

 Source control info:
    Modified by:    $Author: dkline $
                    $Date: 2006/06/21 19:19:01 $
                    $Revision: 1.4 $

 =============================================================================
 History:
 Author: David M. Kline
 -----------------------------------------------------------------------------
 2006-Mar-09  DMK  Development begins.
 2006-May-15  DMK  Removed senInitBoard().
 -----------------------------------------------------------------------------

*/

/* System related include files */
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>


/* DSC related include files */
#include "sensup.h"


/* Forward struct declarations */
typedef struct Dsc Dsc;


/* Declare sensor descriptor structure */
struct Dsc
{
    float scalar;
    unsigned char type;
};


/* Define local variants */
static Dsc senTypes[] =
{
    /* Voltage                */ {0.000005,0x17},{0.00002,0x16},{0.0002,0x15},
    /* Resistence             */ {0.02,0x0A},{0.125,0x14},{31.0,0x20},{1.0,0x0C},
    /* Thermocouple BCEJKNTSR */ {0.10,0x24},{0.10,0x23},{0.10,0x01},{0.10,0x1B},{0.10,0x1C},{0.10,0x22},{0.10,0x1D},{0.10,0x1E},{0.10,0x1F},
    /* RTD                    */ {0.10,0x2C},{0.05,0x18},{0.05,0x19},{0.0125,0x2A},{0.0125,0x2B},{0.10,0x21},
    /* Thermister             */ {0.01,0x1A},
    /* Current loop           */ {0.01,0x11},
    /* pressure/Strain gage   */ {1.0,0x0F},
    /* Disabled channel       */ {1.0,0x13}
};


/****************************************************************************
 * Define public interface methods
 ****************************************************************************/


/*
 * Low-level primitive methods
 */
char senReadB(unsigned short base)
{
    while( (getsen(base,SEN_INPB_STS) & SEN_STS_DAV) == 0);
    return( getsen(base,SEN_INPB_DAT) & 0xFF );
}

void senWriteB(unsigned short base,unsigned char dat)
{
    while( (getsen(base,SEN_INPB_STS) & SEN_STS_CRMT) == 0);
    putsen(base,SEN_OUTB_CMD,dat);
}

short senReadW(unsigned short base)
{
    short hb = senReadB(base) << 8;
    return( hb | (senReadB(base) & 0xFF) );
}

void senWriteW(unsigned short base,short dat)
{
    senWriteB(base,(dat & 0xFF00) >> 8);
    senWriteB(base,(dat & 0x00FF));
}


/*
 * High-level composite methods
 */
void senReset(unsigned short base)
{
    putsen(base,SEN_OUTB_CTL,0);
    while( getsen(base,SEN_INPB_STS) & SEN_STS_FAULT );
}

short senGetStatus(unsigned short base)
{
    return( getsen(base,SEN_INPB_STS) );
}

short senGetChan(unsigned short base,short chn)
{
    senWriteB(base,chn);
    return( senReadW(base) );
}

void senGetChans(unsigned short base,short* dat)
{
    int i;

    senWriteB(base,88);
    for(i=0; i<SEN_CHNS; ++i) dat[i] = senReadW(base);
}

float senGetTemp(unsigned short base)
{
    senWriteB(base,240);
    senWriteB(base,5);
    senWriteB(base,0);

    return( 1.8 * ((float)senReadW(base) / 10.0) + 32.0 );
}

float senGetVersion(unsigned short base)
{
    senWriteB(base,240);
    senWriteB(base,5);
    senWriteB(base,0);

    return( (float)(senReadW(base)&0xFF) / 100.0 );
}

short senGetProductID(unsigned short base)
{
    senWriteB(base,240);
    senWriteB(base,4);
    senWriteB(base,0);

    return( senReadW(base) );
}

short senGetAlarmFlags(unsigned short base)
{
    senWriteB(base,48);
    return( senReadW(base) );
}

void senSetSpeed(unsigned short base)
{
    senWriteB(base,240);
    senWriteB(base,8);
    senWriteB(base,0);
}

short senGetType(short typ)
{
    return( senTypes[typ].type );
}

float senGetScalar(short typ)
{
    return( senTypes[typ].scalar );
}

void senSetType(unsigned short base,short chn,short typ)
{
    senWriteB(base,16+chn);
    senWriteB(base,senTypes[typ].type);
}
