/*

                          Argonne National Laboratory
                            APS Operations Division
                     Beamline Controls and Data Acquisition

                    Support for Diamond Systems Incorporated
		 Setup Application for the Emerald-MM-8P Module



 -----------------------------------------------------------------------------
                                COPYRIGHT NOTICE
 -----------------------------------------------------------------------------
   Copyright (c) 2002 The University of Chicago, as Operator of Argonne
      National Laboratory.
   Copyright (c) 2002 The Regents of the University of California, as
      Operator of Los Alamos National Laboratory.
   Synapps Versions 4-5
   and higher are distributed subject to a Software License Agreement found
   in file LICENSE that is included with this distribution.
 -----------------------------------------------------------------------------

 Description
    This module provides a mechanism to setup the Emerald-MM-8P modules during
    system initialization.

 Developer notes:
    Development will continue when a need arises to configure the board from
    the command line.

 Source control info:
    Modified by:    $Author: dkline $
                    $Date: 2006/03/29 16:47:25 $
                    $Revision: 1.1 $

 =============================================================================
 History:
 Author: David M. Kline
 -----------------------------------------------------------------------------
 2006-Mar-22  DMK  Development started (taken from DSC's eppconf application).
 2006-Mar-29  DMK  Development will continue when a need arises to configure
                   the board from the command line.
 -----------------------------------------------------------------------------

*/


/* System related include files */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/io.h>
#include <unistd.h>


/* Declare local variants */
static int base;
static int port;
static int portirq;
static int portaddr;
static char portproto[8];
static int help;
static int enable;
static int disable;
static int optcount;


/* Declare forward references */
static void usage(void);
static void setPort(void);
static void processOptions(void);
static int options(int argc,char** argv);

static void enablePort(void);
static void disablePort(void);

static void readProm(int addr,int* data);
static void writeProm(int addr,int data);


/****************************************************************************
 * Define primary entry point
 ****************************************************************************/
int main(int argc,char** argv)
{
int s;

    /* Change IO privilege level */
    iopl(3);

    /* Collect options */
    s = options(argc,argv);
    if(s) return(s);

    /* Process options */
    processOptions();

    return(0);
}

static void usage(void)
{
    printf("\nsetemm version 1.0 %s\n\n",__DATE__);

    printf("usage:\tsetemm [-base b [-port p -addr a -irq n -proto m] [-ena] [-dis] ] [-help] \n");

    printf("\nAvailable options:\n");
    printf("\t-base\t- Board base address (hex)\n");
    printf("\t-port\t- Number (0..7)\n");
    printf("\t-addr\t- Address (hex)\n");
    printf("\t-irq\t- IRQ level (2..7)\n");
    printf("\t-proto\t- Protocol (232,422,485)\n");
    printf("\t-ena\t- Enable serial ports\n");
    printf("\t-dis\t- Disable serial ports\n");
    printf("\t-help\t- Display command usage\n");

    printf("\nUsage a leading '0x' for hex numbers\n");
}

static int options(int argc,char** argv)
{
    int i;

    if( argc == 1 )
    {
        usage();
        return(0);
    }

    for(i=1; i<argc; ++i)
    {
        if( strcmp("-base",argv[i]) == 0 )
        {
            ++optcount;
            sscanf(argv[++i],"%x",&base);
            continue;
        }

        if( strcmp("-port",argv[i]) == 0 )
        {
            ++optcount;
            sscanf(argv[++i],"%d",&port);
            continue;
        }

        if( strcmp("-addr",argv[i]) == 0 )
        {
            ++optcount;
            sscanf(argv[++i],"%x",&portaddr);
            continue;
        }

        if( strcmp("-irq",argv[i]) == 0 )
        {
            ++optcount;
            sscanf(argv[++i],"%d",&portirq);
            continue;
        }

        if( strcmp("-proto",argv[i]) == 0 )
        {
            ++optcount;
            strcpy(portproto,argv[++i]);
            continue;
        }

        if( strcmp("-ena",argv[i]) == 0 )
        {
            enable = 1;
            continue;
        }

        if( strcmp("-dis",argv[i]) == 0 )
        {
            disable = 1;
            continue;
        }

        if( strcmp("-help",argv[i]) == 0 )
        {
            help = 1;
            continue;
        }

    }

    return(0);
}

static void processOptions(void)
{
    if( enable && disable )
    {
        printf("setemm: cannot enable/disable a port\n");
        return;
    }

    if( help ) {usage(); return;}
    if( enable ) {enablePort(); return;}
    if( disable ) {disablePort(); return;}

    setPort();
}

static void enablePort(void)
{
    outb(0x80,base);
}

static void disablePort(void)
{
    outb(0x00,base);
}

static void setPort(void)
{
    int p,r,s;

    if(optcount < 5)
    {
        printf("setemm: not enoungh options specified\n");
        return;
    }

    r = 16 + (port/4);
    if(strcmp(portproto,"232") == 0)
        p = 0 << (port % 7);
    else if(strcmp(portproto,"422") == 0)
        p = 1;
    else if(strcmp(portproto,"485") == 0)
        p = 3;
    else
    {
        printf("setemm: invalid protocol \"%s\"\n",portproto);
        return;
    }

#if 0
    writeProm(port,portaddr);
    writeProm(port+8,portirq);

    readProm(r,&s);
    writeProm(r,(s | p));
#else
printf("setPort: base-0x%X,port-%d,portaddr-0x%X,irq-%d,proto-%s\n",base,port,portaddr,portirq,portproto);
printf("setPort: p-%d,r-%d,s-%d\n",p,r,s);
#endif
}

static void readProm(int addr,int* data)
{
    outb(addr,base+4);
    while(inb(base+4) >= 128);
    *data = inb(base+5) << 3;
}

static void writeProm(int addr,int data)
{
    outb((data>>3),base+5);
    outb(addr+128,base+4);
    while(inb(base+4) >= 128);
}
